﻿/* CircuitNetwork.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics;
using Popolo.Numerics.MatrixOperation;

namespace Popolo.HVAC.Circuit
{
  /// <summary>回路網</summary>
  public class CircuitNetwork
  {

    #region インスタンス変数・プロパティ

    /// <summary>流路リスト</summary>
    private List<ICircuitBranch> branches = new List<ICircuitBranch>();

    /// <summary>節点リスト</summary>
    private List<CircuitNode> nodes = new List<CircuitNode>();

    /// <summary>節点を取得する</summary>
    public CircuitNode[] Nodes { get { return nodes.ToArray(); } }

    /// <summary>圧力基準となる節点を取得する</summary>
    public CircuitNode BasePressureNode { get; private set; }

    /// <summary>基準圧力[kPa]を取得する</summary>
    public double BasePressure { get; private set; }

    /// <summary>反復計算回数を取得する</summary>
    public int Iteration { get; private set; }

    /// <summary>誤差を取得する</summary>
    public double Error { get; private set; }

    #endregion

    #region 回路網構築処理

    /// <summary>節点を追加する</summary>
    /// <returns>節点番号</returns>
    public CircuitNode AddNode()
    {
      CircuitNode nd = new CircuitNode();
      nodes.Add(nd);
      if (nodes.Count == 1) SetBasePressure(nd, 0);
      return nd;
    }

    /// <summary>節点を削除する</summary>
    /// <param name="node"></param>
    /// <returns>削除成功の真偽</returns>
    public bool RemoveNode(CircuitNode node)
    {
      if (nodes.Contains(node))
      {
        //接続中の節点は削除不可
        foreach (ICircuitBranch br in branches)
          if (br.UpStreamNode == node || br.DownStreamNode == node) return false;
        nodes.Remove(node);
        BasePressureNode = null;
        return true;
      }
      else return false;
    }

    /// <summary>流路で節点を接続する</summary>
    /// <param name="branch">流路</param>
    /// <param name="ndFrom">接続元</param>
    /// <param name="ndTo">接続先</param>
    public void ConnectNode(ICircuitBranch branch, CircuitNode ndFrom, CircuitNode ndTo)
    {
      if (!nodes.Contains(ndFrom) || !nodes.Contains(ndTo)) throw new Exception("Invalid Node Error");
      if (!branches.Contains(branch)) branches.Add(branch);
      //接続処理
      branch.UpStreamNode = ndFrom;
      branch.DownStreamNode = ndTo;
      ndFrom.addOutFlowBranch(branch);
      ndTo.addInFlowBranch(branch);
    }

    /// <summary>流路を削除する</summary>
    /// <param name="branch">流路</param>
    public void RemoveBranch(ICircuitBranch branch)
    {
      branch.UpStreamNode.removeBranch(branch);
      branch.DownStreamNode.removeBranch(branch);
      branches.Remove(branch);
    }

    /// <summary>基準圧力を設定する</summary>
    /// <param name="node">節点</param>
    /// <param name="pressure">基準圧力[kPa]</param>
    public void SetBasePressure(CircuitNode node, double pressure)
    {
      BasePressureNode = node;
      BasePressure = pressure;
    }

    #endregion

    #region 収束計算処理

    /// <summary>回路網を解く</summary>
    public void Solve()
    {
      //入力値ベクトルを作成する
      List<double> inp = new List<double>();
      for (int i = 0; i < nodes.Count; i++)
      {
        if (nodes[i].IsPressureFixed) inp.Add(nodes[i].Inflow);
        else inp.Add(nodes[i].Pressure);
      }
      IVector inputs = new Vector(inp.Count);
      for (int i = 0; i < inp.Count; i++) inputs[i] = inp[i];

      //ニュートン法で解く
      int iter;
      double err;
      MultiRoots.Newton(errorFNC, ref inputs, 1e-5, 1e-5, 100, out iter, out err);
      Iteration = iter;
      Error = err;

      //基準点に合わせて絶対圧力を調整する
      double dp = BasePressure - BasePressureNode.Pressure;
      foreach (CircuitNode nd in nodes) nd.Pressure += dp;
    }

    /// <summary>誤差関数</summary>
    /// <param name="inputs">入力値</param>
    /// <param name="outputs">出力値</param>
    private void errorFNC(IVector inputs, ref IVector outputs)
    {
      //節点圧力および固定流出入量を設定
      int indx = 0;
      for (int i = 0; i < nodes.Count; i++)
      {
        if (nodes[i].IsPressureFixed)
        {
          nodes[i].Inflow = inputs[indx];
          indx++;
        }
        else
        {
          nodes[i].Pressure = inputs[indx];
          indx++;
        }
      }

      //流路流量を更新
      for (int i = 0; i < branches.Count; i++) branches[i].UpdateFlowRateFromNodePressureDifference();

      //節点の質量保存誤差を計算 
      indx = 0;
      for (int i = 0; i < nodes.Count; i++)
      {
        outputs[i] = nodes[i].IntegrateFlow();
        indx++;
      }
    }

    #endregion

  }
}
