﻿/* CircuitNode.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System.Collections.Generic;

namespace Popolo.HVAC.Circuit
{
  /// <summary>回路網の節点</summary>
  public class CircuitNode
  {

    #region インスタンス変数・プロパティ

    /// <summary>接続元流路リスト</summary>
    private List<ICircuitBranch> outFlowBrchs = new List<ICircuitBranch>();

    /// <summary>接続先流路リスト</summary>
    private List<ICircuitBranch> inFlowBrchs = new List<ICircuitBranch>();

    /// <summary>圧力固定か否か</summary>
    public bool IsPressureFixed { get; set; } = false;
    
    /// <summary>圧力[kPa]を設定・取得する</summary>
    public double Pressure { get; set; }

    /// <summary>流入量[m3/s]を設定・取得する</summary>
    public double Inflow { get; set; }

    #endregion

    #region internalメソッド
    
    /// <summary>接続元流路を追加する</summary>
    /// <param name="branch">接続元流路</param>
    internal void addOutFlowBranch(ICircuitBranch branch) { outFlowBrchs.Add(branch); }

    /// <summary>接続先流路を追加する</summary>
    /// <param name="branch">接続先流路</param>
    internal void addInFlowBranch(ICircuitBranch branch) { inFlowBrchs.Add(branch); }

    /// <summary>流路を削除する</summary>
    /// <param name="branch">流路</param>
    internal void removeBranch(ICircuitBranch branch)
    {
      outFlowBrchs.Remove(branch);
      inFlowBrchs.Remove(branch);
    }

    #endregion

    #region publicメソッド

    /// <summary>流出入収支[m3/s]を積算する</summary>
    /// <returns>流出入収支[m3/s]</returns>
    /// <remarks>質量保存則判定用</remarks>
    public double IntegrateFlow()
    {
      double sum = Inflow;
      foreach (ICircuitBranch br in outFlowBrchs) sum -= br.VolumetricFlowRate;
      foreach (ICircuitBranch br in inFlowBrchs) sum += br.VolumetricFlowRate;
      return sum;
    }

    /// <summary>接続先流路リストを取得する</summary>
    /// <returns>接続先流路リスト</returns>
    public ImmutableCircuitBranch[] GetInFlowBranches() { return inFlowBrchs.ToArray(); }

    /// <summary>接続元流路リストを取得する</summary>
    /// <returns>接続元流路リスト</returns>
    public ImmutableCircuitBranch[] GetOutFlowBranches() { return outFlowBrchs.ToArray(); }

    #endregion

  }
}