﻿/* FluidMachinery.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.Circuit
{
  /// <summary>流体機械に関する抽象クラス</summary>
  public abstract class FluidMachinery : ImmutableFluidMachinery, ICircuitBranch
  {

    #region インスタンス変数・プロパティ

    /// <summary>効率特性係数</summary>
    protected double[] efficiencyCoefficient;

    /// <summary>揚程特性係数</summary>
    protected double[] pressureCoefficient;

    /// <summary>抵抗係数[kPa/(m3/s)^2]</summary>
    protected double resistanceCoefficient;

    /// <summary>最小回転数比[-]</summary>
    private double minimumRotationRatio = 0.4;

    /// <summary>体積流量[m3/s]を設定・取得する</summary>
    public double VolumetricFlowRate { get; internal set; }

    /// <summary>回転数比[-]を設定・取得する</summary>
    public double RotationRatio { get; set; } = 1.0;

    /// <summary>最小回転数比[-]を設定・取得する</summary>
    public double MinimumRotationRatio
    {
      get { return minimumRotationRatio; }
      set { minimumRotationRatio = Math.Max(Math.Min(value, 1.0), 0.05); }
    }

    /// <summary>全圧・揚程[kPa]を取得する</summary>
    public double Pressure { get; protected set; }

    /// <summary>実揚程[kPa]を取得する</summary>
    public double ActualHead { get; private set; }

    /// <summary>インバータ機か否か</summary>
    public bool HasInverter { get; private set; }

    /// <summary>設計流量[m3/s]を取得する</summary>
    public double DesignFlowRate { get; private set; }

    /// <summary>定格軸動力[kW]を取得する</summary>
    public double NominalShaftPower { get; private set; }

    /// <summary>モーター効率[-]を取得する</summary>
    public double MotorEfficiency { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nominalPressure">定格全圧・揚程[kPa]</param>
    /// <param name="nominalFlowRate">定格流量[m3/s]</param>
    /// <param name="designPressure">設計全圧・揚程[kPa]</param>
    /// <param name="designFlowRate">設計流量[m3/s]</param>
    /// <param name="actualHead">実揚程[kPa]</param>
    /// <param name="hasInverter">インバータ機か否か</param>
    protected FluidMachinery
      (double nominalPressure, double nominalFlowRate, double designPressure,
      double designFlowRate, double actualHead, bool hasInverter)
    {
      DesignFlowRate = designFlowRate;
      ActualHead = actualHead;
      HasInverter = hasInverter;

      //定格軸動力[kW]を計算
      NominalShaftPower = nominalPressure * nominalFlowRate;
      MotorEfficiency = GetMotorEfficiency(NominalShaftPower);

      //抵抗係数[kPa/(m3/s)^2]を計算
      resistanceCoefficient = (designPressure - actualHead) / (designFlowRate * designFlowRate);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>流体機械効率[-]を取得する</summary>
    public virtual double GetFluidMachineryEfficiency()
    { return Math.Max(GetFluidMachineryEfficiency(VolumetricFlowRate, efficiencyCoefficient), 0.05); }

    /// <summary>インバータ効率[-]を取得する</summary>
    public virtual double GetInverterEfficiency()
    {
      if (HasInverter)
        return GetInverterEfficiency
          (VolumetricFlowRate * Pressure / NominalShaftPower, Math.Min(1, RotationRatio));
      else return 1.0;
    }

    /// <summary>総合効率[-]を取得する</summary>
    public virtual double GetTotalEfficiency()
    { return MotorEfficiency * GetFluidMachineryEfficiency() * GetInverterEfficiency(); }

    /// <summary>消費電力[kW]を取得する</summary>
    public virtual double GetElectricConsumption()
    { return VolumetricFlowRate * Pressure / GetTotalEfficiency(); }

    /// <summary>体積流量と圧力から回転数比を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="pressure">圧力[kPa]</param>
    internal void updateWithFlowRateAndPressure(double flowRate, double pressure)
    {
      VolumetricFlowRate = flowRate;
      Pressure = pressure;
      RotationRatio = GetRotationRatio(flowRate, pressure, pressureCoefficient);
    }

    /// <summary>設定圧力とPQ特性が交わる体積流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <param name="pressure">設定圧力[kPa]</param>
    internal void updateWithRotationRatioAndPressure(double rotationRatio, double pressure)
    {
      RotationRatio = rotationRatio;
      Pressure = pressure;
      VolumetricFlowRate = GetFlowRate(rotationRatio, pressureCoefficient, pressure, DesignFlowRate);
    }

    /// <summary>抵抗曲線とPQ特性が交わる体積流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <param name="resistanceCoefficient">抵抗係数[kPa/(m3/s)^2]</param>
    /// <param name="actualHead">実揚程[kPa]</param>
    /// <param name="operatingNumber">運転台数</param>
    /// <param name="totalFlowRate">総流量[m3/s]</param>
    internal void updateWithResistanceAndRotationRatio
      (double rotationRatio, double resistanceCoefficient,
      double actualHead, int operatingNumber, out double totalFlowRate)
    {
      RotationRatio = rotationRatio;
      totalFlowRate = GetFlowRate(RotationRatio, pressureCoefficient, resistanceCoefficient,
        actualHead, operatingNumber, DesignFlowRate * operatingNumber);
      VolumetricFlowRate = totalFlowRate / operatingNumber;
      Pressure = resistanceCoefficient * totalFlowRate * totalFlowRate + actualHead;
    }

    /// <summary>抵抗曲線とPQ特性が交わる体積流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    internal void updateWithResistanceAndRotationRatio(double rotationRatio)
    {
      RotationRatio = rotationRatio;
      VolumetricFlowRate = GetFlowRate(RotationRatio, pressureCoefficient,
        resistanceCoefficient, ActualHead, 1, DesignFlowRate);
      Pressure = resistanceCoefficient * Math.Pow(VolumetricFlowRate, 2) + ActualHead;
    }

    /// <summary>体積流量と回転数比から圧力を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="rotationRatio">回転数比[-]</param>
    internal void updateWithFlowRateAndRotationRatio(double flowRate, double rotationRatio)
    {
      VolumetricFlowRate = flowRate;
      RotationRatio = rotationRatio;
      Pressure = GetPressure(flowRate, rotationRatio, pressureCoefficient);
    }

    /// <summary>停止させる</summary>
    public virtual void ShutOff()
    {
      VolumetricFlowRate = 0;
      Pressure = 0;
      RotationRatio = 0;
    }

    #endregion

    #region ICircuitBranch実装

    /// <summary>上流節点を設定・取得する</summary>
    public CircuitNode UpStreamNode { get; set; }

    /// <summary>下流節点を設定・取得する</summary>
    public CircuitNode DownStreamNode { get; set; }

    /// <summary>前後差圧から体積流量[m3/s]を計算する</summary>
    /// <returns>体積流量[m3/s]</returns>
    public void UpdateFlowRateFromNodePressureDifference()
    {
      double dp = DownStreamNode.Pressure - UpStreamNode.Pressure;
      if (pressureCoefficient[1] < dp) VolumetricFlowRate = 0;
      else VolumetricFlowRate = GetFlowRate(RotationRatio, pressureCoefficient, dp, DesignFlowRate);
    }

    #endregion

    #region public staticメソッド
    
    /// <summary>流体機械効率[-]を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="efficiencyCoef">機械効率[-]</param>
    /// <returns>流体機械効率[-]</returns>
    public static double GetFluidMachineryEfficiency(double flowRate, double[] efficiencyCoef)
    { return getPolynomial(flowRate, efficiencyCoef); }

    /// <summary>モーター効率[-]を計算する</summary>
    /// <param name="shaftPower">軸動力[kW]</param>
    /// <returns>モーター効率[-]</returns>
    public static double GetMotorEfficiency(double shaftPower)
    {
      double lnWT = Math.Log(shaftPower);
      return 8.015e-1 + (6.567e-2 + -8.724e-3 * lnWT) * lnWT;
    }

    /// <summary>インバータ効率[-]を計算する</summary>
    /// <param name="loadRate">負荷率[-]</param>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <returns>インバータ効率[-]</returns>
    public static double GetInverterEfficiency(double loadRate, double rotationRatio)
    {
      double r2 = rotationRatio * rotationRatio;
      double a1 = -0.2287 * r2 + 0.3425 * rotationRatio - 0.4923;
      double a2 = 0.4690 * r2 - 0.8248 * rotationRatio + 0.9172;
      double a3 = -0.3356 * r2 + 0.6319 * rotationRatio + 0.4756;
      return loadRate * (loadRate * a1 + a2) + a3;
    }

    /// <summary>体積流量と圧力から回転数比を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="pressureCoef">圧力特性係数</param>
    /// <returns>回転数比[-]</returns>
    public static double GetRotationRatio(double flowRate, double pressure, double[] pressureCoef)
    {
      Roots.ErrorFunction eFnc = delegate (double rRate)
      { return GetPressure(flowRate, rRate, pressureCoef) - pressure; };
      return Roots.Newton(eFnc, 0.8, 1e-5, 1e-4, 1e-4, 20);
    }

    /// <summary>設定圧力とPQ特性が交わる体積流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <param name="pressureCoef">圧力特性係数</param>
    /// <param name="pressure">設定圧力[kPa]</param>
    /// <param name="initialFlowRate">体積流量初期値[m3/s]</param>
    /// <returns>体積流量[m3/s]</returns>
    public static double GetFlowRate
      (double rotationRatio, double[] pressureCoef, double pressure, double initialFlowRate)
    {
      if (rotationRatio <= 0) return 0;
      double r2 = rotationRatio * rotationRatio;

      //締切り揚程で不足する場合には0を出力
      if (r2 * getPolynomial(0, pressureCoef) < pressure) return 0;

      Roots.ErrorFunction eFnc = delegate (double vf)
      { return r2 * getPolynomial(vf / rotationRatio, pressureCoef) - pressure; };
      Roots.ErrorFunction eFncD = delegate (double vf)
      { return r2 * getPolynomialD(vf, 1 / rotationRatio, pressureCoef); };

      return Roots.Newton(eFnc, eFncD, initialFlowRate, 1e-4, 1e-4, 20);
    }

    /// <summary>抵抗曲線とPQ特性が交わる体積流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <param name="pressureCoef">圧力特性係数</param>
    /// <param name="resistanceCoefficient">抵抗係数[kPa/(m3/s)^2]</param>
    /// <param name="actualHead">実揚程[kPa]</param>
    /// <param name="operatingNumber">運転台数</param>
    /// <param name="initialFlowRate">体積流量初期値[m3/s]</param>
    /// <returns>体積流量[m3/s]</returns>
    public static double GetFlowRate
      (double rotationRatio, double[] pressureCoef, double resistanceCoefficient,
      double actualHead, int operatingNumber, double initialFlowRate)
    {
      double r2 = rotationRatio * rotationRatio;
      Roots.ErrorFunction eFnc = delegate (double vf)
      {
        return r2 * getPolynomial
        (vf / (operatingNumber * rotationRatio), pressureCoef) - (vf * vf * resistanceCoefficient + actualHead);
      };
      Roots.ErrorFunction eFncD = delegate (double vf)
      {
        return r2 * getPolynomialD(vf, 1 / (operatingNumber * rotationRatio), pressureCoef)
          - 2 * vf * resistanceCoefficient;
      };
      return Roots.Newton(eFnc, eFncD, initialFlowRate, 1e-4, 1e-4, 20);
    }

    /// <summary>体積流量と回転数比から圧力を計算する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    /// <param name="rotationRatio">回転数比[-]</param>
    /// <param name="pressureCoef">圧力特性係数</param>
    /// <returns>圧力[kPa]</returns>
    public static double GetPressure(double flowRate, double rotationRatio, double[] pressureCoef)
    { return rotationRatio * rotationRatio * getPolynomial(flowRate / rotationRatio, pressureCoef); }

    #endregion

    #region private staticメソッド

    /// <summary>多項式y=Σa[n]*x^(N-n)の計算を行う</summary>
    /// <param name="x">入力変数</param>
    /// <param name="a">係数配列</param>
    /// <returns>y=Σa[n]*x^n</returns>
    private static double getPolynomial(double x, double[] a)
    {
      double y = a[0];
      for (int i = 1; i < a.Length; i++) y = y * x + a[i];
      return y;
    }

    /// <summary>多項式y=Σa[n]*x^(N-n)*b^(N-n)の微分値の計算を行う</summary>
    /// <param name="x">入力変数</param>
    /// <param name="b">係数b</param>
    /// <param name="a">係数配列</param>
    /// <returns>多項式y=Σa[n]*x^(N-n)*b^(N-n)の微分値</returns>
    private static double getPolynomialD(double x, double b, double[] a)
    {
      int pow = a.Length - 1;
      double y = b * pow * a[0];
      for (int i = 1; i < a.Length - 1; i++) y = b * (x * y + (pow - i) * a[i]);
      return y;
    }

    #endregion

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用の流体機械</summary>
  public interface ImmutableFluidMachinery
  {
    /// <summary>流量[m3/s]を取得する</summary>
    double VolumetricFlowRate { get; }

    /// <summary>回転数比[-]を取得する</summary>
    double RotationRatio { get; }

    /// <summary>最小回転数比[-]を取得する</summary>
    double MinimumRotationRatio { get; }

    /// <summary>全圧・揚程[kPa]を取得する</summary>
    double Pressure { get; }

    /// <summary>実揚程[kPa]を取得する</summary>
    double ActualHead { get; }

    /// <summary>インバータ機か否か</summary>
    bool HasInverter { get; }

    /// <summary>定格軸動力[kW]を取得する</summary>
    double NominalShaftPower { get; }

    /// <summary>モーター効率[-]を取得する</summary>
    double MotorEfficiency { get; }

    /// <summary>流体機械効率[-]を取得する</summary>
    double GetFluidMachineryEfficiency();

    /// <summary>インバータ効率[-]を取得する</summary>
    double GetInverterEfficiency();

    /// <summary>消費電力[kW]を取得する</summary>
    double GetElectricConsumption();

    /// <summary>総合効率[-]を取得する</summary>
    double GetTotalEfficiency();
  }

  #endregion

}
