﻿/* PumpSystem.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit
{
  /// <summary>ポンプシステム</summary>
  public class PumpSystem : ImmutablePumpSystem
  {

    #region インスタンス変数・プロパティ

    /// <summary>遠心ポンプ</summary>
    private CentrifugalPump pump;

    /// <summary>遠心ポンプを取得する</summary>
    public ImmutableCentrifugalPump Pump { get { return pump; } }

    /// <summary>抵抗係数[kPa/(m3/s)^2]</summary>
    private double resistanceCoefficient;

    /// <summary>総流量[m3/s]を設定・取得する</summary>
    public double TotalFlowRate { get; set; }

    /// <summary>バイパス量[m3/s]を取得する</summary>
    public double BypassFlowRate { get; private set; }

    /// <summary>運転台数[台]を取得する</summary>
    public int OperatingNumber { get; private set; }

    /// <summary>ポンプ台数[台]を取得する</summary>
    public int PumpNumber { get; private set; }

    /// <summary>実揚程[kPa]を取得する</summary>
    public double ActualHead { get; private set; }

    /// <summary>圧力設定値[kPa]を設定・取得する</summary>
    public double PressureSetpoint
    {
      get { return pump.PressureSetpoint; }
      set { pump.PressureSetpoint = value; }
    }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="pump">遠心ポンプ</param>
    /// <param name="designPressure">設計圧力[kPa]</param>
    /// <param name="designFlowRate">設計流量[m3/s]</param>
    /// <param name="actualHead">実揚程[kPa]</param>
    /// <param name="pumpNumber">ポンプの台数</param>
    public PumpSystem
      (CentrifugalPump pump, double designPressure, double designFlowRate, double actualHead, int pumpNumber)
    {
      this.pump = pump;
      PumpNumber = pumpNumber;
      PressureSetpoint = designPressure;
      ActualHead = actualHead;

      //抵抗係数[kPa/(m3/s)^2]を計算
      resistanceCoefficient = (designPressure - actualHead) / (designFlowRate * designFlowRate);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>状態を更新する</summary>
    public void UpdateState()
    {
      //運転台数を確定
      OperatingNumber = getOperatingNumber(TotalFlowRate);
      if (OperatingNumber == 0)
      {
        ShutOff();
        return;
      }
      //過負荷の場合
      if (PumpNumber < OperatingNumber)
      {
        OperatingNumber = PumpNumber;
        double tf;
        pump.updateWithResistanceAndRotationRatio(1.0, resistanceCoefficient, ActualHead, PumpNumber, out tf);
        TotalFlowRate = tf;
        BypassFlowRate = 0;
      }
      else
      {
        double vFlow = TotalFlowRate / OperatingNumber;
        //最小吐出圧制御の場合：
        if (Pump.Control == CentrifugalPump.ControlMethod.MinimumPressure)
        {
          double pressure = TotalFlowRate * TotalFlowRate * resistanceCoefficient + ActualHead;
          pump.updateWithFlowRateAndPressure(vFlow, pressure);
          //最小回転数比[-]未満の場合
          if (Pump.RotationRatio < Pump.MinimumRotationRatio)
          {
            double tf;
            pump.updateWithResistanceAndRotationRatio
              (Pump.MinimumRotationRatio, resistanceCoefficient, ActualHead, OperatingNumber, out tf);
          }
        }
        //吐出圧一定制御・バイパス制御の場合
        else pump.UpdateState(vFlow);

        BypassFlowRate = (Pump.VolumetricFlowRate * OperatingNumber) - TotalFlowRate;
      }
    }

    /// <summary>停止させる</summary>
    public void ShutOff()
    {
      pump.ShutOff();
      BypassFlowRate = 0;
      TotalFlowRate = 0;
      OperatingNumber = 0;
    }

    /// <summary>消費電力[kW]を取得する</summary>
    public double GetElectricConsumption()
    { return Pump.GetElectricConsumption() * OperatingNumber; }

    #endregion

    #region staticメソッド

    /// <summary>ポンプの必要運転台数[台]を計算する</summary>
    /// <param name="flowRate">必要水量[m3/s]</param>
    /// <returns>必要運転台数[台]</returns>
    private int getOperatingNumber(double flowRate)
    {
      if (flowRate <= 0) return 0;

      //最小吐出圧制御の場合：1台ずつ増やして確認
      if (Pump.Control == CentrifugalPump.ControlMethod.MinimumPressure)
      {
        double r2 = flowRate * flowRate;
        double ps = resistanceCoefficient * r2 + ActualHead;
        //締切揚程が必要揚程を上回ることの確認
        pump.updateWithFlowRateAndRotationRatio(0, 1.0);
        if (pump.Pressure < ps) throw new Exception("Pump Pressure Error");
        int opNum = 1;
        while (true)
        {
          //最大回転数比で水量が足りるか確認
          double tf;
          pump.updateWithResistanceAndRotationRatio(1.0, resistanceCoefficient, ActualHead, opNum, out tf);
          if (flowRate < tf) break;
          else opNum++;
          if (50 < opNum) throw new Exception("Pump Number Error");
        }
        return opNum;
      }
      //吐出圧一定制御・バイパス制御の場合には台数で流量均等分割
      else
      {
        pump.updateWithRotationRatioAndPressure(1.0, PressureSetpoint);
        return (int)Math.Ceiling(flowRate / Pump.VolumetricFlowRate);
      }
    }

    #endregion

  }

  /// <summary>読み取り専用ポンプシステム</summary>
  public interface ImmutablePumpSystem
  {
    /// <summary>遠心ポンプを取得する</summary>
    ImmutableCentrifugalPump Pump { get; }

    /// <summary>総流量[m3/s]を設定・取得する</summary>
    double TotalFlowRate { get; }

    /// <summary>バイパス量[m3/s]を取得する</summary>
    double BypassFlowRate { get; }

    /// <summary>運転台数[台]を取得する</summary>
    int OperatingNumber { get; }

    /// <summary>ポンプ台数[台]を取得する</summary>
    int PumpNumber { get; }

    /// <summary>実揚程[kPa]を取得する</summary>
    double ActualHead { get; }

    /// <summary>設計全圧・揚程[kPa]を取得する</summary>
    double PressureSetpoint { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double GetElectricConsumption();
  }

}
