﻿/* CrossFinEvaporator.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>プレートフィン付管形蒸発器（冷却除湿着霜コイル）</summary>
  public class CrossFinEvaporator : ImmutableCrossFinEvaporator
  {

    #region 定数宣言

    /// <summary>大気圧[kPa]は固定:海抜0m</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>氷の昇華潜熱[kJ/kg]</summary>
    private const double SUBLIMINATION_LATENT_HEAT = 2837;

    /// <summary>0℃の水の蒸発潜熱[kJ/kg]</summary>
    private const double VAPORIZATION_LATENT_HEAT = 2501;

    /// <summary>乾き空気の定圧比熱[kJ/kg-K]</summary>
    private const double DRYAIR_ISOBARIC_SPECIFIC_HEAT = 1.005;

    /// <summary>水蒸気の定圧比熱[kJ/kg-K]</summary>
    private const double VAPOR_ISOBARIC_SPECIFIC_HEAT = 1.805;

    /// <summary>氷の定圧比熱[kJ/kg-K]</summary>
    private const double ICE_ISOBARIC_SPECIFIC_HEAT = 2.090;

    /// <summary>乾きコイルの熱貫流率[kW/(m2K)]計算係数</summary>
    /// <remarks>プレートフィン2.0m/s基準:空気調和ハンドブックより</remarks>
    private const double CF_A = 0.0236;
    private const double CF_B = 0.5479;

    /// <summary>着霜による熱貫流率効率低下係数</summary>
    private const double F_PENALTY = 0.6;

    #endregion

    #region プロパティ
    
    /// <summary>乾湿境界の相対湿度[%]</summary>
    private double borderRelativeHumidity;

    /// <summary>伝熱面積[m2]を取得する</summary>
    public double SurfaceArea { get; private set; }

    /// <summary>乾き伝熱面積[m2]を取得する</summary>
    public double DrySurfaceArea { get; private set; }

    /// <summary>湿り伝熱面積[m2]を取得する</summary>
    public double WetSurfaceArea { get; private set; }

    /// <summary>着霜伝熱面積[m2]を取得する</summary>
    public double FrostSurfaceArea
    {
      get { return SurfaceArea - (DrySurfaceArea + WetSurfaceArea); }
    }

    /// <summary>定格風量[kg/s]を取得する</summary>
    public double NominalAirFlowRate { get; private set; }

    /// <summary>風量[kg/s]を取得する</summary>
    public double AirFlowRate { get; private set; }

    /// <summary>蒸発温度[C]を取得する</summary>
    public double EvaporatingTemperature { get; private set; }

    /// <summary>乾湿境界の相対湿度[%]を設定・取得する</summary>
    public double BorderRelativeHumidity
    {
      get { return borderRelativeHumidity; }
      set { borderRelativeHumidity = Math.Max(50, value); }
    }

    /// <summary>入口空気乾球温度[C]を設定・取得する</summary>
    public double InletAirTemperature { get; set; }

    /// <summary>入口空気絶対温度[kg/kg]を設定・取得する</summary>
    public double InletAirHumidityRatio { get; set; }

    /// <summary>出口空気乾球温度[C]を取得する</summary>
    public double OutletAirTemperature { get; set; }

    /// <summary>出口空気絶対温度[kg/kg]を取得する</summary>
    public double OutletAirHumidityRatio { get; set; }

    /// <summary>交換熱量[kW]を取得する</summary>
    public double HeatTransfer { get; private set; }

    /// <summary>デフロスト負荷[kW]を取得する</summary>
    public double DefrostLoad { get; private set; }

    /// <summary>停止しているか否か</summary>
    public bool IsShutOff { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="evpTemperature">蒸発温度[C]</param>
    /// <param name="heatTransfer">熱交換能力[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    public CrossFinEvaporator(double evpTemperature, double heatTransfer, double airFlowRate,
      double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity)
    {
      //プロパティ初期化
      NominalAirFlowRate = airFlowRate;
      AirFlowRate = airFlowRate;
      BorderRelativeHumidity = borderRelativeHumidity;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;
      shutOff();

      //伝熱面積を初期化する
      SurfaceArea = GetSurfaceArea(evpTemperature, heatTransfer, airFlowRate, airFlowRate,
        inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity);
    }

    /// <summary>伝熱面積[m2]を計算する</summary>
    /// <param name="evpTemperature">蒸発温度[C]</param>
    /// <param name="heatTransfer">熱交換能力[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <returns>伝熱面積[m2]</returns>
    public static double GetSurfaceArea
      (double evpTemperature, double heatTransfer, double airFlowRate, double nominalAirFlowRate, 
      double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity)
    {
      double epsilon;
      double kD = CF_A * Math.Pow(airFlowRate / nominalAirFlowRate, CF_B);

      //乾湿境界判定
      double rh = MoistAir.GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
        (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      borderRelativeHumidity = Math.Max(rh, borderRelativeHumidity);

      //湿り空気比熱の計算
      double cpmaWB = MoistAir.GetSpecificHeat(inletAirHumidityRatio);

      //乾きコイル面積の計算
      double mca = cpmaWB * airFlowRate;
      double tWB = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
        (inletAirHumidityRatio, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
      double qD = (inletAirTemperature - tWB) * mca;

      //乾きコイルで伝熱が終了する場合
      if (heatTransfer < qD)
      {
        epsilon = heatTransfer / (mca * (inletAirTemperature - evpTemperature));
        return -Math.Log(1 - epsilon) * mca / kD;
      }
      //湿りコイルまで到達する場合
      epsilon = qD / (mca * (inletAirTemperature - evpTemperature));
      double sD = -Math.Log(1 - epsilon) * mca / kD;

      double qW, sW, xFB, tFB, cpmaFB;
      //湿りコイルがある場合
      if (0 < tWB)
      {
        tFB = 0;
        //湿りコイル面積の計算
        xFB = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (0, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        cpmaFB = MoistAir.GetSpecificHeat(xFB);
        double hWB = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (tWB, inletAirHumidityRatio);
        double hEvp = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
          (evpTemperature, 100, ATMOSPHERIC_PRESSURE);
        double hFB = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
          (0, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        qW = (hWB - hFB) * airFlowRate;
        double kW = kD / (0.5 * (cpmaWB + cpmaFB));

        //湿りコイルで伝熱が終了する場合
        if (heatTransfer - qD < qW)
        {
          epsilon = (heatTransfer - qD) / (airFlowRate * (hWB - hEvp));
          return -Math.Log(1 - epsilon) * airFlowRate / kW + sD;
        }
        //着霜コイルまで到達する場合
        epsilon = qW / (airFlowRate * (hWB - hEvp));
        sW = -Math.Log(1 - epsilon) * airFlowRate / kW;
      }
      else
      {
        tFB = tWB;
        xFB = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (tWB, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        cpmaFB = MoistAir.GetSpecificHeat(xFB);
        qW = 0;
        sW = 0;
      }

      //着霜コイル面積の計算
      double kF = kD / cpmaFB * F_PENALTY;
      double hdF = getHD(tFB, borderRelativeHumidity);
      double hdEvp = getHD(evpTemperature, 100);
      epsilon = (heatTransfer - qD - qW) / (airFlowRate * (hdF - hdEvp));
      if (1 <= epsilon) throw new Exception("初期化不能エラー");
      double sF = -Math.Log(1 - epsilon) * airFlowRate / kF;

      return sF + sD + sW;
    }

    /// <summary>乾球温度[C]から固体基準の比エンタルピー[kW]を計算する</summary>
    /// <param name="temperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <returns>固体基準の比エンタルピー[kW]</returns>
    private static double getHD(double temperature, double relativeHumidity)
    {
      double x = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (temperature, relativeHumidity, ATMOSPHERIC_PRESSURE);
      return DRYAIR_ISOBARIC_SPECIFIC_HEAT * temperature 
        + (ICE_ISOBARIC_SPECIFIC_HEAT * temperature + SUBLIMINATION_LATENT_HEAT) * x;
    }

    /// <summary>機器を停止させる</summary>
    private void shutOff()
    {
      AirFlowRate = 0;
      OutletAirTemperature = InletAirTemperature;
      OutletAirHumidityRatio = InletAirHumidityRatio;
      DrySurfaceArea = SurfaceArea;
      WetSurfaceArea = 0;
      HeatTransfer = 0;
      EvaporatingTemperature = 0;
      DefrostLoad = 0;
      IsShutOff = true;
    }

    #endregion

    #region 交換熱量計算処理

    /// <summary>交換熱量[kW]を計算する</summary>
    /// <param name="evpTemperature">蒸発温度[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <returns>交換熱量[kW]</returns>
    public double GetHeatTransfer
      (double evpTemperature, double airFlowRate, double inletAirTemperature, double inletAirHumidityRatio)
    {
      //プロパティ設定
      EvaporatingTemperature = evpTemperature;
      AirFlowRate = airFlowRate;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;

      //運転判定
      if (airFlowRate <= 0 || inletAirTemperature <= evpTemperature)
      {
        shutOff();
        return 0;
      }

      double ht, to, wo, sd, sw, dfl;
      GetHeatTransfer(evpTemperature, airFlowRate, NominalAirFlowRate, SurfaceArea, inletAirTemperature, 
        inletAirHumidityRatio, borderRelativeHumidity, out ht, out to, out wo, out sd, out sw, out dfl);
      OutletAirTemperature = to;
      OutletAirHumidityRatio = wo;
      DrySurfaceArea = sd;
      WetSurfaceArea = sw;
      HeatTransfer = ht;
      DefrostLoad = dfl;
      return ht;
    }

    /// <summary>交換熱量[kW]を計算する</summary>
    /// <param name="evpTemperature">蒸発温度[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="heatTransfer">出力：交換熱量[kW]</param>
    /// <param name="outletAirTemperature">出力：出口空気乾球温度[C]</param>
    /// <param name="outletAirHumidityRatio">出力：出口空気絶対湿度[kg/kg]</param>
    /// <param name="sD">出力：乾きコイル面積[m2]</param>
    /// <param name="sW">出力：湿りコイル面積[m2]</param>
    /// <param name="defrostLoad">出力：除霜負荷[kW]</param>
    public static void GetHeatTransfer
      (double evpTemperature, double airFlowRate, double nominalAirFlowRate, double surfaceArea,
      double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity, 
      out double heatTransfer, out double outletAirTemperature, out double outletAirHumidityRatio,
      out double sD, out double sW, out double defrostLoad)
    {
      //乾きコイルの熱通過率[kW/m2K]
      double kD = CF_A * Math.Pow(airFlowRate / nominalAirFlowRate, CF_B);

      //乾湿境界判定
      double rh = MoistAir.GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
        (inletAirTemperature, inletAirHumidityRatio, ATMOSPHERIC_PRESSURE);
      borderRelativeHumidity = Math.Max(rh, borderRelativeHumidity);
      double tWB = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
        (inletAirHumidityRatio, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);

      //湿り空気比熱[kJ/kgK]の計算
      double cpmaWB = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      double mca = cpmaWB * airFlowRate;

      //乾きコイルの計算
      //露点まで冷却するために必要な面積を計算
      double qD = mca * (inletAirTemperature - tWB);
      double epsilonD = qD / (mca * (inletAirTemperature - evpTemperature));
      if (epsilonD < 1) sD = -Math.Log(1 - epsilonD) * mca / kD;
      else sD = surfaceArea;

      double hWB = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(tWB, inletAirHumidityRatio);
      double hEvp = 
        MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity(evpTemperature, 100, ATMOSPHERIC_PRESSURE);

      //乾きコイルのみで伝熱が終了する場合
      if (surfaceArea <= sD || 1 <= epsilonD || hWB < hEvp)
      {
        sD = surfaceArea;
        sW = 0;
        defrostLoad = 0;
        outletAirHumidityRatio = inletAirHumidityRatio;

        epsilonD = 1 - Math.Exp(-kD * sD / mca);
        qD = epsilonD * mca * (inletAirTemperature - evpTemperature);
        outletAirTemperature = inletAirTemperature - qD / mca;
        heatTransfer = qD;
        return;
      }

      //湿りコイルがある場合
      double tFB, qW, xFB, cpmaFB;
      if (0 < tWB)
      {
        tFB = 0;
        xFB = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (0, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        cpmaFB = MoistAir.GetSpecificHeat(xFB);

        //凝固点（0C）まで冷却するために必要な面積を計算
        double hFB = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
          (0, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);

        qW = (hWB - hFB) * airFlowRate;
        double kW = kD / (0.5 * (cpmaWB + cpmaFB));
        double epsilonW = qW / (airFlowRate * (hWB - hEvp));
        if (epsilonW < 1) sW = -Math.Log(1 - epsilonW) * airFlowRate / kW;
        else sW = surfaceArea - sD;

        //湿りコイルで伝熱が終了する場合
        if (surfaceArea <= sW + sD || 1 <= epsilonW)
        {
          sW = surfaceArea - sD;
          defrostLoad = 0;

          epsilonW = 1 - Math.Exp(-kW * sW / airFlowRate);
          qW = epsilonW * airFlowRate * (hWB - hEvp);
          double ho = hWB - qW / airFlowRate;
          outletAirHumidityRatio = MoistAir.GetHumidityRatioFromEnthalpyAndRelativeHumidity
            (ho, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
          outletAirTemperature = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndEnthalpy
            (outletAirHumidityRatio, ho);
          heatTransfer = qD + qW;
          return;
        }
      }
      else
      {
        qW = 0;
        sW = 0;
        tFB = tWB;
        xFB = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (tWB, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        cpmaFB = MoistAir.GetSpecificHeat(xFB);
      }

      //着霜コイルの計算
      double kF = kD / cpmaFB * F_PENALTY;
      double hdFB = getHD(tFB, borderRelativeHumidity);
      double hdEvp = getHD(evpTemperature, 100);
      double sF = surfaceArea - sD - sW;
      double epsilonF = 1 - Math.Exp(-kF * sF / airFlowRate);
      double qF = epsilonF * airFlowRate * (hdFB - hdEvp);
      double hdo = hdFB - qF / airFlowRate;

      //出口空気温度を収束計算
      double to = tFB;
      double err1 = Math.Abs(getHD(to, borderRelativeHumidity) - hdo);
      const double DELTA = 0.001;
      while (0.01 < err1)
      {
        double err2 = Math.Abs(getHD(to + DELTA, borderRelativeHumidity) - hdo);
        to -= DELTA * err1 / (err2 - err1);
        err1 = Math.Abs(getHD(to, borderRelativeHumidity) - hdo);
      }
      outletAirTemperature = to;
      outletAirHumidityRatio = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (outletAirTemperature, borderRelativeHumidity, ATMOSPHERIC_PRESSURE);

      //除霜負荷を計算
      defrostLoad = airFlowRate * (xFB - outletAirHumidityRatio) 
        * (SUBLIMINATION_LATENT_HEAT - ICE_ISOBARIC_SPECIFIC_HEAT * outletAirTemperature);

      //交換熱量[kW]を集計
      heatTransfer = qD + qW + qF;
    }

    #endregion

    #region 蒸発温度計算処理

    /// <summary>蒸発温度[C]を計算する</summary>
    /// <param name="heatTransfer">交換熱量[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="deductDefrostLoad">除霜負荷を差し引くか否か</param>
    /// <returns>蒸発温度[C]</returns>
    public double GetEvaporatingTemperature(double heatTransfer, double airFlowRate, 
      double inletAirTemperature, double inletAirHumidityRatio, bool deductDefrostLoad)
    {
      //プロパティ設定
      HeatTransfer = heatTransfer;
      AirFlowRate = airFlowRate;
      InletAirTemperature = inletAirTemperature;
      InletAirHumidityRatio = inletAirHumidityRatio;

      //運転判定
      if (airFlowRate <= 0 || heatTransfer <= 0)
      {
        shutOff();
        return 0;
      }

      double te, to, wo, sd, sw, dfl;
      GetEvaporatingTemperature(heatTransfer, airFlowRate, NominalAirFlowRate, SurfaceArea, 
        inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity, deductDefrostLoad, 
        out te, out to, out wo, out sd, out sw, out dfl);
      OutletAirTemperature = to;
      OutletAirHumidityRatio = wo;
      DrySurfaceArea = sd;
      WetSurfaceArea = sw;
      EvaporatingTemperature = te;
      DefrostLoad = dfl;
      return te;
    }

    /// <summary>蒸発温度[C]を計算する</summary>
    /// <param name="heatTransfer">交換熱量[kW]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="surfaceArea">伝熱面積[m2]</param>
    /// <param name="inletAirTemperature">入口空気乾球温度[C]</param>
    /// <param name="inletAirHumidityRatio">入口空気絶対湿度[kg/kg]</param>
    /// <param name="borderRelativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="deductDefrostLoad">除霜負荷を差し引くか否か</param>
    /// <param name="evaporatingTemperature">出力：蒸発温度[C]</param>
    /// <param name="outletAirTemperature">出力：出口空気乾球温度[C]</param>
    /// <param name="outletAirHumidityRatio">出力：出口空気絶対湿度[kg/kg]</param>
    /// <param name="sD">出力：乾きコイル面積[m2]</param>
    /// <param name="sW">出力：湿りコイル面積[m2]</param>
    /// <param name="defrostLoad">出力：除霜負荷[kW]</param>
    public static void GetEvaporatingTemperature
      (double heatTransfer, double airFlowRate, double nominalAirFlowRate,
      double surfaceArea, double inletAirTemperature, double inletAirHumidityRatio, double borderRelativeHumidity,
      bool deductDefrostLoad, out double evaporatingTemperature, out double outletAirTemperature, 
      out double outletAirHumidityRatio, out double sD, out double sW, out double defrostLoad)
    {
      //蒸発温度を仮定
      double cpma = MoistAir.GetSpecificHeat(inletAirHumidityRatio);
      evaporatingTemperature = inletAirTemperature - heatTransfer / (airFlowRate * cpma);

      Roots.ErrorFunction eFnc = delegate (double eTemp)
      {
        double ht, ot, oa, sd, sw, dl;
        GetHeatTransfer(eTemp, airFlowRate, nominalAirFlowRate, surfaceArea, inletAirTemperature,
          inletAirHumidityRatio, borderRelativeHumidity, out ht, out ot, out oa, out sd, out sw, out dl);
        if (deductDefrostLoad) return ht - heatTransfer - dl;
        else return ht - heatTransfer;
      };
      evaporatingTemperature = Roots.Brent(evaporatingTemperature - 10, evaporatingTemperature + 10, 0.00001, eFnc);
      double hTransfer;
      GetHeatTransfer(evaporatingTemperature, airFlowRate, nominalAirFlowRate, surfaceArea,
        inletAirTemperature, inletAirHumidityRatio, borderRelativeHumidity, 
        out hTransfer, out outletAirTemperature, out outletAirHumidityRatio, out sD, out sW, out defrostLoad);
    }

    #endregion

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用のプレートフィン付管形蒸発器</summary>
  public interface ImmutableCrossFinEvaporator
  {
    /// <summary>伝熱面積[m2]を取得する</summary>
    double SurfaceArea { get; }

    /// <summary>乾き伝熱面積[m2]を取得する</summary>
    double DrySurfaceArea { get; }

    /// <summary>湿り伝熱面積[m2]を取得する</summary>
    double WetSurfaceArea { get; }

    /// <summary>着霜伝熱面積[m2]を取得する</summary>
    double FrostSurfaceArea { get; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    double NominalAirFlowRate { get; }

    /// <summary>風量[kg/s]を取得する</summary>
    double AirFlowRate { get; }

    /// <summary>蒸発温度[C]を取得する</summary>
    double EvaporatingTemperature { get; }

    /// <summary>乾湿境界の相対湿度[%]を取得する</summary>
    double BorderRelativeHumidity { get; }

    /// <summary>入口空気乾球温度[C]を取得する</summary>
    double InletAirTemperature { get; }

    /// <summary>入口空気絶対温度[kg/kg]を取得する</summary>
    double InletAirHumidityRatio { get; }

    /// <summary>出口空気乾球温度[C]を取得する</summary>
    double OutletAirTemperature { get; }

    /// <summary>出口空気絶対温度[kg/kg]を取得する</summary>
    double OutletAirHumidityRatio { get; }

    /// <summary>交換熱量[kW]を取得する</summary>
    double HeatTransfer { get; }

    /// <summary>デフロスト負荷[kW]を取得する</summary>
    double DefrostLoad { get; }

  }
 
	#endregion
  
}