﻿/* PlateHeatExchanger.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>プレート熱交換器</summary>
  public class PlateHeatExchanger : ImmutablePlateHeatExchanger
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    /// <summary>伝熱係数[kW/K]を取得する</summary>
    public double HeatTransferCoefficient { get; private set; }

    /// <summary>熱源水流量上限値[kg/s]を取得する</summary>
    public double MaxHeatSourceFlowRate { get; private set; }

    /// <summary>供給流量上限値[kg/s]を取得する</summary>
    public double MaxSupplyFlowRate { get; private set; }

    /// <summary>熱源水流量[kg/s]を取得する</summary>
    public double HeatSourceFlowRate { get; private set; }

    /// <summary>供給流量[kg/s]を取得する</summary>
    public double SupplyFlowRate { get; private set; }

    /// <summary>熱源水入口温度[C]を取得する</summary>
    public double HeatSourceInletTemperature { get; private set; }

    /// <summary>熱源水出口温度[C]を取得する</summary>
    public double HeatSourceOutletTemperature { get; private set; }

    /// <summary>供給温度[C]を取得する</summary>
    public double SupplyTemperature { get; private set; }

    /// <summary>還温度[C]を取得する</summary>
    public double ReturnTemperature { get; private set; }

    /// <summary>供給温度設定値[C]を設定・取得する</summary>
    public double SupplyTemperatureSetpoint { get; set; }

    /// <summary>熱交換量[kW]を取得する</summary>
    public double HeatTransfer { get; private set; }

    #endregion

    #region コンストラクタ・インスタンスメソッド

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="heatTransfer">熱交換量[kW]</param>
    /// <param name="heatsourceTemperature">熱源水温度[C]</param>
    /// <param name="heatsourceFlowRate">熱源水流量[kg/s]</param>
    /// <param name="supplyTemperature">供給温度[C]</param>
    /// <param name="supplyFlowRate">供給流量[kg/s]</param>
    public PlateHeatExchanger(double heatTransfer,
      double heatsourceTemperature, double heatsourceFlowRate, double supplyTemperature, double supplyFlowRate)
    {
      MaxHeatSourceFlowRate = heatsourceFlowRate;
      MaxSupplyFlowRate = supplyFlowRate;
      HeatSourceInletTemperature = heatsourceTemperature;

      bool isHeating = supplyTemperature < heatsourceTemperature;
      if (isHeating)
      {
        ReturnTemperature = supplyTemperature - heatTransfer / (WATER_SPECIFIC_HEAT * supplyFlowRate);
        HeatTransferCoefficient = HeatExchange.GetHeatTransferCoefficient
          (heatsourceTemperature, ReturnTemperature, heatsourceFlowRate * WATER_SPECIFIC_HEAT,
          supplyFlowRate * WATER_SPECIFIC_HEAT, heatTransfer, HeatExchange.FlowType.CounterFlow);
      }
      else
      {
        ReturnTemperature = supplyTemperature + heatTransfer / (WATER_SPECIFIC_HEAT * supplyFlowRate);
        HeatTransferCoefficient = HeatExchange.GetHeatTransferCoefficient
          (ReturnTemperature, heatsourceTemperature, supplyFlowRate * WATER_SPECIFIC_HEAT,
          heatsourceFlowRate * WATER_SPECIFIC_HEAT, heatTransfer, HeatExchange.FlowType.CounterFlow);
      }
      ShutOff();
    }

    /// <summary>出口状態を更新する（成り行き計算）</summary>
    /// <param name="heatsourceTemperature">熱源水入口温度[C]</param>
    /// <param name="returnTemperature">還水温度[C]</param>
    /// <param name="heatsourceFlowRate">熱源水流量[kg/s]</param>
    /// <param name="supplyFlowRate">供給流量[kg/s]</param>
    public void Update
      (double heatsourceTemperature, double returnTemperature, double heatsourceFlowRate, double supplyFlowRate)
    {
      IsOverLoad = false;
      HeatSourceInletTemperature = heatsourceTemperature;
      ReturnTemperature = returnTemperature;
      HeatSourceFlowRate = heatsourceFlowRate;
      SupplyFlowRate = supplyFlowRate;

      double sign;
      bool isHeating = returnTemperature < heatsourceTemperature;
      double mchs = HeatSourceFlowRate * WATER_SPECIFIC_HEAT;
      double mcsp = SupplyFlowRate * WATER_SPECIFIC_HEAT;
      if (isHeating)
      {
        sign = 1;
        HeatTransfer = HeatExchange.GetHeatTransfer(heatsourceTemperature, returnTemperature,
          mchs, mcsp, HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
      }
      else
      {
        sign = -1;
        HeatTransfer = HeatExchange.GetHeatTransfer(returnTemperature, heatsourceTemperature,
          mcsp, mchs, HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
      }

      //出口水温計算
      SupplyTemperature = ReturnTemperature + sign * HeatTransfer / mcsp;
      HeatSourceOutletTemperature = HeatSourceInletTemperature - sign * HeatTransfer / mchs;
    }

    /// <summary>供給温度[C]を制御する</summary>
    /// <param name="heatsourceTemperature">熱源水入口温度[C]</param>
    /// <param name="returnTemperature">還水温度[C]</param>
    /// <param name="supplyFlowRate">供給流量[kg/s]</param>
    public void ControlSupplyTemperature
      (double heatsourceTemperature, double returnTemperature, double supplyFlowRate)
    {
      HeatSourceInletTemperature = heatsourceTemperature;
      ReturnTemperature = returnTemperature;
      SupplyFlowRate = supplyFlowRate;

      //加熱冷却逆転判定
      bool isHeating = returnTemperature < SupplyTemperatureSetpoint;
      bool isRev = (heatsourceTemperature < returnTemperature && isHeating) ||
        (returnTemperature < heatsourceTemperature && !isHeating);

      //流量0以下または加熱冷却逆転の場合には機器を停止
      if (supplyFlowRate <= 0 || isRev)
      {
        ShutOff();
        return;
      }

      double sign;
      double mcSply = supplyFlowRate * WATER_SPECIFIC_HEAT;
      double ql = Math.Abs(returnTemperature - SupplyTemperatureSetpoint) * mcSply;
      //加熱運転
      if (isHeating)
      {
        sign = 1;
        //最大流量での熱交換量計算
        HeatTransfer = HeatExchange.GetHeatTransfer
          (heatsourceTemperature, returnTemperature, MaxHeatSourceFlowRate * WATER_SPECIFIC_HEAT, mcSply,
          HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
        IsOverLoad = HeatTransfer < ql;
        //過負荷判定
        if (IsOverLoad) HeatSourceFlowRate = MaxHeatSourceFlowRate;
        else
        {
          Roots.ErrorFunction eFnc = delegate (double flow) {
            HeatTransfer = HeatExchange.GetHeatTransfer
            (heatsourceTemperature, returnTemperature, flow * WATER_SPECIFIC_HEAT, mcSply,
            HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
            return HeatTransfer - ql;
          };
          HeatSourceFlowRate = Roots.Bisection
            (eFnc, 0, MaxHeatSourceFlowRate, -ql, HeatTransfer - ql, 0, MaxHeatSourceFlowRate * 0.001, 20);
        }
      }
      //冷却運転
      else
      {
        sign = -1;
        //最大流量での熱交換量計算
        HeatTransfer = HeatExchange.GetHeatTransfer
          (returnTemperature, heatsourceTemperature, mcSply, MaxHeatSourceFlowRate * WATER_SPECIFIC_HEAT,
          HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
        IsOverLoad = HeatTransfer < ql;
        //過負荷判定
        if (IsOverLoad) HeatSourceFlowRate = MaxHeatSourceFlowRate;
        else
        {
          Roots.ErrorFunction eFnc = delegate (double flow) {
            HeatTransfer = HeatExchange.GetHeatTransfer
            (returnTemperature, heatsourceTemperature, mcSply, flow * WATER_SPECIFIC_HEAT,
            HeatTransferCoefficient, HeatExchange.FlowType.CounterFlow);
            return HeatTransfer - ql;
          };
          HeatSourceFlowRate = Roots.Bisection
            (eFnc, 0, MaxHeatSourceFlowRate, -ql, HeatTransfer - ql, 0, MaxHeatSourceFlowRate * 0.001, 20);
        }
      }

      //出口水温計算
      SupplyTemperature = ReturnTemperature + sign * HeatTransfer / mcSply;
      HeatSourceOutletTemperature = HeatSourceInletTemperature
        - sign * HeatTransfer / (HeatSourceFlowRate * WATER_SPECIFIC_HEAT);
    }

    /// <summary>停止させる</summary>
    public void ShutOff()
    {
      HeatTransfer = 0;
      SupplyTemperature = ReturnTemperature;
      HeatSourceOutletTemperature = HeatSourceInletTemperature;
      SupplyFlowRate = HeatSourceFlowRate = 0;
    }

    #endregion

  }

  /// <summary>読み取り専用プレート熱交換器</summary>
  public interface ImmutablePlateHeatExchanger
  {
    /// <summary>過負荷か否か</summary>
    bool IsOverLoad { get; }

    /// <summary>伝熱係数[kW/K]を取得する</summary>
    double HeatTransferCoefficient { get; }

    /// <summary>熱源水流量上限値[kg/s]を取得する</summary>
    double MaxHeatSourceFlowRate { get; }

    /// <summary>供給流量上限値[kg/s]を取得する</summary>
    double MaxSupplyFlowRate { get; }

    /// <summary>熱源水流量[kg/s]を取得する</summary>
    double HeatSourceFlowRate { get; }

    /// <summary>供給流量[kg/s]を取得する</summary>
    double SupplyFlowRate { get; }

    /// <summary>熱源水入口温度[C]を取得する</summary>
    double HeatSourceInletTemperature { get; }

    /// <summary>熱源水出口温度[C]を取得する</summary>
    double HeatSourceOutletTemperature { get; }

    /// <summary>供給温度[C]を取得する</summary>
    double SupplyTemperature { get; }

    /// <summary>還温度[C]を取得する</summary>
    double ReturnTemperature { get; }

    /// <summary>供給温度設定値[C]を取得する</summary>
    double SupplyTemperatureSetpoint { get; }

    /// <summary>熱交換量[kW]を取得する</summary>
    double HeatTransfer { get; }
  }

}
