﻿/* RotaryRegenerator.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>回転型熱交換器</summary>
  public class RotaryRegenerator : ImmutableRotaryRegenerator
  {

    #region 定数宣言

    /// <summary>湿り空気の密度[kg/m3]</summary>
    private const double AIR_DENSITY = 1.2;

    /// <summary>大気圧[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>SAとEAの流路の比率は1:1とする</summary>
    private const double SA_AREA_RATE = 0.5;

    /// <summary>ローター見付け面積あたりの周長[m/m2]</summary>
    private const double PERUNIT = 5.76e-3 / (0.196e-6 + 2.584e-6);

    /// <summary>マトリクスの等価直径[m]</summary>
    private const double E_DIAMETER = 1.795e-3;

    /// <summary>ローター見付け面積に対する流路面積比率[-]</summary>
    private const double AIRSTREAM_RATE = 2.584e-6 / (0.196e-6 + 2.584e-6);

    #endregion

    #region インスタンス変数

    /// <summary>マトリクスの熱伝導率[W/(mK)]</summary>
    private double thermalConductivity;

    /// <summary>ロータ前面見付面積[m2]</summary>
    private double frontalArea;

    /// <summary>流路断面積[m2]</summary>
    private double airStreamArea;

    /// <summary>マトリクス断面積[m2]</summary>
    private double matrixArea;

    /// <summary>伝熱面積[m2]</summary>
    private double heatTransferArea;

    /// <summary>マトリクスの熱容量回転量[W/K]</summary>
    private double matrixHeatCapacity;

    /// <summary>定格の熱通過率[-]</summary>
    private double nominalEpsilon;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>詳細モデルか否かを取得する</summary>
    /// <remarks>簡易モデルは風量によらず熱通過有効度一定</remarks>
    public bool IsDetailedModel { get; private set; }

    /// <summary>全熱交換器か否かを取得する</summary>
    public bool IsDesiccantWheel { get; private set; }

    /// <summary>ローターの直径[m]を取得する</summary>
    public double Diameter { get; private set; }

    /// <summary>ローターの奥行き[m]を取得する</summary>
    public double Depth { get; private set; }

    /// <summary>給気風量[CMH]を取得する</summary>
    public double SupplyAirFlowVolume { get; private set; }

    /// <summary>排気風量[CMH]を取得する</summary>
    public double ExhaustAirFlowVolume { get; private set; }

    /// <summary>給気入口乾球温度[C]を取得する</summary>
    public double SupplyAirInletDrybulbTemperature { get; private set; }

    /// <summary>排気入口乾球温度[C]を取得する</summary>
    public double ExhaustAirInletDrybulbTemperature { get; private set; }

    /// <summary>給気出口乾球温度[C]を取得する</summary>
    public double SupplyAirOutletDrybulbTemperature { get; private set; }

    /// <summary>排気出口乾球温度[C]を取得する</summary>
    public double ExhaustAirOutletDrybulbTemperature { get; private set; }

    /// <summary>給気入口絶対湿度[kg/kg]を取得する</summary>
    public double SupplyAirInletHumidityRatio { get; private set; }

    /// <summary>排気入口絶対湿度[kg/kg]を取得する</summary>
    public double ExhaustAirInletHumidityRatio { get; private set; }

    /// <summary>給気出口絶対湿度[kg/kg]を取得する</summary>
    public double SupplyAirOutletHumidityRatio { get; private set; }

    /// <summary>排気出口絶対湿度[kg/kg]を取得する</summary>
    public double ExhaustAirOutletHumidityRatio { get; private set; }

    /// <summary>給気側熱交換効率εSA[-]を取得する</summary>
    public double Efficiency { get; private set; }

    /// <summary>定格消費電力[kW]を設定・取得する</summary>
    public double NominalElectricity { get; set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double Electricity { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nominalEpsilon">定格の熱通過率[-]</param>
    /// <param name="isDesiccantWheel">デシカントローターの場合にはtrue</param>
    /// <param name="nominalElectricity">定格消費電力[kW]</param>
    /// <remarks>風量によらず熱通過有効度を一定とする簡易モデル</remarks>
    public RotaryRegenerator(double nominalEpsilon, bool isDesiccantWheel, double nominalElectricity)
    {
      IsDetailedModel = false;
      this.nominalEpsilon = nominalEpsilon;
      IsDesiccantWheel = isDesiccantWheel;
      NominalElectricity = nominalElectricity;
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="efficiency">定格の給気側熱交換効率εSA[-]</param>
    /// <param name="diameter">ローター直径[m]</param>
    /// <param name="depth">ローター奥行き[m]</param>
    /// <param name="thermalConductivity">マトリクスの熱伝導率[W/(mK)]</param>
    /// <param name="saFlowVolume">給気風量[CMH]</param>
    /// <param name="eaFlowVolume">排気風量[CMH]</param>
    /// <param name="isDesiccantWheel">デシカントローターの場合にはtrue</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="nominalElectricity">定格消費電力</param>
    public RotaryRegenerator
      (double efficiency, double diameter, double depth, double thermalConductivity, double saFlowVolume,
      double eaFlowVolume, bool isDesiccantWheel, double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, double nominalElectricity)
    {
      initialize(efficiency, diameter, depth, thermalConductivity, saFlowVolume, eaFlowVolume,
        isDesiccantWheel, inletSADrybulbTemperature, inletSAHumidityRatio, inletEADrybulbTemperature,
        inletEAHumidityRatio, nominalElectricity);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="depth">ローター奥行き[m]</param>
    /// <param name="saFlowVolume">給気風量[CMH]</param>
    /// <param name="eaFlowVolume">排気風量[CMH]</param>
    /// <param name="isDesiccantWheel">デシカントローターの場合にはtrue</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    public RotaryRegenerator
      (double depth, double saFlowVolume, double eaFlowVolume, bool isDesiccantWheel,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio)
    {
      const double VEL = 3.0;
      double diameter = 2 * Math.Sqrt(saFlowVolume / 3600 / (VEL * 0.433 * Math.PI));
      double mr = saFlowVolume / eaFlowVolume;
      double eff = 1.0884 - 0.0608 * VEL + mr * ((0.0623 + 0.0022 * VEL) + mr * (-0.3093 + mr * 0.0752));
      double nomElec = 1e-5 * saFlowVolume + 0.09;
      IsDesiccantWheel = isDesiccantWheel;

      initialize
        (eff, diameter, depth, 210, saFlowVolume, eaFlowVolume, IsDesiccantWheel, inletSADrybulbTemperature,
        inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio, nomElec);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="efficiency">定格の給気側熱交換効率εSA[-]</param>
    /// <param name="diameter">ローター直径[m]</param>
    /// <param name="depth">ローター奥行き[m]</param>
    /// <param name="thermalConductivity">マトリクスの熱伝導率[W/(mK)]</param>
    /// <param name="saFlowVolume">給気風量[CMH]</param>
    /// <param name="eaFlowVolume">排気風量[CMH]</param>
    /// <param name="isDesiccantWheel">デシカントローターの場合にはtrue</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="nominalElectricity">定格消費電力</param>
    private void initialize
      (double efficiency, double diameter, double depth, double thermalConductivity,
      double saFlowVolume, double eaFlowVolume, bool isDesiccantWheel,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, double nominalElectricity)
    {
      IsDetailedModel = true;

      //幾何学形状を保存・計算
      Depth = depth;
      Diameter = diameter;
      GetRotaryGeometrics
        (diameter, depth, out frontalArea, out airStreamArea, out matrixArea, out heatTransferArea);

      //マトリクスの熱伝導率[W/(mK)]を保存して熱容量を逆算
      this.thermalConductivity = thermalConductivity;
      matrixHeatCapacity = GetMatrixHeatCapacity
        (efficiency, frontalArea, airStreamArea, matrixArea, heatTransferArea, depth, thermalConductivity,
        saFlowVolume, eaFlowVolume, inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio);

      //ローターの種類と消費電力を保存
      IsDesiccantWheel = isDesiccantWheel;
      NominalElectricity = nominalElectricity;

      //定格条件で成り行き計算
      UpdateState(saFlowVolume, eaFlowVolume, 1.0, inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio);
    }

    #endregion

    #region 成り行き計算処理（インスタンスメソッド）

    /// <summary>成り行き状態を計算する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    public void UpdateState
      (double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio)
    {
      //風量を保存
      SupplyAirFlowVolume = supplyAirFlowVolume;
      ExhaustAirFlowVolume = exhaustAirFlowVolume;

      //入口空気状態を保存
      SupplyAirInletDrybulbTemperature = inletSADrybulbTemperature;
      SupplyAirInletHumidityRatio = inletSAHumidityRatio;
      ExhaustAirInletDrybulbTemperature = inletEADrybulbTemperature;
      ExhaustAirInletHumidityRatio = inletEAHumidityRatio;

      //消費電力は回転率に比例
      Electricity = NominalElectricity * rotatingRate;

      //風量・回転数が0の場合
      if (supplyAirFlowVolume <= 0 || exhaustAirFlowVolume <= 0 || rotatingRate <= 0)
      {
        ShutOff();
        return;
      }

      double effSA, effEA;
      if (IsDetailedModel)
      {
        getEfficiency_Detailed
          (supplyAirFlowVolume, exhaustAirFlowVolume, rotatingRate,
          inletSADrybulbTemperature, inletSAHumidityRatio,
          inletEADrybulbTemperature, inletEAHumidityRatio, out effSA, out effEA);
      }
      else
      {
        getEfficiency_Simplified
          (supplyAirFlowVolume, exhaustAirFlowVolume, rotatingRate,
          inletSADrybulbTemperature, inletSAHumidityRatio,
          inletEADrybulbTemperature, inletEAHumidityRatio, out effSA, out effEA);
      }
      Efficiency = effSA;

      //出口空気状態の計算
      SupplyAirOutletDrybulbTemperature = SupplyAirInletDrybulbTemperature -
        effSA * (SupplyAirInletDrybulbTemperature - ExhaustAirInletDrybulbTemperature);
      ExhaustAirOutletDrybulbTemperature = ExhaustAirInletDrybulbTemperature -
        effEA * (ExhaustAirInletDrybulbTemperature - SupplyAirInletDrybulbTemperature);
      //水分交換
      if (IsDesiccantWheel)
      {
        SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio -
          effSA * (SupplyAirInletHumidityRatio - ExhaustAirInletHumidityRatio);
        ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio -
          effEA * (ExhaustAirInletHumidityRatio - SupplyAirInletHumidityRatio);
      }
      else
      {
        SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio;
        ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio;
      }
    }

    /// <summary>熱交換効率[-]を計算する（詳細モデル）</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="effSA">出力:SA側熱交換効率εSA[-]</param>
    /// <param name="effEA">出力:EA側熱交換効率εSA[-]</param>
    private void getEfficiency_Detailed
      (double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, out double effSA, out double effEA)
    {
      //無次元数を計算
      double ntu0m, rmc, cLambda, mcMin;
      bool isMcMinSASide;
      GetDimensionLessVariables
        (frontalArea, airStreamArea, matrixArea, heatTransferArea, Depth,
        thermalConductivity, supplyAirFlowVolume, exhaustAirFlowVolume,
        inletSADrybulbTemperature, inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio,
        out mcMin, out rmc, out ntu0m, out cLambda, out isMcMinSASide);

      //回転数制御の効果を計算
      double rr = (matrixHeatCapacity * rotatingRate) / mcMin;

      //熱通過有効度[-]を計算
      double eff = GetEffectiveness
        (mcMin, rmc, ntu0m, cLambda, rr, supplyAirFlowVolume, exhaustAirFlowVolume,
        inletSADrybulbTemperature, inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio);

      //熱交換効率[-]を計算
      if (isMcMinSASide)
      {
        effSA = eff;
        effEA = eff * rmc;
      }
      else
      {
        effSA = eff * rmc;
        effEA = eff;
      }
    }

    /// <summary>熱交換効率[-]を計算する（簡易モデル）</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="effSA">出力:SA側熱交換効率εSA[-]</param>
    /// <param name="effEA">出力:EA側熱交換効率εSA[-]</param>
    private void getEfficiency_Simplified
      (double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, out double effSA, out double effEA)
    {
      //熱容量流量比[-]の計算
      double svSA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletSADrybulbTemperature, inletSAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double svEA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletEADrybulbTemperature, inletEAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double mSA = supplyAirFlowVolume / 3600d / svSA;
      double mEA = exhaustAirFlowVolume / 3600d / svEA;
      double mcSA = mSA * MoistAir.GetSpecificHeat(inletSAHumidityRatio) * 1000;
      double mcEA = mEA * MoistAir.GetSpecificHeat(inletEAHumidityRatio) * 1000;
      double mcMin = Math.Min(mcSA, mcEA);
      double mcMax = Math.Max(mcSA, mcEA);
      double rmc = mcMin / mcMax;

      if (mcSA == mcMin)
      {
        effSA = nominalEpsilon * rotatingRate;
        effEA = nominalEpsilon * rotatingRate * rmc;
      }
      else
      {
        effSA = nominalEpsilon * rotatingRate * rmc;
        effEA = nominalEpsilon * rotatingRate;
      }
    }

    #endregion

    #region 出口状態制御処理（インスタンスメソッド）

    /// <summary>給気出口温度を制御する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="outletSADrybulbTemperatureSP">給気出口乾球温度設定値[C]</param>
    /// <returns>制御可能か否か</returns>
    public bool ControlOutletTemperature(
      double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, double outletSADrybulbTemperatureSP)
    {
      //成り行き計算実行
      UpdateState
        (supplyAirFlowVolume, exhaustAirFlowVolume, rotatingRate,
        inletSADrybulbTemperature, inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio);

      //冷却運転か否か
      bool cl = SupplyAirOutletDrybulbTemperature < SupplyAirInletDrybulbTemperature;

      //熱交換が無駄な場合は停止
      if ((cl && (SupplyAirInletDrybulbTemperature < outletSADrybulbTemperatureSP)) ||
        (!cl && (outletSADrybulbTemperatureSP < SupplyAirInletDrybulbTemperature)))
      {
        ShutOff();
        return false;
      }

      //処理可能な場合には消費電力と出口条件を修正
      bool canSolve =
        (cl && SupplyAirOutletDrybulbTemperature <= outletSADrybulbTemperatureSP
        && outletSADrybulbTemperatureSP <= SupplyAirInletDrybulbTemperature) ||
        (!cl && outletSADrybulbTemperatureSP < SupplyAirOutletDrybulbTemperature
        && SupplyAirInletDrybulbTemperature <= outletSADrybulbTemperatureSP);
      if (canSolve)
      {
        double rRun = (1d / Efficiency) * (outletSADrybulbTemperatureSP - SupplyAirInletDrybulbTemperature)
          / (ExhaustAirInletDrybulbTemperature - SupplyAirInletDrybulbTemperature);
        Electricity *= rRun;
        SupplyAirOutletDrybulbTemperature = outletSADrybulbTemperatureSP;
        SupplyAirOutletHumidityRatio =
          rRun * SupplyAirOutletHumidityRatio + (1 - rRun) * SupplyAirInletHumidityRatio;
      }
      return canSolve;
    }

    /// <summary>給気出口湿度を制御する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="outletSAHumidityRatioSP">給気出口絶対湿度設定値[kg/kg]</param>
    /// <returns>制御可能か否か</returns>
    public bool ControlOutletHumidity(
      double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, double outletSAHumidityRatioSP)
    {
      //成り行き計算実行
      UpdateState
        (supplyAirFlowVolume, exhaustAirFlowVolume, rotatingRate,
        inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio);

      //除湿運転か否か
      bool cl = SupplyAirOutletHumidityRatio < SupplyAirInletHumidityRatio;

      //熱交換が無駄な場合は停止
      if ((cl && (SupplyAirInletHumidityRatio < outletSAHumidityRatioSP)) ||
        (!cl && (outletSAHumidityRatioSP < SupplyAirInletHumidityRatio)))
      {
        ShutOff();
        return false;
      }

      //制御可能な場合には消費電力と出口条件を修正
      bool canSolve =
        (cl && SupplyAirOutletHumidityRatio <= outletSAHumidityRatioSP
        && outletSAHumidityRatioSP <= SupplyAirInletHumidityRatio) ||
        (!cl && outletSAHumidityRatioSP <= SupplyAirOutletHumidityRatio
        && SupplyAirInletHumidityRatio <= outletSAHumidityRatioSP);
      if (canSolve)
      {
        double rRun = (1d / Efficiency) *
          (outletSAHumidityRatioSP - SupplyAirInletHumidityRatio) /
          (ExhaustAirInletHumidityRatio - SupplyAirInletHumidityRatio);
        Electricity *= rRun;
        SupplyAirOutletDrybulbTemperature = rRun * SupplyAirOutletDrybulbTemperature
          + (1 - rRun) * SupplyAirInletDrybulbTemperature;
        SupplyAirOutletHumidityRatio = outletSAHumidityRatioSP;
      }
      return canSolve;
    }

    /// <summary>給気出口比エンタルピーを制御する</summary>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="rotatingRate">ローター回転率[-]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="outletSAEnthalpySP">給気出口比エンタルピー設定値[kJ/kg]</param>
    /// <returns>制御可能か否か</returns>
    public bool ControlOutletEnthalpy(
      double supplyAirFlowVolume, double exhaustAirFlowVolume, double rotatingRate,
      double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio, double outletSAEnthalpySP)
    {
      //成り行き計算実行
      UpdateState
        (supplyAirFlowVolume, exhaustAirFlowVolume, rotatingRate,
        inletSADrybulbTemperature, inletSAHumidityRatio, inletEADrybulbTemperature, inletEAHumidityRatio);

      //冷却除湿運転か否か
      double hSAi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirInletDrybulbTemperature, SupplyAirInletHumidityRatio);
      double hSAo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirOutletDrybulbTemperature, SupplyAirOutletHumidityRatio);
      bool cl = hSAo < hSAi;

      //熱交換が無駄な場合は停止
      if ((cl && (hSAi < outletSAEnthalpySP)) || (!cl && (outletSAEnthalpySP < hSAi)))
      {
        ShutOff();
        return false;
      }

      //制御可能な場合には消費電力と出口条件を修正
      bool canSolve =
        (cl && hSAo <= outletSAEnthalpySP && outletSAEnthalpySP <= hSAi) ||
        (!cl && outletSAEnthalpySP <= hSAo && hSAi <= outletSAEnthalpySP);
      if (canSolve)
      {
        double ctw = 1.805
          * (ExhaustAirInletDrybulbTemperature - SupplyAirInletDrybulbTemperature)
          * (ExhaustAirInletHumidityRatio - SupplyAirInletHumidityRatio);
        double aRun = Efficiency * Efficiency * ctw;
        double hEAi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (ExhaustAirInletDrybulbTemperature, ExhaustAirInletHumidityRatio);
        double bRun = Efficiency * ((hEAi - hSAi) - ctw);
        double cRun = hSAi - outletSAEnthalpySP;
        double bf = Math.Sqrt(bRun * bRun - 4 * aRun * cRun);
        double rRun = (-bRun + bf) / (2 * aRun);
        if (rRun < 0 || 1 < rRun) rRun = (-bRun - bf) / (2 * aRun);
        Electricity *= rRun;
        SupplyAirOutletDrybulbTemperature = rRun * SupplyAirOutletDrybulbTemperature
          + (1 - rRun) * SupplyAirInletDrybulbTemperature;
        SupplyAirOutletHumidityRatio = rRun * SupplyAirOutletHumidityRatio
          + (1 - rRun) * SupplyAirInletHumidityRatio;
      }
      return canSolve;
    }

    #endregion

    #region その他インスタンスメソッド

    /// <summary>熱回収量[kW]を計算する</summary>
    /// <returns>熱回収量[kW]（給気への加熱が正）</returns>
    public double GetHeatRecovery()
    {
      double hSAo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirOutletDrybulbTemperature, SupplyAirOutletHumidityRatio);
      double hSAi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (SupplyAirInletDrybulbTemperature, SupplyAirInletHumidityRatio);
      return SupplyAirFlowVolume * AIR_DENSITY / 3600 * (hSAo - hSAi);
    }

    /// <summary>停止させる</summary>
    public void ShutOff()
    {
      Efficiency = 0;
      SupplyAirOutletDrybulbTemperature = SupplyAirInletDrybulbTemperature;
      ExhaustAirOutletDrybulbTemperature = ExhaustAirInletDrybulbTemperature;
      SupplyAirOutletHumidityRatio = SupplyAirInletHumidityRatio;
      ExhaustAirOutletHumidityRatio = ExhaustAirInletHumidityRatio;
      Electricity = 0;
    }

    #endregion

    #region staticメソッド

    /// <summary>マトリクスの熱容量回転量[W/K]を計算する</summary>
    /// <param name="efficiency">熱交換効率εSA[-]</param>
    /// <param name="frontalArea">ロータ前面見付面積[m2]</param>
    /// <param name="airStreamArea">流路断面積[m2]</param>
    /// <param name="matrixArea">マトリクス断面積[m2]</param>
    /// <param name="heatTransferArea">伝熱面積[m2]</param>
    /// <param name="length">ローター奥行き[m]</param>
    /// <param name="thermalConductivity">マトリクスの熱伝導率[W/(mK)]</param>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <returns>マトリクスの熱容量回転量[W/K]</returns>
    public static double GetMatrixHeatCapacity
      (double efficiency, double frontalArea, double airStreamArea, double matrixArea,
      double heatTransferArea, double length, double thermalConductivity, double supplyAirFlowVolume,
      double exhaustAirFlowVolume, double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio)
    {
      //無次元数を取得
      double ntu0m, rmc, cLambda, mcMin;
      bool isMcMinSASide;
      GetDimensionLessVariables
        (frontalArea, airStreamArea, matrixArea, heatTransferArea, length, thermalConductivity,
        supplyAirFlowVolume, exhaustAirFlowVolume, inletSADrybulbTemperature, inletSAHumidityRatio,
        inletEADrybulbTemperature, inletEAHumidityRatio,
        out mcMin, out rmc, out ntu0m, out cLambda, out isMcMinSASide);

      double effectiveness;
      if (isMcMinSASide) effectiveness = efficiency;
      else effectiveness = efficiency / rmc;

      double rr;
      if (0.99 < rmc)
      {
        //熱容量流量がほぼ等しい場合
        double ee = ntu0m / (1 + ntu0m);
        rr = effectiveness / ee / (1 - cLambda);
        if (1 <= rr) rr = 100;
        else rr = Math.Pow(1 / (9 * (1 - rr)), 1 / 1.93);
      }
      else
      {
        //熱容量流量が異なる場合
        double em = 2 * rmc * Math.Log((1 - effectiveness) / (1 - effectiveness * rmc));
        em = em / (em + rmc * rmc - 1);
        double rm = em * ((1 + ntu0m) / ntu0m) / (1 - cLambda / (2 - rmc));
        if (1 <= rm) rm = 100;
        else rm = Math.Pow(1 / (9 * (1 - rm)), 1 / 1.93);
        rr = rm * (1 + rmc) / (2 * rmc);
      }
      return rr * mcMin;
    }

    /// <summary>熱通過有効度[-]を計算する</summary>
    /// <param name="mcMin">熱容量流量[W/m]</param>
    /// <param name="rmc">熱容量流量比[-]</param>
    /// <param name="ntu0m">修正移動単位数[-]</param>
    /// <param name="cLambda">流路方向の熱伝導補正係数[-]</param>
    /// <param name="rr">熱容量回転量比[-]</param>
    /// <param name="supplyAirFlowVolume">給気風量[CMH]</param>
    /// <param name="exhaustAirFlowVolume">排気風量[CMH]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <returns>熱通過有効度[-]</returns>
    public static double GetEffectiveness
      (double mcMin, double rmc, double ntu0m, double cLambda, double rr, double supplyAirFlowVolume,
      double exhaustAirFlowVolume, double inletSADrybulbTemperature, double inletSAHumidityRatio,
      double inletEADrybulbTemperature, double inletEAHumidityRatio)
    {
      double e;
      if (0.99 < rmc)
      {
        //熱容量流量がほぼ等しい場合
        double ee = ntu0m / (1 + ntu0m);
        e = ee * (1 - 1 / (9 * Math.Pow(rr, 1.93))) * (1 - cLambda);
      }
      else
      {
        //熱容量流量が異なる場合
        double rrm = 2 * rr * rmc / (1 + rmc);
        double em = ntu0m / (1 + ntu0m) * (1 - 1 / (9 * Math.Pow(rrm, 1.93)));
        em *= 1 - cLambda / (2 - rmc);
        e = Math.Exp(em * (rmc * rmc - 1) / (2 * rmc * (1 - em)));
        e = (1 - e) / (1 - rmc * e);
      }
      return Math.Min(1, Math.Max(e, 0));
    }

    /// <summary>無次元数を計算する</summary>
    /// <param name="frontalArea">ロータ前面見付面積[m2]</param>
    /// <param name="airStreamArea">流路断面積[m2]</param>
    /// <param name="matrixArea">マトリクス断面積[m2]</param>
    /// <param name="heatTransferArea">伝熱面積[m2]</param>
    /// <param name="length">ローター奥行き[m]</param>
    /// <param name="thermalConductivity">マトリクスの熱伝導率[W/(mK)]</param>
    /// <param name="saFlowVolume">給気風量[CMH]</param>
    /// <param name="eaFlowVolume">排気風量[CMH]</param>
    /// <param name="inletSADrybulbTemperature">給気入口乾球温度[C]</param>
    /// <param name="inletSAHumidityRatio">給気入口絶対湿度[kg/kg]</param>
    /// <param name="inletEADrybulbTemperature">排気入口乾球温度[C]</param>
    /// <param name="inletEAHumidityRatio">排気入口絶対湿度[kg/kg]</param>
    /// <param name="mcMin">熱容量流量[W/m]</param>
    /// <param name="rmc">熱容量流量比[-]</param>
    /// <param name="ntu0m">修正移動単位数[-]</param>
    /// <param name="cLambda">流路方向の熱伝導補正係数[-]</param>
    /// <param name="isMcMinSASide">給気側の熱容量流量が小さい場合にtrue</param>
    public static void GetDimensionLessVariables
      (double frontalArea, double airStreamArea, double matrixArea, double heatTransferArea, double length,
      double thermalConductivity, double saFlowVolume, double eaFlowVolume, double inletSADrybulbTemperature,
      double inletSAHumidityRatio, double inletEADrybulbTemperature, double inletEAHumidityRatio,
      out double mcMin, out double rmc, out double ntu0m, out double cLambda, out bool isMcMinSASide)
    {
      //熱容量流量比[-]の計算
      double svSA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletSADrybulbTemperature, inletSAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double svEA = MoistAir.GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (inletEADrybulbTemperature, inletEAHumidityRatio, ATMOSPHERIC_PRESSURE);
      double mSA = saFlowVolume / 3600d / svSA;
      double mEA = eaFlowVolume / 3600d / svEA;
      double mcSA = mSA * MoistAir.GetSpecificHeat(inletSAHumidityRatio) * 1000;
      double mcEA = mEA * MoistAir.GetSpecificHeat(inletEAHumidityRatio) * 1000;
      mcMin = Math.Min(mcSA, mcEA);
      double mcMax = Math.Max(mcSA, mcEA);
      rmc = mcMin / mcMax;
      isMcMinSASide = (mcSA == mcMin);

      //風速[m/s]の計算
      double vSA = saFlowVolume / 3600d / (airStreamArea * SA_AREA_RATE);
      double vEA = eaFlowVolume / 3600d / (airStreamArea * (1 - SA_AREA_RATE));
      //対流熱伝達率[W/(m2K)]の計算
      double ed25 = 1d / Math.Pow(E_DIAMETER, 0.25);
      double alphaSA = (4.13 + 0.195 * inletSADrybulbTemperature / 100) * Math.Pow(vSA, 0.75) * ed25;
      double alphaEA = (4.13 + 0.195 * inletEADrybulbTemperature / 100) * Math.Pow(vEA, 0.75) * ed25;
      //修正移動単位数[-]の計算
      double ha = 1 / (alphaSA * heatTransferArea * SA_AREA_RATE)
        + 1 / (alphaEA * heatTransferArea * (1 - SA_AREA_RATE));
      double ntu0 = 1 / mcMin / ha;
      ntu0m = (2 * ntu0 * rmc) / (1 + rmc);

      //流路方向の熱伝導補正係数[-]の計算
      double lambda = (thermalConductivity * matrixArea) / (length * mcMin);
      double phi = Math.Sqrt((lambda * ntu0m) / (1 + lambda * ntu0m));
      phi = phi * Math.Tanh(ntu0m / phi);
      cLambda = 1 / (1 + ntu0m * (1 + lambda * phi) / (1 + lambda * ntu0m)) - 1 / (1 + ntu0m);
    }

    /// <summary>ローターの幾何学形状を計算する</summary>
    /// <param name="diameter">ローター直径[m]</param>
    /// <param name="length">ローター奥行き[m]</param>
    /// <param name="frontalArea">ロータ前面見付面積[m2]</param>
    /// <param name="airStreamArea">流路断面積[m2]</param>
    /// <param name="matrixArea">マトリクス断面積[m2]</param>
    /// <param name="heatTransferArea">伝熱面積[m2]</param>
    public static void GetRotaryGeometrics
      (double diameter, double length, out double frontalArea, out double airStreamArea,
      out double matrixArea, out double heatTransferArea)
    {
      //見付面積[m2]の計算
      frontalArea = diameter * diameter * Math.PI / 4;
      //流路面積[m2]の計算
      airStreamArea = frontalArea * AIRSTREAM_RATE;
      //マトリクス面積[m2]の計算
      matrixArea = frontalArea - airStreamArea;
      //伝熱面積[m2]の計算
      heatTransferArea = frontalArea * PERUNIT * length;
    }

    #endregion

  }

  #region Immutableインターフェース

  /// <summary>読み取り専用RotaryRegenerator</summary>
  public interface ImmutableRotaryRegenerator
  {

    /// <summary>詳細モデルか否かを取得する</summary>
    bool IsDetailedModel { get; }

    /// <summary>全熱交換器か否かを取得する</summary>
    bool IsDesiccantWheel { get; }

    /// <summary>ローターの直径[m]を取得する</summary>
    double Diameter { get; }

    /// <summary>ローターの奥行き[m]を取得する</summary>
    double Depth { get; }

    /// <summary>給気風量[CMH]を取得する</summary>
    double SupplyAirFlowVolume { get; }

    /// <summary>排気風量[CMH]を取得する</summary>
    double ExhaustAirFlowVolume { get; }

    /// <summary>給気入口乾球温度[C]を取得する</summary>
    double SupplyAirInletDrybulbTemperature { get; }

    /// <summary>排気入口乾球温度[C]を取得する</summary>
    double ExhaustAirInletDrybulbTemperature { get; }

    /// <summary>給気出口乾球温度[C]を取得する</summary>
    double SupplyAirOutletDrybulbTemperature { get; }

    /// <summary>排気出口乾球温度[C]を取得する</summary>
    double ExhaustAirOutletDrybulbTemperature { get; }

    /// <summary>給気入口絶対湿度[kg/kg]を取得する</summary>
    double SupplyAirInletHumidityRatio { get; }

    /// <summary>排気入口絶対湿度[kg/kg]を取得する</summary>
    double ExhaustAirInletHumidityRatio { get; }

    /// <summary>給気出口絶対湿度[kg/kg]を取得する</summary>
    double SupplyAirOutletHumidityRatio { get; }

    /// <summary>排気出口絶対湿度[kg/kg]を取得する</summary>
    double ExhaustAirOutletHumidityRatio { get; }

    /// <summary>給気側熱交換効率εSA[-]を取得する</summary>
    double Efficiency { get; }

    /// <summary>定格消費電力[kW]を取得する</summary>
    double NominalElectricity { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double Electricity { get; }

    /// <summary>熱回収量[kW]を計算する</summary>
    /// <returns>熱回収量[kW]（給気への加熱が正）</returns>
    double GetHeatRecovery();
  }

  #endregion

}
