﻿/* DirectFiredDoubleEffectAbsorptionChiller.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>直だき二重効用吸収冷温水発生機</summary>
  public class DirectFiredAbsorptionChiller: ImmutableDirectFiredAbsorptionChiller
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>冷温水変流量下限[-]</summary>
    private const double MINIMUM_WATERFLOW_RATIO = 0.5;

    /// <summary>水溶液変流量下限[-]</summary>
    private const double MINIMUM_SOLUTIONFLOW_RATIO = 0.5;

    /// <summary>負荷率下限値[-]</summary>
    private const double MINIMUM_PARTIALLOAD = 0.2;

    /// <summary>周囲の温度・燃焼空気温度[C]</summary>
    private const double AMB_TEMP = 25;

    /// <summary>燃焼空気比[-]</summary>
    private const double AIR_RATIO = 1.15;

    /// <summary>排ガス温度[C]</summary>
    private const double SMK_TEMP = 100;

    #endregion

    #region インスタンス変数

    /// <summary>蒸発器伝熱係数[kW/K]</summary>
    private double evaporatorKA;

    /// <summary>凝縮器伝熱係数[kW/K]</summary>
    private double condensorKA;

    /// <summary>低温再生器伝熱係数[kW/K]</summary>
    private double lowDesorborKA;

    /// <summary>低温溶液熱交換器伝熱係数[kW/K]</summary>
    private double thinSolutionHexKA;

    /// <summary>定格水溶液流量[kg/s]</summary>
    private double solutionFlowRate;

    /// <summary>定格高温再生器投入熱量[kW]</summary>
    private double desorbHeat;

    /// <summary>缶体からの熱損失係数[kW/K]</summary>
    private double heatLossCoefficient;

    /// <summary>温水ボイラ</summary>
    private HotWaterBoiler hBoiler;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>冷却運転モードか否かを設定・取得する</summary>
    public bool IsCoolingMode { get; set; }

    /// <summary>燃料の種別を取得する</summary>
    public Boiler.Fuel Fuel { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double OutletWaterTemperature { get; private set; }

    /// <summary>出口水温設定値[C]を設定・取得する</summary>
    public double OutletWaterSetPointTemperature { get; set; }

    /// <summary>入口水温[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>冷水/温水の水量[kg/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate { get; private set; }

    /// <summary>冷水下限流量[kg/s]を取得する</summary>
    public double MinChilledWaterFlowRate { get; private set; }

    /// <summary>温水下限流量[kg/s]を取得する</summary>
    public double MinHotWaterFlowRate { get; private set; }

    /// <summary>冷却水上限流量[kg/s]を取得する</summary>
    public double MaxCoolingWaterFlowRate { get; private set; }

    /// <summary>冷却水下限流量[kg/s]を取得する</summary>
    public double MinCoolingWaterFlowRate { get; private set; }

    /// <summary>定格冷却能力[kW]を取得する</summary>
    public double NominalCoolingCapacity { get; private set; }

    /// <summary>定格加熱能力[kW]を取得する</summary>
    public double NominalHeatingCapacity { get; private set; }

    /// <summary>定格冷却燃料消費量[Nm3/s, kg/s]を取得する</summary>
    public double NominalCoolingFuelConsumption { get; private set; }

    /// <summary>定格加熱燃料消費量[Nm3/s, kg/s]を取得する</summary>
    public double NominalHeatingFuelConsumption { get; private set; }

    /// <summary>容量制御下限値[-]を取得する</summary>
    public double MinimumPartialLoadRatio { get; private set; }

    /// <summary>定格消費電力[kW]を取得する</summary>
    public double NominalElectricConsumption { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>燃料消費量[Nm3/s, kg/s]を取得する</summary>
    public double FuelConsumption { get; private set; }

    /// <summary>加熱負荷[kW]を取得する</summary>
    public double HeatingLoad { get; private set; }

    /// <summary>冷却負荷[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (FuelConsumption == 0) return 0;
        else if(IsCoolingMode)
          return 0.001 * CoolingLoad / (FuelConsumption * Boiler.GetHeatingValue(Fuel, true));
        else return 0.001 * HeatingLoad / (FuelConsumption * Boiler.GetHeatingValue(Fuel, true));
      }
    }

    /// <summary>溶液ポンプのインバータ制御があるか</summary>
    public bool HasSolutionInverterPump { get; set; }

    /// <summary>高濃度溶液の質量分率[-]を取得する</summary>
    public double ThickSolutionMassFraction { get; private set; }

    /// <summary>低濃度溶液の質量分率[-]を取得する</summary>
    public double ThinSolutionMassFraction { get; private set; }

    /// <summary>蒸発温度[C]を取得する</summary>
    public double EvaporatingTemperature { get; private set; }

    /// <summary>凝縮温度[C]を取得する</summary>
    public double CondensingTemperature { get; private set; }

    /// <summary>再生温度[C]を取得する</summary>
    public double DesorbTemperature { get; private set; }

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nominalCoolingFuelConsumption">定格冷却ガス消費[Nm3/s, kg/s]</param>
    /// <param name="nominalHetingFuelConsumption">定格加熱ガス消費[Nm3/s, kg/s]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingWaterOutletTemperature">冷却水出口温度[C]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="hotWaterOutletTemperature">温水出口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    /// <param name="electricConsumption">消費電力[kW]</param>
    /// <param name="fuel">燃料の種別</param>
    public DirectFiredAbsorptionChiller
      (double nominalCoolingFuelConsumption, double nominalHetingFuelConsumption,
      double chilledWaterInletTemperature, double chilledWaterOutletTemperature,
      double coolingWaterInletTemperature, double coolingWaterOutletTemperature,
      double hotWaterInletTemperature, double hotWaterOutletTemperature, double chilledWaterFlowRate,
      double coolingWaterFlowRate, double hotWaterFlowRate, double electricConsumption, Boiler.Fuel fuel)
    {
      this.IsCoolingMode = true;
      this.OutletWaterSetPointTemperature = chilledWaterOutletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.InletWaterTemperature = chilledWaterInletTemperature;
      this.CoolingWaterOutletTemperature = coolingWaterOutletTemperature;
      this.CoolingWaterFlowRate = coolingWaterFlowRate;
      this.WaterFlowRate = chilledWaterFlowRate;
      this.MinCoolingWaterFlowRate = coolingWaterFlowRate * MINIMUM_WATERFLOW_RATIO;
      this.MaxCoolingWaterFlowRate = coolingWaterFlowRate;
      this.MaxChilledWaterFlowRate = chilledWaterFlowRate;
      this.MaxHotWaterFlowRate = hotWaterFlowRate;
      this.MinChilledWaterFlowRate = chilledWaterFlowRate * MINIMUM_WATERFLOW_RATIO;
      this.MinHotWaterFlowRate = hotWaterFlowRate * MINIMUM_WATERFLOW_RATIO;
      this.NominalCoolingCapacity = chilledWaterFlowRate * WATER_SPECIFIC_HEAT
        * (chilledWaterInletTemperature - chilledWaterOutletTemperature);
      this.NominalHeatingCapacity = hotWaterFlowRate * WATER_SPECIFIC_HEAT
        * (hotWaterOutletTemperature - hotWaterInletTemperature);
      this.NominalCoolingFuelConsumption = nominalCoolingFuelConsumption;
      this.NominalHeatingFuelConsumption = nominalHetingFuelConsumption;
      this.Fuel = fuel;
      this.NominalElectricConsumption = electricConsumption;

      //吸収冷凍サイクルの各種性能を計算
      AbsorptionRefrigerationCycle.GetHeatTransferCoefficients(chilledWaterInletTemperature,
        chilledWaterOutletTemperature, chilledWaterFlowRate, coolingWaterInletTemperature,
        coolingWaterOutletTemperature, coolingWaterFlowRate, out evaporatorKA, out condensorKA,
        out lowDesorborKA, out thinSolutionHexKA, out solutionFlowRate, out desorbHeat);
      desorbHeat *= 1.001;  //定格性能を担保するための処理

      //缶体熱損失係数の計算
      heatLossCoefficient = Boiler.GetHeatLossCoefficient
        (desorbHeat, AbsorptionRefrigerationCycle.NOM_DSB_LIQ_TEMP, AMB_TEMP, Fuel,
        SMK_TEMP, AIR_RATIO, NominalCoolingFuelConsumption, AMB_TEMP);

      //温水ボイラ初期化
      hBoiler = new HotWaterBoiler(hotWaterInletTemperature, hotWaterOutletTemperature, hotWaterFlowRate,
        NominalHeatingFuelConsumption, 0, AMB_TEMP, AIR_RATIO, Fuel, SMK_TEMP);

      //運転停止
      ShutOff();
    }

    #endregion

    #region publicメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="inletWaterTemperature">冷温水入口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="waterFlowRate">冷温水質量流量[kg/s]</param>
    public void Update(double coolingWaterInletTemperature, double inletWaterTemperature,
      double coolingWaterFlowRate, double waterFlowRate)
    {
      //状態値を保存
      this.InletWaterTemperature = inletWaterTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.CoolingWaterFlowRate = Math.Max(coolingWaterFlowRate, MinCoolingWaterFlowRate);

      //冷却運転
      if (IsCoolingMode && (OutletWaterSetPointTemperature < InletWaterTemperature))
      {
        //極少負荷対応のための入口水温・水量補正
        this.WaterFlowRate = Math.Max(waterFlowRate, MinChilledWaterFlowRate);
        double pl = (InletWaterTemperature - OutletWaterSetPointTemperature)
          * WATER_SPECIFIC_HEAT * WaterFlowRate / NominalCoolingCapacity;
        double ti = (InletWaterTemperature - OutletWaterSetPointTemperature)
          / Math.Min(1, pl / MINIMUM_PARTIALLOAD) + OutletWaterSetPointTemperature;

        double dsbH = 0;
        double tcdo, tdsb, tevp, tcnd, wtn, wtk;
        //定格の高温再生器投入熱量で出口状態を計算
        double cht = AbsorptionRefrigerationCycle.GetChilledWaterOutletTemperature
          (ti, WaterFlowRate, CoolingWaterInletTemperature, CoolingWaterFlowRate, evaporatorKA,
          condensorKA, lowDesorborKA, thinSolutionHexKA, solutionFlowRate, desorbHeat,
          out tcdo, out tdsb, out tevp, out tcnd, out wtn, out wtk);

        IsOverLoad = OutletWaterSetPointTemperature <= cht;
        //過負荷の場合
        if (IsOverLoad)
        {
          OutletWaterTemperature = cht;
          dsbH = desorbHeat;
        }
        //処理可能の場合
        else
        {
          if (HasSolutionInverterPump)
          {
            Minimization.MinimizeFunction mFnc = delegate (double sFlow)
            {
              dsbH = AbsorptionRefrigerationCycle.GetDesorbHeat
              (ti, WaterFlowRate, CoolingWaterInletTemperature, CoolingWaterFlowRate, evaporatorKA, condensorKA,
              lowDesorborKA, thinSolutionHexKA, sFlow, OutletWaterSetPointTemperature,
              out tcdo, out tdsb, out tevp, out tcnd, out wtn, out wtk);
              return dsbH;
            };
            double sf = solutionFlowRate * MINIMUM_SOLUTIONFLOW_RATIO;
            Minimization.GoldenSection(ref sf, solutionFlowRate, mFnc);
          }
          else
          {
            dsbH = AbsorptionRefrigerationCycle.GetDesorbHeat
              (ti, WaterFlowRate, CoolingWaterInletTemperature, CoolingWaterFlowRate, evaporatorKA, condensorKA,
              lowDesorborKA, thinSolutionHexKA, solutionFlowRate, OutletWaterSetPointTemperature,
              out tcdo, out tdsb, out tevp, out tcnd, out wtn, out wtk);
          }
          OutletWaterTemperature = OutletWaterSetPointTemperature;
        }
        CoolingWaterOutletTemperature = tcdo;
        CoolingLoad = (InletWaterTemperature - OutletWaterTemperature) * WATER_SPECIFIC_HEAT * waterFlowRate;
        HeatingLoad = 0;
        FuelConsumption = Boiler.GetFuelConsumption
          (dsbH, tdsb, AMB_TEMP, Fuel, SMK_TEMP, AIR_RATIO, heatLossCoefficient,
          AMB_TEMP, AbsorptionRefrigerationCycle.NOM_DSB_LIQ_TEMP);
        ElectricConsumption = CoolingLoad / NominalCoolingCapacity * NominalElectricConsumption;
        DesorbTemperature = tdsb;
        EvaporatingTemperature = tevp;
        CondensingTemperature = tcnd;
        ThickSolutionMassFraction = wtk;
        ThinSolutionMassFraction = wtn;
      }
      //加熱運転
      else if (!IsCoolingMode && (InletWaterTemperature < OutletWaterSetPointTemperature))
      {
        //水量設定
        CoolingWaterFlowRate = 0;
        this.WaterFlowRate = Math.Max(waterFlowRate, MinHotWaterFlowRate);

        //温水ボイラを更新
        hBoiler.OutletWaterSetPointTemperature = this.OutletWaterSetPointTemperature;
        hBoiler.Update(InletWaterTemperature, WaterFlowRate);
        IsOverLoad = hBoiler.IsOverLoad;
        HeatingLoad = hBoiler.HeatLoad;
        CoolingLoad = 0;
        OutletWaterTemperature = hBoiler.OutletWaterTemperature;
        FuelConsumption = hBoiler.FuelConsumption;
        ElectricConsumption = HeatingLoad / NominalHeatingCapacity * NominalElectricConsumption;
      }
      //運転停止
      else ShutOff();
    }

    /// <summary>運転を停止させる</summary>
    public void ShutOff()
    {
      OutletWaterTemperature = InletWaterTemperature;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      FuelConsumption = CoolingLoad = HeatingLoad = CoolingWaterFlowRate = WaterFlowRate = 0;
      ElectricConsumption = 0;
    }

    #endregion

  }

  #region Immutableインターフェース

  /// <summary>読み取り専用の直だき二重効用冷温水機</summary>
  public interface ImmutableDirectFiredAbsorptionChiller
  {
    /// <summary>冷却運転モードか否かを取得する</summary>
    bool IsCoolingMode { get; }

    /// <summary>燃料の種別を取得する</summary>
    Boiler.Fuel Fuel { get; }

    /// <summary>出口水温[C]を取得する</summary>
    double OutletWaterTemperature { get; }

    /// <summary>出口水温設定値[C]を取得する</summary>
    double OutletWaterSetPointTemperature { get; }

    /// <summary>入口水温[C]を取得する</summary>
    double InletWaterTemperature { get; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    double CoolingWaterOutletTemperature { get; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    double CoolingWaterInletTemperature { get; }

    /// <summary>冷水/温水の水量[kg/s]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    double CoolingWaterFlowRate { get; }

    /// <summary>冷水量上限値[kg/s]を取得する</summary>
    double MaxChilledWaterFlowRate { get; }

    /// <summary>温水量上限値[kg/s]を取得する</summary>
    double MaxHotWaterFlowRate { get; }

    /// <summary>冷水/温水の水量下限値[kg/s]を取得する</summary>
    double MinChilledWaterFlowRate { get; }

    /// <summary>冷却水流量上限値[kg/s]を取得する</summary>
    double MaxCoolingWaterFlowRate { get; }

    /// <summary>冷却水流量下限値[kg/s]を取得する</summary>
    double MinCoolingWaterFlowRate { get; }

    /// <summary>定格冷却能力[kW]を取得する</summary>
    double NominalCoolingCapacity { get; }

    /// <summary>定格加熱能力[kW]を取得する</summary>
    double NominalHeatingCapacity { get; }

    /// <summary>定格冷却燃料消費量[Nm3/s, kg/s]を取得する</summary>
    double NominalCoolingFuelConsumption { get; }

    /// <summary>定格加熱燃料消費量[Nm3/s, kg/s]を取得する</summary>
    double NominalHeatingFuelConsumption { get; }

    /// <summary>容量制御下限値[-]を取得する</summary>
    double MinimumPartialLoadRatio { get; }

    /// <summary>定格消費電力[kW]を取得する</summary>
    double NominalElectricConsumption { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>燃料消費量[Nm3/s, kg/s]を取得する</summary>
    double FuelConsumption { get; }

    /// <summary>加熱負荷[kW]を取得する</summary>
    double HeatingLoad { get; }

    /// <summary>冷却負荷[kW]を取得する</summary>
    double CoolingLoad { get; }

    /// <summary>COP[-]を取得する</summary>
    double COP { get; }

    /// <summary>溶液ポンプのインバータ制御があるか</summary>
    bool HasSolutionInverterPump { get; }

    /// <summary>高濃度溶液の質量分率[-]を取得する</summary>
    double ThickSolutionMassFraction { get; }

    /// <summary>低濃度溶液の質量分率[-]を取得する</summary>
    double ThinSolutionMassFraction { get; }

    /// <summary>蒸発温度[C]を取得する</summary>
    double EvaporatingTemperature { get; }

    /// <summary>凝縮温度[C]を取得する</summary>
    double CondensingTemperature { get; }

    /// <summary>再生温度[C]を取得する</summary>
    double DesorbTemperature { get; }

    /// <summary>過負荷か否か</summary>
    bool IsOverLoad { get; }
  }

  #endregion

}
