﻿/* HotWaterBoiler.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>温水ボイラ</summary>
  public class HotWaterBoiler: ImmutableHotWaterBoiler
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>空気比[-]</summary>
    private double airRatio = 1.1;

    /// <summary>熱損失係数[kW/K]</summary>
    private double heatLossCoefficient = 0.0;

    /// <summary>定格出口水温[C]</summary>
    private double nomOutletWaterTemperature;

    /// <summary>定格排ガス温度[C]</summary>
    private double nominalSmokeTemperature;

    /// <summary>電力の一次エネルギー換算係数[MJ/kWh]を設定・取得する</summary>
    public double PrimaryEnergyFactor { get; set; } = 9.76;

    /// <summary>燃料種別を取得する</summary>
    public Boiler.Fuel Fuel { get; private set;}

    /// <summary>温水出口温度[C]を取得する</summary>
    public double OutletWaterTemperature { get; private set; }

    /// <summary>温水出口温度設定値[C]を設定・取得する</summary>
    public double OutletWaterSetPointTemperature { get; set; }

    /// <summary>温水入口温度[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxWaterFlowRate { get; private set; }

    /// <summary>温水下限流量比[-]を設定・取得する</summary>
    public double MinWaterFlowRatio { get; set; } = 0.2;

    /// <summary>定格加熱能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>定格燃料消費量[kg/s, Nm3/s]を取得する</summary>
    public double NominalFuelConsumption { get; private set; }
    
    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>燃料消費量[kg/s, Nm3/s]を取得する</summary>
    public double FuelConsumption { get; private set; }

    /// <summary>暖房能力[kW]を取得する</summary>
    public double HeatLoad { get; private set; }
    
    /// <summary>周囲の温度[C]を設定・取得する</summary>
    public double AmbientTemperature { get; set; }

    /// <summary>空気比を設定・取得する</summary>
    public double AirRatio
    {
      get { return airRatio; }
      set { airRatio = Math.Max(1.0, Math.Min(1.5, value)); }
    }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (HeatLoad == 0) return 0;
        else
          return 0.001 * HeatLoad / (ElectricConsumption * PrimaryEnergyFactor 
            + FuelConsumption * Boiler.GetHeatingValue(Fuel, true));
      }
    }

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="inletWaterTemperature">入口温水温度[C]</param>
    /// <param name="outletWaterTemperature">出口温水温度[C]</param>
    /// <param name="waterFlowRate">温水流量[kg/s]</param>
    /// <param name="fuelConsumption">燃料消費量[kg/s, Nm3/s]</param>
    /// <param name="electricConsumption">消費電力[kW]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="fuel">燃料種別</param>
    /// <param name="smokeTemperature">排ガス温度[C]</param>
    public HotWaterBoiler
      (double inletWaterTemperature, double outletWaterTemperature, double waterFlowRate,
      double fuelConsumption, double electricConsumption, double ambientTemperature,
      double airRatio, Boiler.Fuel fuel, double smokeTemperature)
    {
      //プロパティ保存
      Fuel = fuel;
      AirRatio = airRatio;
      nominalSmokeTemperature = smokeTemperature;
      NominalFuelConsumption = fuelConsumption;
      ElectricConsumption = electricConsumption;
      AmbientTemperature = ambientTemperature;
      OutletWaterSetPointTemperature = nomOutletWaterTemperature = outletWaterTemperature;
      MaxWaterFlowRate  = WaterFlowRate = waterFlowRate;
      NominalCapacity = (outletWaterTemperature - inletWaterTemperature) * WaterFlowRate * WATER_SPECIFIC_HEAT;

      //熱損失係数の計算
      heatLossCoefficient = Boiler.GetHeatLossCoefficient
        (NominalCapacity, outletWaterTemperature, AmbientTemperature, fuel,
        nominalSmokeTemperature, AirRatio, NominalFuelConsumption, AmbientTemperature);

      //停止させる
      ShutOff();
    }

    #endregion

    #region publicメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    public void Update(double inletWaterTemperature, double waterFlowRate)
    {
      //入力条件保存
      InletWaterTemperature = inletWaterTemperature;
      WaterFlowRate = waterFlowRate;

      //流量0、設定温度<入口温度で停止
      if (WaterFlowRate <= 0 || OutletWaterSetPointTemperature < InletWaterTemperature) ShutOff();

      //燃料消費量を計算
      double hl = WATER_SPECIFIC_HEAT * waterFlowRate * (OutletWaterSetPointTemperature - inletWaterTemperature);
      FuelConsumption = Boiler.GetFuelConsumption(hl, OutletWaterSetPointTemperature, AmbientTemperature, Fuel,
        nominalSmokeTemperature, airRatio, heatLossCoefficient, AmbientTemperature, nomOutletWaterTemperature);

      //過負荷の場合には成り行き計算
      IsOverLoad = NominalFuelConsumption < FuelConsumption;
      if (IsOverLoad)
      {
        FuelConsumption = NominalFuelConsumption;
        double to;
        Boiler.GetOutletWaterTemperature(InletWaterTemperature, WaterFlowRate, 
          NominalFuelConsumption, AmbientTemperature, Fuel, nominalSmokeTemperature, 
          AirRatio, heatLossCoefficient, AmbientTemperature, nomOutletWaterTemperature, out to, out hl);
        OutletWaterTemperature = to;
      }
      else OutletWaterTemperature = OutletWaterSetPointTemperature;
      HeatLoad = hl;
    }

    /// <summary>停止する</summary>
    public void ShutOff()
    {
      OutletWaterTemperature = InletWaterTemperature;
      ElectricConsumption = WaterFlowRate = HeatLoad = FuelConsumption = 0;     
    }

    #endregion

  }

  #region 読み取り専用温水ボイラインターフェース

  /// <summary>読み取り専用温水ボイラ</summary>
  public interface ImmutableHotWaterBoiler
  {
    
    /// <summary>燃料種別を取得する</summary>
    Boiler.Fuel Fuel { get; }

    /// <summary>温水出口温度[C]を取得する</summary>
    double OutletWaterTemperature { get; }

    /// <summary>温水出口温度設定値[C]を取得する</summary>
    double OutletWaterSetPointTemperature { get; }

    /// <summary>温水入口温度[C]を取得する</summary>
    double InletWaterTemperature { get; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    double MaxWaterFlowRate { get; }

    /// <summary>温水下限流量比[-]を取得する</summary>
    double MinWaterFlowRatio { get; }

    /// <summary>定格加熱能力[kW]を取得する</summary>
    double NominalCapacity { get; }

    /// <summary>定格燃料消費量[kg/s, Nm3/s]を取得する</summary>
    double NominalFuelConsumption { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>燃料消費量[kg/s, Nm3/s]を取得する</summary>
    double FuelConsumption { get; }

    /// <summary>暖房能力[kW]を取得する</summary>
    double HeatLoad { get; }

    /// <summary>周囲の温度[C]を取得する</summary>
    double AmbientTemperature { get; }

    /// <summary>空気比を取得する</summary>
    double AirRatio { get; }

    /// <summary>COP[-]を取得する</summary>
    double COP { get; }

    /// <summary>過負荷か否か</summary>
    bool IsOverLoad { get; }
  }

  #endregion

}
