﻿/* IndoorUnit.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.HVAC.HeatExchanger;

namespace Popolo.HVAC.MultiplePackagedHeatPump
{
  /// <summary>室内機</summary>
  public class IndoorUnit : ImmutableIndoorUnit
  {

    #region 定数宣言

    /// <summary>大気圧[kPa]は固定:海抜0m</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>定格能力と最大能力の比率[-]</summary>
    private const double MAX_CAP_RATIO = 1.2;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>暖房用熱交換器</summary>
    private CrossFinCondensor heatingHEX;

    /// <summary>冷房用熱交換器</summary>
    private CrossFinEvaporator coolingHEX;

    /// <summary>定格風量[kg/s]を取得する</summary>
    public double NominalAirFlowRate
    {
      get
      {
        if (IsCooling) return coolingHEX.NominalAirFlowRate;
        else return heatingHEX.NominalAirFlowRate;
      }
    }

    /// <summary>ファンのON/OFF制御を設定・取得する</summary>
    public bool OnOffControlledFan { get; set; }

    /// <summary>室内空気条件を設定・取得する</summary>
    public ImmutableMoistAir IndoorAir { get; set; }

    /// <summary>出口温度[C]を取得する</summary>
    public double OutletAirDrybulbTemperature
    {
      get
      {
        if (IsCooling) return coolingHEX.OutletAirTemperature;
        else return heatingHEX.OutletAirTemperature;
      }
    }

    /// <summary>出口絶対湿度[kg/kg]を取得する</summary>
    public double OutletAirHumidityRatio
    {
      get
      {
        if (IsCooling) return coolingHEX.OutletAirHumidityRatio;
        else return heatingHEX.OutletAirHumidityRatio;
      }
    }

    /// <summary>定格消費電力を設定・取得する</summary>
    public double NominalElectricConsumption { get; private set; }

    /// <summary>稼働率[-]を取得する</summary>
    public double OperatingRate { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption
    {
      get
      {
        if (!OnOffControlledFan) return NominalElectricConsumption;
        else return NominalElectricConsumption * OperatingRate;
      }
    }

    /// <summary>冷房モードか否か</summary>
    public bool IsCooling { get; set; }

    /// <summary>熱負荷[kW]を設定・取得する（加熱負荷が正）</summary>
    public double HeatLoad { get; set; }

    /// <summary>未処理負荷[kW]を取得する（加熱負荷が正）</summary>
    public double OverLoad { get; private set; }

    /// <summary>発停を設定・取得する</summary>
    public bool IsOperating { get; set; }

    /// <summary>暖房時配管長補正係数[-]を設定・取得する</summary>
    public double PipeLengthCorrectionFactor_Heating { get; set; }

    /// <summary>冷房時配管長補正係数[-]を設定・取得する</summary>
    public double PipeLengthCorrectionFactor_Cooling { get; set; }

    /// <summary>最大暖房能力[kW]を取得する</summary>
    public double MaxHeatingCapacity { get; private set; }

    /// <summary>最大冷房能力[kW]を取得する</summary>
    public double MaxCoolingCapacity { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ（JIS8627条件）</summary>
    /// <param name="evaporatingTemperature">定格蒸発温度[C]</param>
    /// <param name="condensingTemperature">定格凝縮温度[C]</param>
    /// <param name="coolingCapacity">定格冷却能力[kW]</param>
    /// <param name="heatingCapacity">定格暖房能力[kW]</param>
    /// <param name="electricConsumption">定格消費電力[kW]</param>
    /// <param name="borderRalativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="onOffControlledFan">On/Off制御ファンか否か</param>
    public IndoorUnit
      (double evaporatingTemperature, double condensingTemperature,
      double coolingCapacity, double heatingCapacity,
      double electricConsumption, double borderRalativeHumidity, bool onOffControlledFan)
    {
      //冷却能力から風量[kg/s]を推定
      double aflow = ((0.014 * coolingCapacity + 2.1609) * coolingCapacity) * 1.2 / 60d;

      initialize(evaporatingTemperature, condensingTemperature,
        27, 20, 0.011, 0.0072, coolingCapacity, heatingCapacity, aflow, aflow,
        electricConsumption, borderRalativeHumidity, onOffControlledFan);
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="evaporatingTemperature">定格蒸発温度[C]</param>
    /// <param name="condensingTemperature">定格凝縮温度[C]</param>
    /// <param name="coolingAirDrybulbTemperature">冷房時定格乾球温度[C]</param>
    /// <param name="heatingAirDrybulbTemperature">暖房時定格乾球温度[C]</param>
    /// <param name="coolingAirHumidityRatio">冷房時定格絶対湿度[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">暖房時定格絶対湿度[kg/kg]</param>
    /// <param name="coolingCapacity">定格冷却能力[kW]</param>
    /// <param name="heatingCapacity">定格暖房能力[kW]</param>
    /// <param name="electricConsumption">定格消費電力[kW]</param>
    /// <param name="borderRalativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="onOffControlledFan">On/Off制御ファンか否か</param>
    public IndoorUnit
      (double evaporatingTemperature, double condensingTemperature,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingCapacity, double heatingCapacity,
      double electricConsumption, double borderRalativeHumidity, bool onOffControlledFan)
    {
      //冷却能力から風量[kg/s]を推定
      double aflow = ((0.014 * coolingCapacity + 2.1609) * coolingCapacity) * 1.2 / 60d;

      initialize(evaporatingTemperature, condensingTemperature, coolingAirDrybulbTemperature,
        heatingAirDrybulbTemperature, coolingAirHumidityRatio, heatingAirHumidityRatio,
        coolingCapacity, heatingCapacity, aflow, aflow,
        electricConsumption, borderRalativeHumidity, onOffControlledFan);
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="evaporatingTemperature">定格蒸発温度[C]</param>
    /// <param name="condensingTemperature">定格凝縮温度[C]</param>
    /// <param name="coolingAirDrybulbTemperature">冷房時定格乾球温度[C]</param>
    /// <param name="heatingAirDrybulbTemperature">暖房時定格乾球温度[C]</param>
    /// <param name="coolingAirHumidityRatio">冷房時定格絶対湿度[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">暖房時定格絶対湿度[kg/kg]</param>
    /// <param name="coolingCapacity">定格冷却能力[kW]</param>
    /// <param name="heatingCapacity">定格暖房能力[kW]</param>
    /// <param name="coolingAirFlowRate">冷房時定格風量[kg/s]</param>
    /// <param name="heatingAirFlowRate">暖房時定格風量[kg/s]</param>
    /// <param name="electricConsumption">定格消費電力[kW]</param>
    /// <param name="borderRalativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="onOffControlledFan">On/Off制御ファンか否か</param>
    public IndoorUnit
      (double evaporatingTemperature, double condensingTemperature,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingCapacity, double heatingCapacity,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double electricConsumption, double borderRalativeHumidity, bool onOffControlledFan)
    {
      initialize(evaporatingTemperature, condensingTemperature, coolingAirDrybulbTemperature, 
        heatingAirDrybulbTemperature, coolingAirHumidityRatio, heatingAirHumidityRatio,
        coolingCapacity, heatingCapacity, coolingAirFlowRate, heatingAirFlowRate,
        electricConsumption, borderRalativeHumidity, onOffControlledFan);
    }

    /// <summary>初期化処理</summary>
    /// <param name="evaporatingTemperature">定格蒸発温度[C]</param>
    /// <param name="condensingTemperature">定格凝縮温度[C]</param>
    /// <param name="coolingAirDrybulbTemperature">冷房時定格乾球温度[C]</param>
    /// <param name="heatingAirDrybulbTemperature">暖房時定格乾球温度[C]</param>
    /// <param name="coolingAirHumidityRatio">冷房時定格絶対湿度[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">暖房時定格絶対湿度[kg/kg]</param>
    /// <param name="coolingCapacity">定格冷却能力[kW]</param>
    /// <param name="heatingCapacity">定格暖房能力[kW]</param>
    /// <param name="coolingAirFlowRate">冷房時定格風量[kg/s]</param>
    /// <param name="heatingAirFlowRate">暖房時定格風量[kg/s]</param>
    /// <param name="electricConsumption">定格消費電力[kW]</param>
    /// <param name="borderRalativeHumidity">乾湿境界相対湿度[%]</param>
    /// <param name="onOffControlledFan">On/Off制御ファンか否か</param>
    private void initialize
      (double evaporatingTemperature, double condensingTemperature,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingCapacity, double heatingCapacity,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double electricConsumption, double borderRalativeHumidity, bool onOffControlledFan)
    {

      //熱交換器初期化
      coolingHEX = new CrossFinEvaporator
        (evaporatingTemperature, coolingCapacity, coolingAirFlowRate,
        coolingAirDrybulbTemperature, coolingAirHumidityRatio, borderRalativeHumidity);
      heatingHEX = new CrossFinCondensor
        (condensingTemperature, heatingCapacity, heatingAirFlowRate,
        heatingAirDrybulbTemperature, heatingAirHumidityRatio);

      //プロパティ初期化
      NominalElectricConsumption = electricConsumption;
      OperatingRate = 1;
      OnOffControlledFan = onOffControlledFan;
      IndoorAir = new MoistAir(27, 0.011);
      IsOperating = true;
      PipeLengthCorrectionFactor_Cooling = 1.0;
      PipeLengthCorrectionFactor_Heating = 1.0;
      MaxHeatingCapacity = heatingCapacity * MAX_CAP_RATIO;
      MaxCoolingCapacity = -coolingCapacity * MAX_CAP_RATIO;
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>必要な冷媒温度[C]を計算する</summary>
    /// <param name="outletAirDrybulbTemperature">出口乾球温度設定値[C]</param>
    /// <returns>必要な冷媒温度[C]</returns>
    public double GetRefrigerantTemperature
      (double outletAirDrybulbTemperature)
    {
      if (IsCooling)
      {
        double hr = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (outletAirDrybulbTemperature, coolingHEX.BorderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        if (hr < IndoorAir.HumidityRatio)
        {
          double ho = MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
            (outletAirDrybulbTemperature, hr);
          HeatLoad = Math.Min(0, (ho - IndoorAir.Enthalpy) * coolingHEX.AirFlowRate);
        }
        else
        {
          HeatLoad = Math.Min(0, MoistAir.GetSpecificHeat(IndoorAir.HumidityRatio) * coolingHEX.AirFlowRate
            * (outletAirDrybulbTemperature - IndoorAir.DrybulbTemperature));
        }
      }
      else
      {
        HeatLoad = Math.Max(0, MoistAir.GetSpecificHeat(IndoorAir.HumidityRatio) * coolingHEX.AirFlowRate
            * (outletAirDrybulbTemperature - IndoorAir.DrybulbTemperature));
      }
      return GetRefrigerantTemperature();
    }

    /// <summary>必要な冷媒温度[C]を計算する</summary>
    /// <returns>必要な冷媒温度[C]</returns>
    public double GetRefrigerantTemperature()
    {
      if (IsCooling)
      {
        if (!IsOperating) return 100;
        else return coolingHEX.GetEvaporatingTemperature
          (Math.Max(-HeatLoad, MaxCoolingCapacity) / PipeLengthCorrectionFactor_Cooling, coolingHEX.NominalAirFlowRate,
          IndoorAir.DrybulbTemperature, IndoorAir.HumidityRatio, false);
      }
      else
      {
        if (!IsOperating) return -100;
        else return heatingHEX.GetCondensingTemperature
          (Math.Min(HeatLoad,MaxHeatingCapacity) / PipeLengthCorrectionFactor_Heating, heatingHEX.NominalAirFlowRate,
          IndoorAir.DrybulbTemperature, IndoorAir.HumidityRatio);
      }
    }

    /// <summary>顕熱負荷[kW]から全熱負荷[kW]を設定する</summary>
    /// <param name="sensibleHeatLoad">顕熱負荷[kW]（加熱が正）</param>
    public void SetHeatLoad(double sensibleHeatLoad)
    {
      const double DELTA = 0.001;
      const double ERR_TOL = 0.00001;
      if (IsCooling)
      {
        double dbt = MoistAir.GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
          (IndoorAir.HumidityRatio, coolingHEX.BorderRelativeHumidity, ATMOSPHERIC_PRESSURE);
        double sh = (dbt - IndoorAir.DrybulbTemperature)
          * MoistAir.GetSpecificHeat(IndoorAir.HumidityRatio) * NominalAirFlowRate;
        if (sh < sensibleHeatLoad) HeatLoad = Math.Min(0, sensibleHeatLoad);
        else
        {
          dbt = IndoorAir.DrybulbTemperature + sensibleHeatLoad / NominalAirFlowRate;
          double hrt = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
              (dbt, coolingHEX.BorderRelativeHumidity, ATMOSPHERIC_PRESSURE);
          double val = sensibleHeatLoad / NominalAirFlowRate + MoistAir.GetSpecificHeat(hrt) * IndoorAir.DrybulbTemperature;
          double err1 = val - MoistAir.GetSpecificHeat(hrt) * dbt;          
          while (ERR_TOL < Math.Abs(err1))
          {
            hrt = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
              (dbt + DELTA, coolingHEX.BorderRelativeHumidity, ATMOSPHERIC_PRESSURE);
            double err2 = val - MoistAir.GetSpecificHeat(hrt) * (dbt + DELTA);
            dbt -= err1 * DELTA / (err2 - err1);
            hrt = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
              (dbt, coolingHEX.BorderRelativeHumidity, ATMOSPHERIC_PRESSURE);
            err1 = val - MoistAir.GetSpecificHeat(hrt) * dbt;
          }
          HeatLoad = (MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(IndoorAir.DrybulbTemperature, IndoorAir.HumidityRatio)
            - MoistAir.GetEnthalpyFromDryBulbTemperatureAndHumidityRatio(dbt, hrt)) * NominalAirFlowRate;
          HeatLoad = Math.Min(0, - HeatLoad);
        }
      }
      else HeatLoad = Math.Max(0, sensibleHeatLoad);
    }

    /// <summary>冷媒温度[C]にもとづき状態を更新する</summary>
    /// <param name="refrigerantTemperature">冷媒温度[C]</param>
    public void UpdateState(double refrigerantTemperature)
    {
      double q;
      if (!IsOperating) q = OverLoad = 0;
      else if (IsCooling)
      {
        q = coolingHEX.GetHeatTransfer
          (refrigerantTemperature, coolingHEX.NominalAirFlowRate,
          IndoorAir.DrybulbTemperature, IndoorAir.HumidityRatio)
          * PipeLengthCorrectionFactor_Cooling;
        if (q < -HeatLoad) OverLoad = HeatLoad + q;
        else OverLoad = 0;
      }
      else
      {
        q = heatingHEX.GetHeatTransfer
          (refrigerantTemperature, coolingHEX.NominalAirFlowRate,
          IndoorAir.DrybulbTemperature, IndoorAir.HumidityRatio)
          * PipeLengthCorrectionFactor_Heating;
        if (q < HeatLoad) OverLoad = HeatLoad - q;
        else OverLoad = 0;
      }
      //稼働率の計算
      if (!IsOperating) OperatingRate = 0;
      else if (OnOffControlledFan) OperatingRate = (HeatLoad - OverLoad) / q;
      else OperatingRate = 1.0;
    }

    #endregion    

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用室内機</summary>
  public interface ImmutableIndoorUnit
  {
    /// <summary>定格風量[kg/s]を取得する</summary>
    double NominalAirFlowRate { get; }

    /// <summary>ファンのON/OFF制御を取得する</summary>
    bool OnOffControlledFan { get; }

    /// <summary>室内空気条件を取得する</summary>
    ImmutableMoistAir IndoorAir { get; }

    /// <summary>出口温度[C]を取得する</summary>
    double OutletAirDrybulbTemperature { get; }

    /// <summary>出口絶対湿度[kg/kg]を取得する</summary>
    double OutletAirHumidityRatio { get; }

    /// <summary>定格消費電力を取得する</summary>
    double NominalElectricConsumption { get; }

    /// <summary>稼働率[-]を取得する</summary>
    double OperatingRate { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>冷房モードか否か</summary>
    bool IsCooling { get; }

    /// <summary>熱負荷[kW]を取得する（加熱負荷が正）</summary>
    double HeatLoad { get; }

    /// <summary>未処理負荷[kW]を取得する（加熱負荷が正）</summary>
    double OverLoad { get; }

    /// <summary>暖房時配管長補正係数[-]を取得する</summary>
    double PipeLengthCorrectionFactor_Heating { get; }

    /// <summary>冷房時配管長補正係数[-]を取得する</summary>
    double PipeLengthCorrectionFactor_Cooling { get; }

  }

  #endregion

}
