﻿/* AirHeatSourceModularChillersSystem.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.HVAC.HeatSource;
using Popolo.ThermophysicalProperty;
using Popolo.HVAC.Circuit;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>空気熱源モジュールヒートポンプによる熱源サブシステム</summary>
  public class AirHeatSourceModularChillersSystem : IHeatSourceSubSystem
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>空気熱源ヒートポンプ</summary>
    private AirHeatSourceModularChillers mChiller;

    /// <summary>冷水・温水ポンプ</summary>
    private CentrifugalPump chwPump, hwPump;

    /// <summary>空気熱源ヒートポンプを取得する</summary>
    public ImmutableAirHeatSourceModularChillers AirHeatSourceModularChillers { get { return mChiller; } }

    /// <summary>冷水ポンプを取得する</summary>
    public ImmutableCentrifugalPump ChilledWaterPump { get { return chwPump; } }

    /// <summary>温水ポンプを取得する</summary>
    public ImmutableCentrifugalPump HotWaterPump { get { return hwPump; } }

    /// <summary>モジュールチラーのセット数を取得する</summary>
    public int ChillerNumber { get; private set; }

    /// <summary>運転台数を取得する</summary>
    public int OperatingChillerNumber { get; private set; }

    #endregion

    #region IHeatSourceSubSystem実装

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode { get; }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; } = 40;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate
    { get { return mChiller.MaxHotWaterFlowRate * ChillerNumber; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio
    { get { return mChiller.MinHotWaterFlowRate / MaxHotWaterFlowRate; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; } = 12;

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint { get; set; } = 7;

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate
    { get { return mChiller.MaxChilledWaterFlowRate * ChillerNumber; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio
    { get { return mChiller.MinChilledWaterFlowRate / MaxChilledWaterFlowRate; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_C = IsOverLoad_H = false;
      OperatingChillerNumber = 0;
      mChiller.ShutOff();
      chwPump.ShutOff();
      hwPump.ShutOff();
    }

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;

      if (Mode == HeatSourceSystemModel.OperatingMode.Cooling) forecastCooling(chilledWaterFlowRate);
      else if (Mode == HeatSourceSystemModel.OperatingMode.Heating) forecastHeating(hotWaterFlowRate);
      else ShutOff();
    }

    /// <summary>将来状態を予測する（暖房）</summary>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    private void forecastHeating(double hotWaterFlowRate)
    {
      if (hotWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }
      mChiller.Mode = HeatSource.AirHeatSourceModularChillers.OperatingMode.Heating;
      ChilledWaterSupplyTemperature = ChilledWaterReturnTemperature;

      OperatingChillerNumber = (int)Math.Ceiling(hotWaterFlowRate / mChiller.MaxHotWaterFlowRate);
      OperatingChillerNumber = Math.Min(ChillerNumber, OperatingChillerNumber);
      mChiller.WaterOutletSetPointTemperature = HotWaterSupplyTemperatureSetpoint;
      while (true)
      {
        //ポンプによる昇温を評価
        double hwFlow = hotWaterFlowRate / OperatingChillerNumber;
        hwPump.UpdateState(0.001 * hwFlow);
        double twi = HotWaterReturnTemperature + hwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * hwFlow);

        mChiller.Update(twi, hwFlow, OutdoorAir.DrybulbTemperature);
        IsOverLoad_H = mChiller.IsOverLoad;
        if (OperatingChillerNumber == ChillerNumber || !IsOverLoad_H) break;
        else OperatingChillerNumber++;
      }

      if (mChiller.IsOverLoad) HotWaterSupplyTemperature = mChiller.WaterOutletTemperature;
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;
    }

    /// <summary>将来状態を予測する（冷房）</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    private void forecastCooling(double chilledWaterFlowRate)
    {
      if (chilledWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }
      mChiller.Mode = HeatSource.AirHeatSourceModularChillers.OperatingMode.Cooling;
      HotWaterSupplyTemperature = HotWaterReturnTemperature;

      OperatingChillerNumber = (int)Math.Ceiling(chilledWaterFlowRate / mChiller.MaxChilledWaterFlowRate);
      OperatingChillerNumber = Math.Min(ChillerNumber, OperatingChillerNumber);
      mChiller.WaterOutletSetPointTemperature = ChilledWaterSupplyTemperatureSetpoint;
      while (true)
      {
        //ポンプによる昇温を評価
        double chwFlow = chilledWaterFlowRate / OperatingChillerNumber;
        chwPump.UpdateState(0.001 * chwFlow);
        double twi = ChilledWaterReturnTemperature + chwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * chwFlow);

        mChiller.Update(twi, chwFlow, OutdoorAir.DrybulbTemperature);
        IsOverLoad_C = mChiller.IsOverLoad;
        if (OperatingChillerNumber == ChillerNumber || !IsOverLoad_C) break;
        else OperatingChillerNumber++;
      }

      if (mChiller.IsOverLoad) ChilledWaterSupplyTemperature = mChiller.WaterOutletTemperature;
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="mChiller">モジュールチラー</param>
    /// <param name="chwPump">冷水ポンプ</param>
    /// <param name="hwPump">温水ポンプ</param>
    /// <param name="number">モジュールチラーの数</param>
    public AirHeatSourceModularChillersSystem
      (AirHeatSourceModularChillers mChiller, CentrifugalPump chwPump, CentrifugalPump hwPump, int number)
    {
      this.mChiller = mChiller;
      this.ChillerNumber = number;
      this.chwPump = chwPump;
      this.hwPump = hwPump;

      //加熱運転・冷却運転対応
      SelectableMode = HeatSourceSystemModel.OperatingMode.Cooling | HeatSourceSystemModel.OperatingMode.Heating;
      Mode = HeatSourceSystemModel.OperatingMode.ShutOff;
    }

    #endregion

  }
}
