﻿using System;

using Popolo.HVAC.HeatSource;
using Popolo.HVAC.Circuit;
using Popolo.HVAC.HeatExchanger;
using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>ターボ冷凍機による熱源サブシステム</summary>
  public class CentrifugalChillerSystem : IHeatSourceSubSystem
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>ターボ冷凍機</summary>
    private ICentrifugalChiller chiller;

    /// <summary>ポンプ</summary>
    private CentrifugalPump cdwPump, chwPump;

    /// <summary>冷却塔</summary>
    private CoolingTower cTower;

    /// <summary>ターボ冷凍機を取得する</summary>
    public ImmutableCentrifugalChiller Chiller { get { return chiller; } }

    /// <summary>冷水ポンプを取得する</summary>
    public ImmutableCentrifugalPump ChilledWaterPump { get { return chwPump; } }

    /// <summary>冷却水ポンプを取得する</summary>
    public ImmutableCentrifugalPump CoolingWaterPump { get { return cdwPump; } }

    /// <summary>冷却塔を取得する</summary>
    public ImmutableCoolingTower CoolingTower { get { return cTower; } }

    /// <summary>冷凍機の台数を取得する</summary>
    public int ChillerNumber { get; private set; }

    /// <summary>冷却塔のセル数（1台の冷凍機あたり）を取得する</summary>
    public int CoolingTowerNumber { get; private set; }

    /// <summary>運転台数を取得する</summary>
    public int OperatingChillerNumber { get; private set; }

    /// <summary>冷却塔を1対1で稼働させるか否か</summary>
    public bool OperateCoolingTowerOneOnOne { get; set; } = true;

    /// <summary>冷却水温度を制御するか否か</summary>
    public bool ControlCoolingWaterTemperature { get; set; }

    /// <summary>冷却水流量を制御するか否か（負荷率比例）</summary>
    public bool ControlCoolingWaterFlowRate { get; set; }

    /// <summary>冷却水温度設定値[C]を設定・取得する</summary>
    public double CoolingWaterTemperatureSetpoint { get; set; } = 32;

    /// <summary>冷却水流量[kg/s]を設定・取得する</summary>
    public double CoolingWaterFlowSetpoint { get; set; }

    #endregion

    #region IHeatSourceSubSystem実装

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode
    { get { return HeatSourceSystemModel.OperatingMode.Cooling; } }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; } = 40;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate
    { get { return 0; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio
    { get { return 0; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; } = 12;

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint { get; set; } = 7;

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate
    { get { return chiller.MaxChilledWaterFlowRate * ChillerNumber; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio
    { get { return chiller.MinChilledWaterFlowRatio / ChillerNumber; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_C = IsOverLoad_H = false;
      OperatingChillerNumber = 0;
      chiller.ShutOff();
      chwPump.ShutOff();
      cdwPump.ShutOff();
      cTower.ShutOff();
    }

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;

      if (Mode != HeatSourceSystemModel.OperatingMode.Cooling || chilledWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }

      chiller.IsOperating = true;
      cTower.SetOutdoorAirState(OutdoorAir.WetbulbTemperature, OutdoorAir.HumidityRatio);

      OperatingChillerNumber = (int)Math.Ceiling(chilledWaterFlowRate / chiller.MaxChilledWaterFlowRate);
      OperatingChillerNumber = Math.Min(ChillerNumber, OperatingChillerNumber);
      chiller.ChilledWaterOutletSetPointTemperature = ChilledWaterSupplyTemperatureSetpoint;
      while (true)
      {
        //冷水・冷却水流量を計算
        double chwFlow = chilledWaterFlowRate / OperatingChillerNumber;
        double pLoad = chwFlow / chiller.MaxChilledWaterFlowRate;
        double cdwFlow;
        if (ControlCoolingWaterFlowRate) cdwFlow = CoolingWaterFlowSetpoint; 
        else cdwFlow = cTower.MaxWaterFlowRate * CoolingTowerNumber * pLoad;
        if (!OperateCoolingTowerOneOnOne) cdwFlow /= ChillerNumber;
        cTower.WaterFlowRate = cdwFlow / CoolingTowerNumber;

        //ポンプによる昇温を評価
        cdwPump.UpdateState(0.001 * cdwFlow);
        double dCDT = cdwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * cdwFlow);
        chwPump.UpdateState(0.001 * chwFlow);
        double twi = ChilledWaterReturnTemperature
          + chwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * chwFlow);

        //冷却塔と冷凍機の連成計算（冷却水温度の計算）
        bool needIteration = !ControlCoolingWaterTemperature;
        if (ControlCoolingWaterTemperature)
        {
          chiller.Update(CoolingWaterTemperatureSetpoint + dCDT, twi, cdwFlow, chwFlow);
          cTower.OutletWaterSetPointTemperature = CoolingWaterTemperatureSetpoint;
          cTower.Update(chiller.CoolingWaterOutletTemperature, true);
          if (cTower.IsOverLoad) needIteration = true;
        }
        //過負荷または冷却水温度成行の場合には収束計算
        if (needIteration)
        {
          Roots.ErrorFunction eFnc = delegate (double cdt)
          {
            chiller.Update(cdt + dCDT, twi, cdwFlow, chwFlow);
            cTower.Update(chiller.CoolingWaterOutletTemperature, cTower.MaxAirFlowRate);
            return cTower.OutletWaterTemperature - cdt;
          };
          if (eFnc(10) < 0) break;
          else if (0 < eFnc(32)) break;
          else Roots.Bisection(eFnc, 10, 32, 0.01, 0.001, 10);
        }

        if (OperatingChillerNumber == ChillerNumber || !chiller.IsOverLoad) break;
        else OperatingChillerNumber++;
      }

      IsOverLoad_C = chiller.IsOverLoad;
      if (IsOverLoad_C) ChilledWaterSupplyTemperature = chiller.ChilledWaterOutletTemperature;
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="chiller">ターボ冷凍機</param>
    /// <param name="chwPump">冷水ポンプ</param>
    /// <param name="cdwPump">冷却水ポンプ</param>
    /// <param name="cTower">冷却塔</param>
    /// <param name="chillerNumber">冷凍機台数</param>
    /// <param name="coolingTowerNumber">冷却塔台数（1台の冷凍機あたりの台数）</param>
    public CentrifugalChillerSystem
      (ICentrifugalChiller chiller, CentrifugalPump chwPump,
      CentrifugalPump cdwPump, CoolingTower cTower, int chillerNumber, int coolingTowerNumber)
    {
      this.chiller = chiller;
      this.chwPump = chwPump;
      this.cdwPump = cdwPump;
      this.cTower = cTower;
      this.ChillerNumber = chillerNumber;
      this.CoolingTowerNumber = coolingTowerNumber;

      CoolingWaterFlowSetpoint = cTower.MaxWaterFlowRate;
      Mode = HeatSourceSystemModel.OperatingMode.ShutOff;
    }

    #endregion

  }
}
