﻿using System;

using Popolo.Numerics.MatrixOperation;

namespace Popolo.HVAC.ThermalStorage
{
  /// <summary>地中熱採熱クラス</summary>
  /// <remarks>2016.10.10作成中・熱収支が合わない!!</remarks>
  public class GroundHeatSource
  {

    #region インスタンス変数・プロパティ定義

    /// <summary>初期化要否フラグ</summary>
    private bool needInitialize = true;

    /// <summary>係数行列</summary>
    private SparseMatrix mat;

    /// <summary>温度ベクトル（計算用×2・一時保存用）</summary>
    private IVector tempBack, tempA, tempB;

    /// <summary>温度分布</summary>
    private double[,] temperature;

    /// <summary>土壌の熱伝導率[W/(mK)]</summary>
    private double[] thermalConductivity;

    /// <summary>土壌の容積比熱[J/(m3K)]</summary>
    private double[] vSpecificHeat;

    /// <summary>熱容量[J/K]</summary>
    private double[,] hCap;
    
    /// <summary>各質点の幅と深さ[m]</summary>
    private double[] lenR, lenZ;

    /// <summary>質点間距離[m]</summary>
    private double[] deltaR, deltaZ;

    /// <summary>各質点の半径[m]</summary>
    private double[] radius;

    /// <summary>杭からの熱流[W/m2]</summary>
    private double[] heatFlow;

    /// <summary>総合熱伝達率[W/(m2K)]</summary>
    private double heatTransferCoefficient = 23;

    /// <summary>計算時間間隔[sec]</summary>
    private double timeStep = 3600;

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set
      {
        if (0 < value)
        {
          timeStep = value;
          needInitialize = true;
        }
      }
    }

    /// <summary>地表面の相当温度[C]を設定・取得する</summary>
    public double SolAirTemperature { get; set; } = 20;

    /// <summary>計算対象の土壌の半径[m]を取得する</summary>
    public double GroundRadius
    {
      get
      {
        double sum = PileRadius;
        for (int i = 0; i < RadialSplitNumber; i++) sum += lenR[i];
        return sum;
      }
    }

    /// <summary>計算対象の土壌の深さ[m]を取得する</summary>
    public double GroundDepth
    {
      get
      {
        double sum = 0;
        for (int i = 0; i < VerticalSplitNumber; i++) sum += lenZ[i];
        return sum;
      }
    }

    /// <summary>埋設杭の半径[m]を取得する</summary>
    public double PileRadius { get; private set; }

    /// <summary>埋設杭の長さ[m]を取得する</summary>
    public double PileLength { get; private set; }

    /// <summary>杭頭までの埋設深さ[m]を取得する</summary>
    public double PileDepth { get; private set; }

    /// <summary>半径方向への分割数[-]を取得する</summary>
    public int RadialSplitNumber { get; private set; }

    /// <summary>深さ方向への分割数[-]を取得する</summary>
    public int VerticalSplitNumber { get; private set; }

    /// <summary>地表面の総合熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double HeatTransferCoefficient
    {
      get { return heatTransferCoefficient; }
      set
      {
        needInitialize = true;
        heatTransferCoefficient = Math.Max(0, value);
      }
    }
    
    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="groundRadius">計算対象の土壌の半径[m]</param>
    /// <param name="groundDepth">計算対象の土壌の深さ[m]</param>
    /// <param name="pileRadius">埋設杭の半径[m]</param>
    /// <param name="pileLength">埋設杭の長さ[m]</param>
    /// <param name="pileDepth">杭頭までの埋設深さ[m]</param>
    /// <param name="rSplitNumber">半径方向への分割数[-]</param>
    /// <param name="vSplitNumber">深さ方向への分割数[-]</param>
    /// <param name="heatTransferCoef">地表面の総合熱伝達率[W/(m2K)]</param>
    /// <param name="thermalConductivity">土壌の熱伝導率[W/(mK)]</param>
    /// <param name="specificHeat">土壌の容積比熱[J/(m3K)]</param>
    public GroundHeatSource
      (double groundRadius, double groundDepth, double pileRadius, double pileLength, double pileDepth,
      int rSplitNumber, int vSplitNumber, double heatTransferCoef, double thermalConductivity, double specificHeat)
    {
      if (vSplitNumber < 2 || rSplitNumber < 2) throw new Exception("Split Number Error");

      //計算条件保存
      this.PileRadius = pileRadius;
      this.PileLength = pileLength;
      this.PileDepth = pileDepth;
      this.RadialSplitNumber = rSplitNumber;
      this.VerticalSplitNumber = vSplitNumber;
      this.HeatTransferCoefficient = heatTransferCoef;
      this.thermalConductivity = new double[VerticalSplitNumber];
      this.vSpecificHeat = new double[VerticalSplitNumber];
      for (int i = 0; i < VerticalSplitNumber; i++)
      {
        this.thermalConductivity[i] = thermalConductivity;
        this.vSpecificHeat[i] = specificHeat;
      }

      //配列初期化
      hCap = new double[RadialSplitNumber, VerticalSplitNumber];
      lenZ = new double[VerticalSplitNumber];

      double bf = groundDepth / VerticalSplitNumber;
      for (int i = 0; i < VerticalSplitNumber; i++) lenZ[i] = bf;
      lenR = new double[RadialSplitNumber];
      bf = (groundRadius - PileRadius) / RadialSplitNumber;
      for (int i = 0; i < RadialSplitNumber; i++) lenR[i] = bf;

      /*double bf = groundDepth / (VerticalSplitNumber - 1);
      lenZ[0] = lenZ[VerticalSplitNumber - 1] = bf * 0.5;
      for (int i = 1; i < VerticalSplitNumber - 1; i++) lenZ[i] = bf;
      lenR = new double[RadialSplitNumber];
      bf = (groundRadius - PileRadius) / (RadialSplitNumber - 1);
      lenR[0] = lenR[RadialSplitNumber - 1] = bf * 0.5;
      for (int i = 1; i < RadialSplitNumber - 1; i++) lenR[i] = bf;*/

      deltaR = new double[RadialSplitNumber - 1];
      deltaZ = new double[VerticalSplitNumber - 1];
      radius = new double[RadialSplitNumber];
      heatFlow = new double[VerticalSplitNumber];

      int size = RadialSplitNumber * VerticalSplitNumber;
      tempA = new Vector(size);
      tempB = new Vector(size);
      tempBack = new Vector(size);
      mat = new SparseMatrix(size, size);
      temperature = new double[RadialSplitNumber, VerticalSplitNumber];
      for (int i = 0; i < temperature.GetLength(0); i++)
        for (int j = 0; j < temperature.GetLength(1); j++)
          temperature[i, j] = 20;
    }

    #endregion

    #region privateメソッド

    /// <summary>初期化する</summary>
    private void initialize()
    {
      if (!needInitialize) return;

      //質点間距離と半径を初期化
      int lst = deltaR.Length - 1;
      deltaR[0] = lenR[0] + 0.5 * lenR[1];
      deltaR[lst] = 0.5 * lenR[lst] + lenR[lst + 1];
      for (int i = 1; i < lst; i++) deltaR[i] = 0.5 * (lenR[i] + lenR[i + 1]);
      lst = deltaZ.Length - 1;
      deltaZ[0] = lenZ[0] + 0.5 * lenZ[1];
      deltaZ[lst] = 0.5 * lenZ[lst] + lenZ[lst + 1];
      for (int i = 1; i < lst; i++) deltaZ[i] = 0.5 * (lenZ[i] + lenZ[i + 1]);
      radius[0] = PileRadius;
      for (int i = 1; i < RadialSplitNumber; i++) radius[i] = radius[i - 1] + deltaR[i - 1];
      if (radius[0] <= 0.5 * deltaR[0]) throw new Exception("r_0 < 0.5 * Delta r_0");

      //熱拡散率を初期化
      double[] am = new double[VerticalSplitNumber];
      double[] aUm = new double[VerticalSplitNumber];
      double[] aDm = new double[VerticalSplitNumber];
      lst = VerticalSplitNumber - 1;
      for (int i = 0; i < lst; i++)
      {
        am[i] = thermalConductivity[i] / vSpecificHeat[i];
        double lamD = 2 * thermalConductivity[i] * thermalConductivity[i + 1] * lenZ[i] * lenZ[i + 1]
          / (thermalConductivity[i + 1] * (lenZ[i] - lenZ[i + 1]) + thermalConductivity[i] * lenZ[i + 1]);
        aDm[i] = lamD / vSpecificHeat[i];
        aUm[i + 1] = lamD / vSpecificHeat[i + 1];
      }
      aUm[0] = aDm[0];
      am[lst] = thermalConductivity[lst] / vSpecificHeat[lst];
      aDm[lst] = aUm[lst];

      //係数行列を作成
      mat.Initialize(0);
      double dt = TimeStep;
      int n1 = RadialSplitNumber - 2;
      int m1 = VerticalSplitNumber - 2;

      //地表面端部(m=0, n=0)
      int col = 0;
      mat[col, col + 1] = -2 * am[0] * dt / (deltaR[0] * deltaR[0]) / lenR[0];
      mat[col, col + RadialSplitNumber] = -2 * aDm[0] * dt / (deltaZ[0] * deltaZ[0]) / lenZ[0];
      mat[col, col] = 1 + 2 * am[0] * dt / (deltaR[0] * deltaR[0]) / lenR[0]
        + 2 * dt / deltaZ[0] / lenZ[0] * (HeatTransferCoefficient / vSpecificHeat[0] + aDm[0] / deltaZ[0]);

      //地表面端部(m=0, n=N)
      col = RadialSplitNumber - 1;
      mat[col, col - 1] = -2 * am[0] * dt / (deltaR[n1] * deltaR[n1]) / lenR[RadialSplitNumber - 1];
      mat[col, col + RadialSplitNumber] = -2 * aDm[0] * dt / (deltaZ[0] * deltaZ[0]) / lenZ[0];
      mat[col, col] = 1 + 2 * am[0] * dt / (deltaR[n1] * deltaR[n1]) / lenR[RadialSplitNumber - 1]
        + 2 * dt / deltaZ[0] / lenZ[0] * (HeatTransferCoefficient / vSpecificHeat[0] + aDm[0] / deltaZ[0]);

      //地表面(m=0, n!=0, n!=N)
      for (int n = 1; n < RadialSplitNumber - 1; n++)
      {
        col = n;
        double drav = 0.5 * (deltaR[n - 1] + deltaR[n]);
        mat[col, col - 1] = -am[0] * dt / drav * (1 / deltaR[n - 1] - 0.5 / radius[n]) / lenR[n];
        mat[col, col + 1] = -am[0] * dt / drav * (1 / deltaR[n] + 0.5 / radius[n]) / lenR[n];
        mat[col, col + RadialSplitNumber] = -2 * aDm[0] * dt / (deltaZ[0] * deltaZ[0]) / lenZ[0];
        mat[col, col] = 1 + 2 * am[0] * dt / (deltaR[n - 1] * deltaR[n]) / lenR[n]
          + 2 * dt / deltaZ[0] / lenZ[0] * (HeatTransferCoefficient / vSpecificHeat[0] + aDm[0] / deltaZ[0]);
      }

      //最下部端部(m=M, n=0)
      col = RadialSplitNumber * (VerticalSplitNumber - 1);
      mat[col, col + 1] = -2 * am[m1] * dt / (deltaR[0] * deltaR[0]) / lenR[0];
      mat[col, col - RadialSplitNumber] = -2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];
      mat[col, col] = 1 + 2 * am[m1] * dt / (deltaR[0] * deltaR[0]) / lenR[0]
        + 2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];

      //最下部端部(m=M, n=N)
      col = RadialSplitNumber * VerticalSplitNumber - 1;
      mat[col, col - 1] = -2 * am[m1] * dt / (deltaR[n1] * deltaR[n1]) / lenR[RadialSplitNumber - 1];
      mat[col, col - RadialSplitNumber] = -2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];
      mat[col, col] = 1 + 2 * am[m1] * dt / (deltaR[n1] * deltaR[n1]) / lenR[RadialSplitNumber - 1]
        + 2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];

      //最下部(m=M, n!=0, n!=N)
      for (int n = 1; n < RadialSplitNumber - 1; n++)
      {
        col = n + RadialSplitNumber * (VerticalSplitNumber - 1);
        double drav = 0.5 * (deltaR[n - 1] + deltaR[n]);
        mat[col, col - 1] = -am[m1] * dt / drav * (1 / deltaR[n - 1] - 0.5 / radius[n]) / lenR[n];
        mat[col, col + 1] = -am[m1] * dt / drav * (1 / deltaR[n] + 0.5 / radius[n]) / lenR[n];
        mat[col, col - RadialSplitNumber] = -2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];
        mat[col, col] = 1 + 2 * am[m1] * dt / (deltaR[n - 1] * deltaR[n]) / lenR[n]
          + 2 * aUm[m1] * dt / (deltaZ[m1] * deltaZ[m1]) / lenZ[VerticalSplitNumber - 1];
      }

      //円周部(n=N, m!=0, m!=M)
      for (int m = 1; m < VerticalSplitNumber - 1; m++)
      {
        col = RadialSplitNumber * (m + 1) - 1;
        double dzav = 0.5 * (deltaZ[m - 1] + deltaZ[m]);
        mat[col, col - 1] = -2 * am[m] * dt / (deltaR[n1] * deltaR[n1]) / lenR[RadialSplitNumber - 1];
        mat[col, col - RadialSplitNumber] = -aUm[m] * dt / (deltaZ[m - 1] * dzav) / lenZ[m];
        mat[col, col + RadialSplitNumber] = -aDm[m] * dt / (deltaZ[m] * dzav) / lenZ[m];
        mat[col, col] = 1 + 2 * am[m] * dt / (deltaR[n1] * deltaR[n1])/ lenR[RadialSplitNumber - 1]
          + (aUm[m] + aDm[m]) * dt / (deltaZ[m - 1] * deltaZ[m]) / lenZ[m];
      }

      //地中熱交換器隣接部(n=0, m!=0, m!=M)
      for (int m = 1; m < VerticalSplitNumber - 1; m++)
      {
        col = RadialSplitNumber * m;
        double dzav = 0.5 * (deltaZ[m - 1] + deltaZ[m]);
        mat[col, col + 1] = -2 * am[m] * dt / (deltaR[0] * deltaR[0]) / lenR[0];
        mat[col, col - RadialSplitNumber] = -aUm[m] * dt / (deltaZ[m - 1] * dzav) / lenZ[m];
        mat[col, col + RadialSplitNumber] = -aDm[m] * dt / (deltaZ[m] * dzav) / lenZ[m];
        mat[col, col] = 1 + 2 * am[m] * dt / (deltaR[0] * deltaR[0]) / lenR[0]
          + (aUm[m] + aDm[m]) * dt / (deltaZ[m - 1] * deltaZ[m]) / lenZ[m];
      }

      //土壌内部(n!=0, n!=N, m!=0, m!=M)
      for (int m = 1; m < VerticalSplitNumber - 1; m++)
      {
        for (int n = 1; n < RadialSplitNumber - 1; n++)
        {
          col = RadialSplitNumber * m + n;
          double drav = 0.5 * (deltaR[n - 1] + deltaR[n]);
          double dzav = 0.5 * (deltaZ[m - 1] + deltaZ[m]);
          mat[col, col - 1] = -am[m] * dt / drav * (1 / deltaR[n - 1] - 0.5 / radius[n]) / lenR[n];
          mat[col, col + 1] = -am[m] * dt / drav * (1 / deltaR[n] + 0.5 / radius[n]) / lenR[n];
          mat[col, col - RadialSplitNumber] = -aUm[m] * dt / (deltaZ[m - 1] * dzav) / lenZ[m];
          mat[col, col + RadialSplitNumber] = -aDm[m] * dt / (deltaZ[m] * dzav) / lenZ[m];
          mat[col, col] = 1 + 2 * am[m] * dt / (deltaR[n - 1] * deltaR[n]) / lenR[n]
            + (aUm[m] + aDm[m]) * dt / (deltaZ[m - 1] * deltaZ[m]) / lenZ[m];
        }
      }

      needInitialize = false;
    }

    /// <summary>温度ベクトルを二次元配列に設定する</summary>
    /// <param name="vec">温度ベクトル</param>
    /// <param name="temps">二次元配列</param>
    private static void setTemperatures(IVector vec, ref double[,] temps)
    {
      int RNum = temps.GetLength(0);
      int VNum = temps.GetLength(1);
      for (int n = 0; n < RNum; n++)
        for (int m = 0; m < VNum; m++)
          temps[n, m] = vec[n + RNum * m];
    }

    /// <summary>二次元配列を温度ベクトルに設定する</summary>
    /// <param name="temps">二次元配列</param>
    /// <param name="vec">温度ベクトル</param>
    private static void setTemperatures(double[,] temps, ref IVector vec)
    {
      int RNum = temps.GetLength(0);
      int VNum = temps.GetLength(1);
      for (int n = 0; n < RNum; n++)
        for (int m = 0; m < VNum; m++)
          vec[n + RNum * m] = temps[n, m];
    }

    #endregion

    #region モデル作成処理

    /// <summary>質点の深さ[m]を設定する</summary>
    /// <param name="vSplitNumber">質点の番号（深さ方向）</param>
    /// <param name="depth">深さ[m]</param>
    public void SetDepth(int vSplitNumber, double depth)
    {
      needInitialize = true;
      lenZ[vSplitNumber] = depth;
    }

    /// <summary>質点の幅[m]を設定する</summary>
    /// <param name="rSplitNumber">質点の番号（半径方向）</param>
    /// <param name="length">深さ[m]</param>
    public void SetLength(int rSplitNumber, double length)
    {
      needInitialize = true;
      lenR[rSplitNumber] = length;
    }

    /// <summary>地中熱交換器からの熱流[W/m2]を設定する</summary>
    /// <param name="vSplitNumber">深さ方向の番号</param>
    /// <param name="heatFlow">熱流[W/m2]</param>
    public void SetHeatFlow(int vSplitNumber, double heatFlow)
    { this.heatFlow[vSplitNumber] = heatFlow; }

    #endregion

    #region 状態更新・取得処理

    /// <summary>状態を予測する</summary>
    public void Update()
    {
      initialize();

      //係数ベクトルを作成
      int n1 = RadialSplitNumber - 2;
      int m1 = VerticalSplitNumber - 2;
      setTemperatures(temperature, ref tempA);

      //地表面(m=0)
      double bf = 2 * HeatTransferCoefficient * TimeStep / (vSpecificHeat[0] * deltaZ[0]) * SolAirTemperature / lenZ[0];
      for (int n = 0; n < RadialSplitNumber; n++) tempA[n] += bf;

      //地中熱交換器隣接部(n=0)
      double depth = 0;
      bf = (2 * TimeStep) * (1 / deltaR[0] - 0.5 / radius[0]) / lenR[0];
      for (int m = 0; m < VerticalSplitNumber; m++)
      {
        if (PileDepth <= depth && depth <= PileDepth + PileLength)
          tempA[RadialSplitNumber * m] += bf * heatFlow[m] / vSpecificHeat[m];
        depth += lenZ[m];
      }

      /*//deBUG
      for (int i = 0; i < mat.Columns; i++)
      {
        double sum = 0;
        for (int j = 0; j < mat.Rows; j++) sum += mat[i, j];
        Console.WriteLine(sum);
      }*/

      //疎行列による連立方程式を解く
      mat.SolveLinearEquation(tempA, ref tempB);      
      setTemperatures(tempB, ref temperature);
    }

    /// <summary>地中温度[C]を取得する</summary>
    /// <param name="n">半径方向分割番号</param>
    /// <param name="m">深さ方向分割番号</param>
    /// <returns>地中温度[C]</returns>
    public double GetTemperature(int n, int m) { return temperature[n, m]; }

    /// <summary>土壌の熱量[MJ]を積算する</summary>
    /// <param name="refTemperature">基準温度[C]</param>
    /// <returns>土壌の熱量[MJ]</returns>
    public double GetHeatStorage(double refTemperature)
    {
      initialize();
      double sum = 0;
      for (int i = 0; i < RadialSplitNumber; i++)
      {
        double bf = 0;
        for (int j = 0; j < VerticalSplitNumber; j++)
          bf += (temperature[i, j] - refTemperature) * lenZ[j] * vSpecificHeat[j];
        sum += bf * radius[i] * lenR[i];
      }
      return sum * 2 * Math.PI * 1e-6;
    }

    /// <summary>地中温度を初期化する</summary>
    /// <param name="temp">地中温度[C]</param>
    public void InitializeTemperature(double temp)
    {
      for (int i = 0; i < temperature.GetLength(0); i++)
        for (int j = 0; j < temperature.GetLength(1); j++)
          temperature[i, j] = temp;
    }

    #endregion

  }
}
