﻿using System;
using System.Collections.Generic;

using Popolo.Numerics;

namespace Popolo.HumanBody
{
  /// <summary>事務所執務者</summary>
  public class OfficeWorker : ImmutableOfficeWorker
  {

    #region 列挙型定義

    /// <summary>曜日</summary>
    [Flags]
    public enum DaysOfWeek
    {
      /// <summary>日曜</summary>
      Sunday = 1,
      /// <summary>月曜</summary>
      Monday = 2,
      /// <summary>火曜</summary>
      Tuesday = 4,
      /// <summary>水曜</summary>
      Wednesday = 8,
      /// <summary>木曜</summary>
      Thursday = 16,
      /// <summary>金曜</summary>
      Friday = 32,
      /// <summary>土曜</summary>
      Saturday = 64
    }

    /// <summary>雇用形態</summary>
    public enum EmploymentStatus
    {
      /// <summary>派遣・契約社員</summary>
      TemporaryStaff,
      /// <summary>一般平社員</summary>
      NoTitle,
      /// <summary>管理職</summary>
      Manager,
      /// <summary>役員</summary>
      Administrator
    }

    /// <summary>昼休み</summary>
    private enum LunchBreakTake
    {
      /// <summary>常に社外</summary>
      AlwaysOutGo = 0,
      /// <summary>常に社内</summary>
      NeverOutGo = 1,
      /// <summary>時に社外</summary>
      SomeTimeOutGo = 2
    }

    #endregion

    #region クラス変数

    /// <summary>出社モデルのパラメータ（平日）</summary>
    private static readonly double[] sttMdl = new double[] { -40.23, 32.61, 1.101, -0.452 };

    /// <summary>出社モデルのパラメータ（休日）</summary>
    private static readonly double[] sttMdl_Hol = new double[] { 24.45, 98.71, 1.273, -0.060 };

    /// <summary>退社モデルのパラメータ（平日:M20-M40）</summary>
    private static readonly double[] endMdl_M24 = new double[] { 82.3, 109.28, 1.529, 0.438 };

    /// <summary>退社モデルのパラメータ（平日:その他）</summary>
    private static readonly double[] endMdl = new double[] { 40.86, 72.69, 1.309, 0.286 };

    /// <summary>退社モデル（滞在時間）のパラメータ（休日）</summary>
    private static readonly double[] endMdl_Hol = new double[] { -73.75, 203.35, 1.465, 0.683 };

    /// <summary>昼休み開始モデルのパラメータ</summary>
    private static readonly double[] lncStMdl = new double[] { -0.337, 1.881, 0.502, 0.303 };

    /// <summary>昼休み終了モデルのパラメータ</summary>
    private static readonly double[] lncEdMdl = new double[] { -5.661, 74.93, 0.686, 0.905 };

    /// <summary>社内→社内の状態推移確率</summary>
    private static readonly double[] tProbII = new double[] { 0.973, 0.979, 0.986, 0.952, 0.960, 0.991, 0.964, 0.973, 0.989, 0.976 };

    /// <summary>社外→社外の状態推移確率</summary>
    private static readonly double[] tProbOO = new double[] { 0.880, 0.773, 0.850, 0.765, 0.816, 0.905, 0.556, 0.417, 0.556, 0.710 };

    /// <summary>業種による補正係数（技術職では0.15としておく）</summary>
    private const double cFactor = 0.00;

    #endregion

    #region インスタンス変数

    /// <summary>一様乱数発生器</summary>
    private readonly MersenneTwister urGen;

    /// <summary>正規乱数発生器</summary>
    private readonly NormalRandom nrGen;

    /// <summary>執務者グループ:M20,M30,M40,M50,M60,F20,F30,F40,F50,F60</summary>
    private readonly int group;

    /// <summary>徹夜作業を行う可能性のある執務者か否か</summary>
    private bool allNightWorkable;

    /// <summary>休日出勤を行う可能性のある執務者か否か</summary>
    private bool holidayWorkable;

    /// <summary>昼休みの取り方</summary>
    private readonly LunchBreakTake lunchBreak;

    /// <summary>徹夜作業中か</summary>
    private bool allNightWorker = false;

    /// <summary>休日出勤中か</summary>
    private bool holidayWorker = false;

    /// <summary>昼休み外出時刻</summary>
    private DateTime lnchOutGoTime;

    /// <summary>昼休み帰社時刻</summary>
    private DateTime lnchComeBackTime;

    /// <summary>外出開始時刻リスト</summary>
    private List<DateTime> outGoTime = new List<DateTime>();

    /// <summary>帰社時刻リスト</summary>
    private List<DateTime> comeBackTime = new List<DateTime>();

    #endregion

    #region プロパティ定義

    /// <summary>男性か否かを取得する</summary>
    public bool IsMale { get; private set; }

    /// <summary>雇用形態を取得する</summary>
    public EmploymentStatus Status { get; private set; }

    /// <summary>年齢を取得する</summary>
    public int Age { get; private set; }

    /// <summary>休日を取得する</summary>
    public DaysOfWeek Holidays { get; private set; }

    /// <summary>始業時刻（時）を取得する</summary>
    public int StartHour { get; private set; }

    /// <summary>始業時刻（分）を取得する</summary>
    public int StartMinute { get; private set; }

    /// <summary>終業時刻（時）を取得する</summary>
    public int EndHour { get; private set; }

    /// <summary>終業時刻（分）を取得する</summary>
    public int EndMinute { get; private set; }

    /// <summary>昼休み開始時刻（時）を取得する</summary>
    public int LunchStartHour { get; private set; }

    /// <summary>昼休み開始時刻（分）を取得する</summary>
    public int LunchStartMinute { get; private set; }

    /// <summary>昼休み終了時刻（時）を取得する</summary>
    public int LunchEndHour { get; private set; }

    /// <summary>昼休み終了時刻（分）を取得する</summary>
    public int LunchEndMinute { get; private set; }

    /// <summary>出社時刻を取得する</summary>
    public DateTime ArriveTime { get; private set; }

    /// <summary>退社時刻を取得する</summary>
    public DateTime LeaveTime { get; private set; }

    /// <summary>休暇取得中か否かを取得する</summary>
    public bool IsLOA { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="isMale">男性か否か</param>
    /// <param name="age">年齢</param>
    /// <param name="isPermanent">正社員か否か</param>
    /// <param name="holidays">休日の曜日（複数設定可）</param>
    /// <param name="startHour">始業時刻（時）</param>
    /// <param name="startMinute">始業時刻（分）</param>
    /// <param name="endHour">終業時刻（時）</param>
    /// <param name="endMinute">就業時刻（分）</param>
    /// <param name="lunchStartHour">昼休み開始時刻（時）</param>
    /// <param name="lunchStartMinute">昼休み開始時刻（分）</param>
    /// <param name="lunchEndHour">昼休み終了時刻（時）</param>
    /// <param name="lunchEndMinute">昼休み終了時刻（分）</param>
    public OfficeWorker(bool isMale, int age, bool isPermanent, DaysOfWeek holidays,
      int startHour, int startMinute, int endHour, int endMinute,
      int lunchStartHour, int lunchStartMinute, int lunchEndHour, int lunchEndMinute) :
      this(isMale, age, isPermanent, holidays, startHour, startMinute, endHour, endMinute,
        lunchStartHour, lunchStartMinute, lunchEndHour, lunchEndMinute, System.DateTime.Now.Millisecond)
    { }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="isMale">男性か否か</param>
    /// <param name="age">年齢</param>
    /// <param name="isPermanent">正社員か否か</param>
    /// <param name="holidays">休日の曜日（複数設定可）</param>
    /// <param name="startHour">始業時刻（時）</param>
    /// <param name="startMinute">始業時刻（分）</param>
    /// <param name="endHour">終業時刻（時）</param>
    /// <param name="endMinute">就業時刻（分）</param>
    /// <param name="lunchStartHour">昼休み開始時刻（時）</param>
    /// <param name="lunchStartMinute">昼休み開始時刻（分）</param>
    /// <param name="lunchEndHour">昼休み終了時刻（時）</param>
    /// <param name="lunchEndMinute">昼休み終了時刻（分）</param>
    /// <param name="seed">乱数発生シード</param>
    public OfficeWorker(bool isMale, int age, bool isPermanent, DaysOfWeek holidays,
      int startHour, int startMinute, int endHour, int endMinute,
      int lunchStartHour, int lunchStartMinute, int lunchEndHour, int lunchEndMinute, int seed)
    {
      IsMale = isMale;
      Age = age;
      Holidays = holidays;
      StartHour = startHour;
      StartMinute = startMinute;
      EndHour = endHour;
      EndMinute = endMinute;
      LunchStartHour = lunchStartHour;
      LunchEndHour = lunchEndHour;
      LunchEndMinute = lunchEndMinute;
      urGen = new MersenneTwister(seed);
      nrGen = new NormalRandom(seed);

      //世代と年齢によるグループ判定
      if (IsMale)
      {
        if (age < 30) group = 0;
        else if (30 <= age && age < 40) group = 1;
        else if (40 <= age && age < 50) group = 2;
        else if (50 <= age && age < 60) group = 3;
        else group = 4;
      }
      else
      {
        if (age < 30) group = 5;
        else if (30 <= age && age < 40) group = 6;
        else if (40 <= age && age < 50) group = 7;
        else if (50 <= age && age < 60) group = 8;
        else group = 9;
      }

      //雇用形態判定
      if (!isPermanent) Status = EmploymentStatus.TemporaryStaff;
      else
      {
        double[] mngRate = new double[] { 4 / 80d, 9 / 80d, 23 / 81d, 31 / 83d, 18 / 59d, 1 / 72d, 1 / 78d, 4 / 79d, 6 / 67d, 7 / 67d };  //管理職比率
        double[] admRate = new double[] { 1 / 80d, 0 / 80d, 4 / 81d, 11 / 83d, 22 / 59d, 1 / 72d, 0 / 78d, 0 / 79d, 4 / 67d, 13 / 67d };  //役員比率
        double rnd = nrGen.NextDouble();
        if (mngRate[group] < rnd) Status = EmploymentStatus.Manager;
        else if (mngRate[group] + admRate[group] < rnd) Status = EmploymentStatus.Administrator;
        else Status = EmploymentStatus.NoTitle;
      }

      //徹夜作業と休日出勤可能性判定
      switch (Status)
      {
        case EmploymentStatus.NoTitle:
          allNightWorkable = urGen.NextDouble() < 0.093;
          holidayWorkable = urGen.NextDouble() < 0.355;
          break;
        case EmploymentStatus.Manager:
          allNightWorkable = urGen.NextDouble() < 0.215;
          holidayWorkable = urGen.NextDouble() < 0.527;
          break;
        case EmploymentStatus.Administrator:
          allNightWorkable = urGen.NextDouble() < 0.088;
          holidayWorkable = urGen.NextDouble() < 0.471;
          break;
        default:
          allNightWorkable = urGen.NextDouble() < 0.045;
          holidayWorkable = urGen.NextDouble() < 0.306;
          break;
      }

      //昼休みの取り方
      double[] aGoOut = new double[] { 0.176, 0.157, 0.179, 0.122, 0.221, 0.068, 0.070, 0.141, 0.179, 0.155 };  //常に外出
      double[] nGoOut = new double[] { 0.392, 0.446, 0.488, 0.451, 0.453, 0.534, 0.581, 0.500, 0.488, 0.452 };  //常に社内
      double bf = urGen.NextDouble();
      if (bf < nGoOut[group]) lunchBreak = LunchBreakTake.NeverOutGo;
      else if (bf < nGoOut[group] + aGoOut[group]) lunchBreak = LunchBreakTake.AlwaysOutGo;
      else lunchBreak = LunchBreakTake.SomeTimeOutGo;
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>休日か否か</summary>
    /// <param name="dayOfWeek">曜日</param>
    /// <returns>休日か否か</returns>
    private bool isHoliday(DayOfWeek dayOfWeek)
    {
      switch (dayOfWeek)
      {
        case DayOfWeek.Saturday:
          return (Holidays & DaysOfWeek.Saturday) == DaysOfWeek.Saturday;
        case DayOfWeek.Monday:
          return (Holidays & DaysOfWeek.Monday) == DaysOfWeek.Monday;
        case DayOfWeek.Tuesday:
          return (Holidays & DaysOfWeek.Tuesday) == DaysOfWeek.Tuesday;
        case DayOfWeek.Wednesday:
          return (Holidays & DaysOfWeek.Wednesday) == DaysOfWeek.Wednesday;
        case DayOfWeek.Thursday:
          return (Holidays & DaysOfWeek.Thursday) == DaysOfWeek.Thursday;
        case DayOfWeek.Friday:
          return (Holidays & DaysOfWeek.Friday) == DaysOfWeek.Friday;
        default:
          return (Holidays & DaysOfWeek.Sunday) == DaysOfWeek.Sunday;
      }
    }

    /// <summary>1日の執務スケジュールを更新する</summary>
    /// <param name="dTime">日にち</param>
    public void Update(DateTime dTime)
    {
      //休日出勤判定
      IsLOA = isHoliday(dTime.DayOfWeek);
      if (!IsLOA && urGen.NextDouble() < 0.035)
      {
        IsLOA = true;
        holidayWorker = false;
      }
      else holidayWorker = IsLOA && holidayWorkable && urGen.NextDouble() < 0.188;
      //休日出勤者の34.1%は平日相当の業務時間
      if (holidayWorker && 0.341 < urGen.NextDouble()) IsLOA = false;

      //出社時刻計算****************************************
      //前日徹夜作業の場合には出社時刻=0:00,翌日は非休日
      bool spendNight = allNightWorker; //一次保存
      if (allNightWorker)
      {
        ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0);
        IsLOA = false;
      }
      //休日の場合
      else if (IsLOA && !holidayWorker)
        ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0).AddDays(1);
      //平日出勤・休日出勤の場合
      else
      {
        double[] par;
        if (!IsLOA) par = sttMdl; //平日出勤
        else par = sttMdl_Hol;    //休日出勤
        double min = par[0] + par[1] * Math.Sinh((nrGen.NextDouble() - par[3]) / par[2]);
        ArriveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, StartHour, StartMinute, 0).AddMinutes(min);
      }

      //退社時刻計算****************************************
      //当日徹夜作業の場合には退社時刻=0:00
      allNightWorker = (urGen.NextDouble() < 0.012 && allNightWorkable);
      if (allNightWorker) LeaveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0).AddDays(1);
      //休日の場合
      else if (IsLOA && !holidayWorker) LeaveTime = ArriveTime;
      else
      {
        //平日出勤
        if (!IsLOA)
        {
          double[] par;
          if (group == 0 || group == 1 || group == 2) par = endMdl_M24;
          else par = endMdl;
          double min = par[0] + par[1] * Math.Sinh((nrGen.NextDouble() - par[3]) / par[2]);
          LeaveTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, EndHour, EndMinute, 0).AddMinutes(min);
        }
        //休日出勤
        else
        {
          double min = endMdl_Hol[0] + endMdl_Hol[1] * Math.Sinh((nrGen.NextDouble() - endMdl_Hol[3]) / endMdl_Hol[2]);
          min += 60 * (EndHour - StartHour) + (EndMinute - StartMinute);
          LeaveTime = ArriveTime.AddMinutes(min);
        }
        if (LeaveTime.CompareTo(ArriveTime) < 0) LeaveTime = ArriveTime;
      }

      //昼休み**********************************************
      //外出する場合
      if (lunchBreak == LunchBreakTake.AlwaysOutGo || urGen.NextDouble() < 0.488)
      {
        double min = lncStMdl[0] + lncStMdl[1] * Math.Sinh((nrGen.NextDouble() - lncStMdl[3]) / lncStMdl[2]);
        lnchOutGoTime = new DateTime(dTime.Year, dTime.Month, dTime.Day, LunchStartHour, LunchStartMinute, 0).AddMinutes(min);

        double z = (nrGen.NextDouble() - lncEdMdl[3]) / lncEdMdl[2];
        min = (lncEdMdl[0] + lncEdMdl[1] * Math.Exp(z)) / (Math.Exp(z) + 1);
        min = Math.Max(0, 60 * (LunchEndHour - LunchStartHour) + (LunchEndMinute - LunchStartMinute) - min);
        lnchComeBackTime = lnchOutGoTime.AddMinutes(min);
      }
      //社内に留まる場合
      else lnchOutGoTime = lnchComeBackTime = ArriveTime;

      //その他一時外出**************************************
      outGoTime.Clear();
      comeBackTime.Clear();
      DateTime dtNow;
      if (spendNight) dtNow = new DateTime(dTime.Year, dTime.Month, dTime.Day, StartHour, StartMinute, 0); //前日泊まった執務者は定時より外出し得る
      else dtNow = ArriveTime;
      bool goOut = urGen.NextDouble() < ((1.0 - (tProbII[group] - cFactor)) / (2.0 - tProbOO[group] - (tProbII[group] - cFactor)));
      if (goOut) outGoTime.Add(dtNow);
      dtNow = dtNow.AddHours(1);
      while (dtNow.CompareTo(LeaveTime) < 0)
      {
        if (goOut && (urGen.NextDouble() < 1.0 - tProbOO[group]))
        {
          goOut = false;
          comeBackTime.Add(dtNow);
        }
        else if (!goOut && (urGen.NextDouble() < 1.0 - (tProbII[group] - cFactor)))
        {
          goOut = true;
          outGoTime.Add(dtNow);
        }
        dtNow = dtNow.AddHours(1);
      }
      if (goOut) comeBackTime.Add(LeaveTime);
    }

    /// <summary>社内に滞在しているか否か</summary>
    /// <param name="dTime">時刻</param>
    /// <returns>社内に滞在しているか否か</returns>
    public bool StayInOffice(DateTime dTime)
    {
      //出社退社時間外の場合は不在
      if (dTime.CompareTo(ArriveTime) < 0) return false;
      if (0 < dTime.CompareTo(LeaveTime)) return false;

      //昼休み滞在不在確認
      if (0 < dTime.CompareTo(lnchOutGoTime) && dTime.CompareTo(lnchComeBackTime) <= 0) return false;

      //その他一時外出確認
      for (int i = 0; i < outGoTime.Count; i++)
        if (0 < dTime.CompareTo(outGoTime[i]) && dTime.CompareTo(comeBackTime[i]) <= 0) return false;

      return true;
    }

    /// <summary>就業形式を設定する</summary>
    /// <param name="allNightWorkable">徹夜作業をし得るか否か</param>
    /// <param name="holidayWorkable">休日出勤をし得るか否か</param>
    public void SetWorkStyle(bool allNightWorkable, bool holidayWorkable)
    {
      this.allNightWorkable = allNightWorkable;
      this.holidayWorkable = holidayWorkable;
    }

    #endregion

  }

  #region 読み取り専用interface

  /// <summary>読み取り専用の事務所執務者</summary>
  public interface ImmutableOfficeWorker
  {

    /// <summary>男性か否かを取得する</summary>
    bool IsMale { get; }

    /// <summary>雇用形態を取得する</summary>
    OfficeWorker.EmploymentStatus Status { get; }

    /// <summary>年齢を取得する</summary>
    int Age { get; }

    /// <summary>休日を取得する</summary>
    OfficeWorker.DaysOfWeek Holidays { get; }

    /// <summary>始業時刻（時）を取得する</summary>
    int StartHour { get; }

    /// <summary>始業時刻（分）を取得する</summary>
    int StartMinute { get; }

    /// <summary>終業時刻（時）を取得する</summary>
    int EndHour { get; }

    /// <summary>終業時刻（分）を取得する</summary>
    int EndMinute { get; }

    /// <summary>昼休み開始時刻（時）を取得する</summary>
    int LunchStartHour { get; }

    /// <summary>昼休み開始時刻（分）を取得する</summary>
    int LunchStartMinute { get; }

    /// <summary>昼休み終了時刻（時）を取得する</summary>
    int LunchEndHour { get; }

    /// <summary>昼休み終了時刻（分）を取得する</summary>
    int LunchEndMinute { get; }

    /// <summary>出社時刻を取得する</summary>
    DateTime ArriveTime { get; }

    /// <summary>退社時刻を取得する</summary>
    DateTime LeaveTime { get; }

    /// <summary>休暇取得中か否かを取得する</summary>
    bool IsLOA { get; }

    /// <summary>社内に滞在しているか否か</summary>
    /// <param name="dTime">時刻</param>
    /// <returns>社内に滞在しているか否か</returns>
    bool StayInOffice(DateTime dTime);

  }

  #endregion

}
