﻿/* TwoNodeModel.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HumanBody
{
  /// <summary>2-Nodeモデルに関する静的メソッドを提供する</summary>
  public static class TwoNodeModel
  {

    #region 定数宣言

    /// <summary>黒体の放射定数[W/(m2K4)]</summary>
    private const double BLACK_CONSTANT = 5.67e-8;

    /// <summary>衣服の面積率[-]</summary>
    private const double K_CLO = 0.25;

    /// <summary>衣服の透湿係数[K/kPa]</summary>
    private const double I_CLS = 0.45;

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>代謝量換算係数[(W/m2)/met]</summary>
    private const double CONVERT_MET_TO_W = 58.2;

    #endregion

    #region staticメソッド

    /// <summary>定常状態を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[CDB]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="velocity">相対気流速度[m/s]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="basalMetabolism">基礎代謝量[W/m2]</param>
    /// <param name="externalWork">外部仕事量[W/m2]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <param name="skinTemperature">出力:皮膚温度[C]</param>
    /// <param name="coreTemperature">出力:コア温度[C]</param>
    /// <param name="bodyTemperature">出力:平均体温[C]</param>
    /// <param name="clothTemperature">出力:衣服温度[C]</param>
    /// <param name="sensibleHFSkin">出力:皮膚からの顕熱損失[W/m2]</param>
    /// <param name="latentHFSkin">出力:皮膚からの潜熱損失[W/m2]</param>
    /// <param name="sensibleRespiration">出力:呼吸による顕熱損失[W/m2]</param>
    /// <param name="latentRespiration">出力:呼吸による潜熱損失[W/m2]</param>
    /// <param name="wettedness">出力:皮膚の平均濡れ率[-]</param>
    public static void GetSteadyState
      (double drybulbTemperature, double meanRadiantTemperature, double relativeHumidity, double velocity, 
      double clothing, double basalMetabolism, double externalWork, double atmosphericPressure,
      out double skinTemperature, out double coreTemperature, out double bodyTemperature, 
      out double clothTemperature, out double sensibleHFSkin, out double latentHFSkin, 
      out double sensibleRespiration, out double latentRespiration, out double wettedness)
    {
      //定数宣言
      const double WEIGHT = 70d;              //標準体重[kg]
      const double BODY_SURFACE = 1.8;        //標準体表面積[m2]
      const double C_SWEATING = 47.2;         //発汗の係数[mg/(m2 s K)]
      const double C_VASODILATATION = 55.6;   //血管拡張の係数[L/(m2 s K)]
      const double C_VASOCONSTRICTION = 0.1;  //血管収縮の係数[1/K]
      const double SETPOINT_SKIN = 33.7;      //皮膚セットポイント温度[C]
      const double SETPOINT_CORE = 36.8;      //コアセットポイント温度[C]
      const double SETPOINT_BODY = 36.49;     //体温セットポイント温度[C]
      const double NORMAL_BLOOD_FLOW = 1.75;  //標準状態の皮膚血流量[mL/(m2 s)]
      const double CRITICAL_WETTEDNESS = 0.85;//最大濡れ率[-]

      //水蒸気分圧[kPa]の計算
      double pa = relativeHumidity / 100 * Water.GetSaturationPressure(drybulbTemperature);
      //着衣抵抗[m2K/W]の計算
      double rcl = 0.155 * clothing;
      //着衣面積率[-]の計算
      double clothRate = 1d + K_CLO * clothing;

      //初期値を設定する
      sensibleHFSkin = sensibleRespiration = latentRespiration = wettedness = 0;
      skinTemperature = SETPOINT_SKIN;
      coreTemperature = SETPOINT_CORE;
      bodyTemperature = SETPOINT_BODY;
      double skinBloodFlow = NORMAL_BLOOD_FLOW;
      double mshiv = 0;
      double alpha = 0.1;
      latentHFSkin = 0.1 * basalMetabolism;
      double metabolism = basalMetabolism;   //ふるえ産熱は0とする

      //対流熱伝達率の計算（代謝量か気流で決定）
      double chcv = 8.6 * Math.Pow(Math.Max(0.15, velocity), 0.53);
      double chcm = 5.66 * Math.Pow(Math.Max(0, basalMetabolism / CONVERT_MET_TO_W - 0.85), 0.39);
      double convectiveHTransCoef = Math.Max(chcv, chcm);

      //Δt=1minとして60minの繰り返し計算を行う
      const int DELTA_T = 1;
      clothTemperature = (skinTemperature + drybulbTemperature) / 2d;
      for (int tim = 0; tim < 60; tim += DELTA_T)
      {
        //衣服の表面温度を収束計算
        double operatingTemp, ra;
        while (true)
        {
          double ctOld = clothTemperature;
          //放射熱伝達率[W/(m2K)]の計算
          double hr = 4d * BLACK_CONSTANT * 0.72 
            * Math.Pow((clothTemperature + meanRadiantTemperature) / 2d + CONVERT_C_TO_K, 3);
          //総合熱伝達率[W/(m2K)]の計算
          double hcr = hr + convectiveHTransCoef;
          //空気層顕熱抵抗[(m2K)/W]の計算
          ra = 1 / (clothRate * hcr);
          //作用温度[C]の計算
          operatingTemp = (hr * meanRadiantTemperature + convectiveHTransCoef * drybulbTemperature) / hcr;
          //衣服温度[C]の計算
          clothTemperature = (ra * skinTemperature + rcl * operatingTemp) / (ra + rcl);
          //衣服温度の更新量が0.01C以下で収束と判定
          if (Math.Abs(ctOld - clothTemperature) < 0.01) break;
        }

        //皮膚からの顕熱損失量[W/(m2K)]の計算
        sensibleHFSkin = (skinTemperature - operatingTemp) / (ra + rcl);
        //コアから皮膚への熱流[W/m2]の計算
        double hfcs = (coreTemperature - skinTemperature) * (5.28 + 3.842 * skinBloodFlow);

        //呼吸による熱損失量[W/m2]の計算
        latentRespiration = metabolism * 0.017251 * (5.8662 - pa);
        sensibleRespiration = metabolism * 0.0014 * (34 - drybulbTemperature);

        //コアと皮膚への熱流[W/m2]を計算
        double scr = metabolism - hfcs - latentRespiration - sensibleRespiration - externalWork;
        double ssk = hfcs - sensibleHFSkin - latentHFSkin;
        //コアと皮膚の熱容量[J/K]を計算
        double tccr = 3492 * (1 - alpha) * WEIGHT;
        double tcsk = 3492 * alpha * WEIGHT;
        //コアと皮膚の温度変化量[K/s]を計算
        double dtcr = (scr * BODY_SURFACE) / tccr;
        double dtsk = (ssk * BODY_SURFACE) / tcsk;
        //コアと皮膚の温度を更新
        coreTemperature = coreTemperature + dtcr * (DELTA_T * 60);
        skinTemperature = skinTemperature + dtsk * (DELTA_T * 60);

        //重み付けして体温を計算
        bodyTemperature = alpha * skinTemperature + (1 - alpha) * coreTemperature;

        //制御量の計算
        double sDil = Math.Max(0, coreTemperature - SETPOINT_CORE);
        double sStr = Math.Max(0, SETPOINT_SKIN - skinTemperature);
        double sSw1 = Math.Max(0, bodyTemperature - SETPOINT_BODY);
        double sSw2 = Math.Max(0, skinTemperature - SETPOINT_SKIN);
        double sShv1 = Math.Max(0, SETPOINT_SKIN - skinTemperature);
        double sShv2 = Math.Max(0, SETPOINT_CORE - coreTemperature);

        //皮膚血流量[L/(m2 s)]の計算
        skinBloodFlow = (NORMAL_BLOOD_FLOW + C_VASODILATATION * sDil) / (1d + C_VASOCONSTRICTION * sStr);
        skinBloodFlow = Math.Max(Math.Min(skinBloodFlow, 25), 0.139);
        //皮膚・コアの重量比を更新
        alpha = 0.0417737 + 0.2069953 / (skinBloodFlow + 0.1626158);

        //調節発汗による蒸発熱損失[W/m2]を計算
        double msw = C_SWEATING * sSw1 * Math.Exp(sSw2 / 10.7);
        double esw = 2.501 * msw;

        //不感蒸泄による蒸発熱損失[W/m2]を計算
        double lewis = 0.0555 * (skinTemperature + CONVERT_C_TO_K);
        double latentHTransCoef = 1 / (rcl / (I_CLS * lewis) + 1d / (clothRate * convectiveHTransCoef * lewis));
        double emax = latentHTransCoef * (Water.GetSaturationPressure(skinTemperature) - pa);
        double wSw = esw / emax;
        double eb = 0.06 * (emax - esw);
        double wB = eb / emax;
        wettedness = wSw + wB;
        latentHFSkin = esw + eb;

        //ぬれ率上限を上回る場合には蒸発熱損失等を修正
        if (CRITICAL_WETTEDNESS < wettedness)
        {
          wettedness = CRITICAL_WETTEDNESS;
          wSw = (CRITICAL_WETTEDNESS - 0.06) / 0.94;
          esw = wSw * emax;
          eb = 0.06 * (1 - wSw) * emax;
          latentHFSkin = esw + eb;
        }
        //皮膚表面が露点以下の場合
        if (emax < 0)
        {
          eb = 0;
          esw = 0;
          wB = CRITICAL_WETTEDNESS;
          wSw = CRITICAL_WETTEDNESS;
          latentHFSkin = emax;
        }

        //ふるえ熱産生量[W/m2]を計算
        mshiv = 19.4 * sShv1 * sShv2;
        //基礎代謝とふるえ熱産生を加算した値を代謝量とする
        metabolism = basalMetabolism + mshiv;

        //衣服の温度を更新
        clothTemperature = (ra * skinTemperature + rcl * operatingTemp) / (ra + rcl);
      }
    }

    /// <summary>外界条件からSET*[C]を直接計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[CDB]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="velocity">相対気流速度[m/s]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="basalMetabolism">基礎代謝量[W/m2]</param>
    /// <param name="externalWork">外部仕事量[W/m2]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>SET*[C]</returns>
    public static double GetSETStarFromAmbientCondition
      (double drybulbTemperature, double meanRadiantTemperature, double relativeHumidity, double velocity, 
      double clothing, double basalMetabolism, double externalWork, double atmosphericPressure)
    {
      double st, ct, bt, clt, ss, ls, sr, lr, wd;
      GetSteadyState(drybulbTemperature, meanRadiantTemperature, relativeHumidity,
        velocity, clothing, basalMetabolism, externalWork, atmosphericPressure,
        out st, out ct, out bt, out clt, out ss, out ls, out sr, out lr, out wd);
      return GetSETStar(meanRadiantTemperature, basalMetabolism, externalWork, clt, st, ss, ls, wd);
    }

    /// <summary>SET*[C]を計算する</summary>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="basalMetabolism">基礎代謝量[W/m2]</param>
    /// <param name="externalWork">外部仕事量[W/m2]</param>
    /// <param name="clothTemperature">衣服温度[C]</param>
    /// <param name="skinTemperature">皮膚温度[C]</param>
    /// <param name="sensibleHFSkin">皮膚からの顕熱損失[W/m2]</param>
    /// <param name="latentHFSkin">皮膚からの潜熱損失[W/m2]</param>
    /// <param name="wettedness">皮膚の平均濡れ率[-]</param>
    /// <returns>SET*[C]</returns>
    public static double GetSETStar
      (double meanRadiantTemperature, double basalMetabolism, double externalWork, double clothTemperature,
      double skinTemperature, double sensibleHFSkin, double latentHFSkin, double wettedness)
    {
      //放射熱伝達率[W/(m2K)]の計算
      double radiativeHTransCoef = 4d * BLACK_CONSTANT * 0.72
        * Math.Pow((clothTemperature + meanRadiantTemperature) / 2d + CONVERT_C_TO_K, 3);
      //対流熱伝達率[W/(m2 K)]の計算
      double convectiveHTransCoef = 5.66 * Math.Pow(Math.Max(0, basalMetabolism / CONVERT_MET_TO_W - 0.85), 0.39);
      convectiveHTransCoef = Math.Max(convectiveHTransCoef, 3d);

      //標準着衣量[clo]の計算
      double sClothing = 1.3264 / ((basalMetabolism - externalWork) / CONVERT_MET_TO_W + 0.7383) - 0.0953;

      //着衣抵抗[m2K/W]の計算
      double rcl = 0.155 * sClothing;
      //着衣面積率[-]の計算
      double clothRate = 1d + K_CLO * sClothing;

      //潜熱伝達率[W/(m2 K)]の計算
      double lewis = 0.0555 * (skinTemperature + CONVERT_C_TO_K);
      double latentHTransCoef = 1 / (rcl / (I_CLS * lewis) + 1d / (clothRate * convectiveHTransCoef * lewis));

      //顕熱伝達率[W/(m2K)]の計算
      double hcr = radiativeHTransCoef + convectiveHTransCoef;
      double sensibleHTransCoef = 1 / (1 / (clothRate * hcr) + rcl);

      //反復計算でSET*を計算
      double hfSkin = sensibleHFSkin + latentHFSkin;
      double psk = Water.GetSaturationPressure(skinTemperature);
      Roots.ErrorFunction eFnc = delegate (double setStar)
      {
        return hfSkin - sensibleHTransCoef * (skinTemperature - setStar)
        - wettedness * latentHTransCoef * (psk - Water.GetSaturationPressure(setStar) * 0.5);
      };
      return Roots.Newton(eFnc, 26, 1e-3, 1e-3, 1e-3, 20);
    }

    #endregion

  }
}
