﻿/* NMSimplex.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Numerics
{
  /// <summary>NelderとMeadによる滑降シンプレックス法</summary>
  public static class NMSimplex
  {

    #region 定数宣言

    private const int MAX_ITERATION = 5000;

    private const double ERR_TOLERANCE = 1e-5;

    #endregion

    #region インスタンス変数

    private static double alpha = 1.0;

    private static double beta = 0.5;

    private static double gamma = 2.0;

    private static double rho = 1.0;

    private static double pow = 2.0;

    #endregion

    #region プロパティ

    /// <summary>反射の係数αを設定・取得する</summary>
    public static double Alpha
    {
      get { return alpha; }
      set { if (0 < value) alpha = value; }
    }

    /// <summary>膨張の係数βを設定・取得する</summary>
    public static double Beta
    {
      get { return beta; }
      set { if (0 < value && value < 1) beta = value; }
    }

    /// <summary>収縮の係数γを設定・取得する</summary>
    public static double Gamma
    {
      get { return gamma; }
      set { if (1 < value) gamma = value; }
    }

    /// <summary>ペナルティパラメータを設定・取得する</summary>
    public static double Rho
    {
      get { return rho; }
      set { if (0 < value) rho = value; }
    }

    /// <summary>ペナルティ乗数を設定・取得する</summary>
    public static double Pow
    {
      get { return pow; }
      set { if (0 < value) pow = value; }
    }

    #endregion

    #region delegate定義

    /// <summary>最適化する関数</summary>
    /// <param name="x">入力値</param>
    /// <returns>出力値</returns>
    public delegate double OptimizeFunction(double[] x);

    /// <summary>最適化する関数</summary>
    /// <param name="x">入力値</param>
    /// <param name="iteration">反復回数[回]</param>
    /// <returns>出力値</returns>
    private delegate double optimizeFunction(double[] x, int iteration);

    #endregion

    #region publicメソッド

    /// <summary>関数を最小化する入力を探索する</summary>
    /// <param name="mFnc">最小化関数</param>
    /// <param name="minX">Xの最小値リスト</param>
    /// <param name="maxX">Xの最大値リスト</param>
    /// <param name="success">収束成功の真偽</param>
    /// <returns>ユーザー関数が最小値をとる入力ベクトル</returns>
    public static double[] GetSolution
      (OptimizeFunction mFnc, double[] minX, double[] maxX, out bool success)
    {
      optimizeFunction fnc = delegate (double[] x, int iteration)
      {
        return mFnc(x);
      };
      double[][] points = makeInitialPoints(minX, maxX);
      return getSolution(fnc, points, false, out success);
    }

    /// <summary>関数を最小化する入力を探索する</summary>
    /// <param name="mFnc">最小化関数</param>
    /// <param name="cFnc">制約関数（f(x)=0とする）</param>
    /// <param name="minX">Xの最小値リスト</param>
    /// <param name="maxX">Xの最大値リスト</param>
    /// <param name="success">収束成功の真偽</param>
    /// <returns>ユーザー関数が最小値をとる入力ベクトル</returns>
    public static double[] GetSolution(OptimizeFunction mFnc,
      OptimizeFunction cFnc, double[] minX, double[] maxX, out bool success)
    {
      optimizeFunction fnc2 = delegate (double[] x, int iteration)
      {
        return mFnc(x) + (Rho * (iteration + 5)) * Math.Pow(Math.Abs(cFnc(x)), pow);
        //return mFnc(x) + Rho * Math.Pow(Math.Abs(cFnc(x)), pow);
      };
      double[][] points = makeInitialPoints(minX, maxX);
      return getSolution(fnc2, points, true, out success);
    }

    #endregion

    #region privateメソッド

    /// <summary>関数を最小化する入力を探索する</summary>
    /// <param name="fnc">ユーザー関数</param>
    /// <param name="points">探索範囲を示す配列</param>
    /// <param name="hasConstraint">制約付き問題か否か</param>
    /// <param name="success">収束成功の真偽</param>
    /// <returns>ユーザー関数が最小値をとる入力ベクトル</returns>
    private static double[] getSolution
      (optimizeFunction fnc, double[][] points, bool hasConstraint, out bool success)
    {
      //関数の次数を取得
      int num = points[0].Length;
      int iMin = 0;

      //計算用配列
      double[] newPt1 = new double[num];
      double[] newPt2 = new double[num];

      //評価値保持配列
      double[] ypi = new double[num + 1];

      //座標値を積算//重心計算に使用
      double[] sum = new double[num];
      summatePoints(points, ref sum);

      int iterNum = 0;
      success = true;
      while (true)
      {
        //評価値を再計算
        if (iterNum == 0 || hasConstraint)
          for (int i = 0; i <= num; i++) ypi[i] = fnc(points[i], iterNum);

        //最大値、二番目に大きい値、最小値の要素番号を計算
        double ave = 0;
        int iMax, iSec;
        if (ypi[0] < ypi[1]) { iMax = 1; iSec = 0; }
        else { iMax = 0; iSec = 1; }
        for (int i = 0; i <= num; i++)
        {
          ave += ypi[i];
          if (ypi[i] < ypi[iMin]) iMin = i;
          else if (ypi[iMax] < ypi[i])
          {
            iSec = iMax;
            iMax = i;
          }
          else if ((ypi[iSec] < ypi[i]) && (i != iMax)) iSec = i;
        }

        //収束判定
        ave /= (num + 1);
        double err = 0;
        for (int i = 0; i <= num; i++) err += Math.Abs(ypi[i] - ave);
        if (err / (num + 1) < ERR_TOLERANCE)
        {
          break;
        }

        //反射処理
        double yt1 = tryPoint(fnc, iterNum, points[iMax], sum, -Alpha, ref newPt1);
        if (yt1 < ypi[iMin])
        {
          //膨張処理
          double yt2 = tryPoint(fnc, iterNum, points[iMax], sum, Gamma, ref newPt2);
          if (yt2 < ypi[iMin])
          {
            switchPoint(ref points, ref sum, newPt2, iMax);
            ypi[iMax] = yt2;
          }
          else
          {
            switchPoint(ref points, ref sum, newPt1, iMax);
            ypi[iMax] = yt1;
          }
        }
        else
        {
          bool ltNxt = true;
          for (int i = 0; i < num; i++)
          {
            if (i != iMax && yt1 <= ypi[i])
            {
              ltNxt = false;
              break;
            }
          }
          if (ltNxt)
          {
            if (yt1 <= ypi[iMax])
            {
              switchPoint(ref points, ref sum, newPt1, iMax);
              ypi[iMax] = yt1;
            }
            //収縮処理
            yt1 = tryPoint(fnc, iterNum, points[iMax], sum, Beta, ref newPt1);
            if (ypi[iMax] < yt1)
            {
              //多次元収縮処理
              for (int i = 0; i <= num; i++)
              {
                if (i != iMin)
                {
                  for (int j = 0; j < num; j++)
                    points[i][j] = 0.5 * (points[i][j] + points[iMin][j]);
                }
                if (!hasConstraint) ypi[i] = fnc(points[i], iterNum);
              }
              summatePoints(points, ref sum);
            }
            else
            {
              //一次元収縮処理
              switchPoint(ref points, ref sum, newPt1, iMax);
              ypi[iMax] = yt1;
            }
          }
          else
          {
            switchPoint(ref points, ref sum, newPt1, iMax);
            ypi[iMax] = yt1;
          }
        }

        iterNum++;
        if (MAX_ITERATION < iterNum)
        {
          success = false;
          break;
        }
      }
      return points[iMin];
    }

    /// <summary>関数を最小化する入力を探索する</summary>
    /// <param name="fnc">ユーザー関数</param>
    /// <param name="points">探索範囲を示す配列</param>
    /// <param name="hasConstraint">制約付き問題か否か</param>
    /// <param name="success">収束成功の真偽</param>
    /// <returns>ユーザー関数が最小値をとる入力ベクトル</returns>
    private static double[] getSolution2
      (optimizeFunction fnc, double[][] points, bool hasConstraint, out bool success)
    {
      //関数の次数を取得
      int num = points[0].Length;
      int iMin = 0;

      //計算用配列
      double[] newPt1 = new double[num];
      double[] newPt2 = new double[num];

      //評価値保持配列
      double[] ypi = new double[num + 1];

      //座標値を積算//重心計算に使用
      double[] sum = new double[num];
      summatePoints(points, ref sum);

      int iterNum = 0;
      success = true;
      while (true)
      {
        //評価値を再計算
        if (iterNum == 0 || hasConstraint)
          for (int i = 0; i <= num; i++) ypi[i] = fnc(points[i], iterNum);

        //最大値、二番目に大きい値、最小値の要素番号を計算
        double ave = 0;
        int iMax, iSec;
        if (ypi[0] < ypi[1]) { iMax = 1; iSec = 0; }
        else { iMax = 0; iSec = 1; }
        for (int i = 0; i <= num; i++)
        {
          ave += ypi[i];
          if (ypi[i] < ypi[iMin]) iMin = i;
          else if (ypi[iMax] < ypi[i])
          {
            iSec = iMax;
            iMax = i;
          }
          else if ((ypi[iSec] < ypi[i]) && (i != iMax)) iSec = i;
        }

        //収束判定
        ave /= (num + 1);
        double err = 0;
        for (int i = 0; i <= num; i++) err += Math.Abs(ypi[i] - ave);
        if (err / (num + 1) < ERR_TOLERANCE)
        {
          break;
        }

        //反射処理
        double yt1 = tryPoint(fnc, iterNum, points[iMax], sum, -Alpha, ref newPt1);
        if (yt1 < ypi[iMax])
        {
          switchPoint(ref points, ref sum, newPt1, iMax);
          ypi[iMax] = yt1;
          if (yt1 < ypi[iMin])
          {
            //膨張処理
            double yt2 = tryPoint(fnc,iterNum, points[iMax], sum, Gamma, ref newPt2);
            if (yt2 < ypi[iMin])
            {
              switchPoint(ref points, ref sum, newPt2, iMax);
              ypi[iMax] = yt2;
            }
          }
        }
        else
        {
          //一次元収縮処理
          yt1 = tryPoint(fnc, iterNum, points[iMax], sum, Beta, ref newPt1);
          if (yt1 < ypi[iMax])
          {
            switchPoint(ref points, ref sum, newPt1, iMax);
            ypi[iMax] = yt1;
          }
          else
          {
            //多次元収縮処理
            for (int i = 0; i <= num; i++)
            {
              if (i != iMin)
              {
                for (int j = 0; j < num; j++)
                {
                  points[i][j] = 0.5 * (points[i][j] + points[iMin][j]);
                  if(! hasConstraint) ypi[i] = fnc(points[i], iterNum);
                }
              }
            }
            summatePoints(points, ref sum);
          }
        }
        iterNum++;
        if (MAX_ITERATION < iterNum)
        {
          success = false;
          break;
        }
      }
      return points[iMin];
    }

    /// <summary>初期値リストを作成する</summary>
    /// <param name="minX">Xの最小値リスト</param>
    /// <param name="maxX">Xの最大値リスト</param>
    /// <returns>初期値リスト</returns>
    private static double[][] makeInitialPoints(double[] minX, double[] maxX)
    {
      double[][] pnts = new double[minX.Length + 1][];
      for (int i = 0; i < pnts.Length; i++) pnts[i] = new double[minX.Length];

      MersenneTwister mt = new MersenneTwister(DateTime.Now.Second);
      for (int i = 0; i < pnts.Length; i++)
      {
        for (int j = 0; j < minX.Length; j++)
        {
          pnts[i][j] = minX[j] + (maxX[j] - minX[j]) * mt.NextDouble();
        }
      }
      return pnts;
    }

    /// <summary>座標を入れ替える</summary>
    /// <param name="points">座標リスト</param>
    /// <param name="sum">全座標の合算値</param>
    /// <param name="newPt">新しい座標</param>
    /// <param name="iMax">入れ替える要素番号</param>
    private static void switchPoint
      (ref double[][] points, ref double[] sum, double[] newPt, int iMax)
    {
      int num = newPt.Length;
      for (int i = 0; i < num; i++)
      {
        double buff = (newPt[i] - points[iMax][i]);
        sum[i] += buff;
        points[iMax][i] = newPt[i];
      }
    }

    /// <summary>新しい座標値を作成して関数を評価する</summary>
    /// <param name="fnc">ユーザー関数</param>
    /// <param name="iteration">反復回数</param>
    /// <param name="pt">旧座標</param>
    /// <param name="sum">全座標の合算値</param>
    /// <param name="cf">新座標取得のための係数</param>
    /// <param name="newPt">新座標</param>
    /// <returns>評価値</returns>
    private static double tryPoint
      (optimizeFunction fnc, int iteration, double[] pt, double[] sum, double cf, ref double[] newPt)
    {
      double cf1 = (1.0 - cf) / pt.Length;
      double cf2 = cf - cf1;
      for (int i = 0; i < pt.Length; i++) newPt[i] = sum[i] * cf1 + pt[i] * cf2;
      return fnc(newPt, iteration);
    }

    /// <summary>各座標を合算する</summary>
    /// <param name="points">座標配列</param>
    /// <param name="sum">出力：合算値</param>
    private static void summatePoints(double[][] points, ref double[] sum)
    {
      int num = points[0].Length;
      for (int i = 0; i < num; i++) sum[i] = 0;
      for (int i = 0; i < points.Length; i++)
      {
        for (int j = 0; j < num; j++) sum[j] += points[i][j];
      }
    }

    #endregion

  }
}
