﻿/* PCMWallLayer.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections.Generic;

namespace Popolo.ThermalLoad
{
  /// <summary>潜熱蓄熱材料を用いた壁層</summary>
  public class PCMWallLayer : WallLayer
  {

    #region 列挙型

    /// <summary>相</summary>
    [Flags]
    public enum State
    {
      /// <summary>固体</summary>
      Solid = 1,
      /// <summary>平衡</summary>
      Equilibrium = 2,
      /// <summary>液体</summary>
      Liquid = 4
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>各相の物性</summary>
    private Dictionary<State, WallLayer> layers = new Dictionary<State, WallLayer>();

    /// <summary>B側の現在の相を取得する</summary>
    public State CurrentState_B { get; private set; }

    /// <summary>F側の現在の相を取得する</summary>
    public State CurrentState_F { get; private set; }

    /// <summary>B側の過去の相を取得する</summary>
    public State LastState_B { get; private set; }

    /// <summary>F側の過去の相を取得する</summary>
    public State LastState_F { get; private set; }

    /// <summary>凝固点[C]を取得する</summary>
    public double FreezingTemperature { get; private set; }

    /// <summary>融解点[C]を取得する</summary>
    public double MeltingTemperature { get; private set; }

    #endregion

    #region インスタンスメソッド

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="freezingTemperature">凝固点[C]</param>
    /// <param name="meltingTemperature">融点[C]</param>
    /// <param name="thickness">壁層の厚み[m]</param>
    /// <param name="solidLayer">固体状態の物性</param>
    /// <param name="equilibriumLayer">平衡状態の物性</param>
    /// <param name="liquidLayer">液体状態の物性</param>
    public PCMWallLayer(string name, double freezingTemperature, double meltingTemperature,
      double thickness, WallLayer solidLayer, WallLayer equilibriumLayer, WallLayer liquidLayer)

    {
      IsVariableProperties = true;
      FreezingTemperature = freezingTemperature;
      MeltingTemperature = meltingTemperature;
      Thickness = thickness;
      layers.Add(State.Solid, solidLayer);
      layers.Add(State.Equilibrium, equilibriumLayer);
      layers.Add(State.Liquid, liquidLayer);

      CurrentState_F = CurrentState_B = State.Equilibrium;
      UpdateState(freezingTemperature, freezingTemperature);
    }

    /// <summary>温度から状態を判定する</summary>
    /// <param name="temp">温度</param>
    /// <returns>状態（固体・平衡・液相の別）</returns>
    private State getState(double temp)
    {
      if (temp <= FreezingTemperature) return State.Solid;
      if (temp < MeltingTemperature) return State.Equilibrium;
      else return State.Liquid;
    }

    /// <summary>熱容量[J/(m2K)]を計算する</summary>
    /// <param name="state">相</param>
    /// <returns>熱容量[J/(m2K)]</returns>
    public double GetHeatCapacity(State state)
    { return 0.5 * layers[state].VolSpecificHeat * Thickness * 1000d; }

    /// <summary>壁層端部の温度に応じて状態値を更新する</summary>
    /// <param name="temperatureF">F側温度</param>
    /// <param name="temperatureB">B側温度</param>
    /// <returns>状態値変更の有無</returns>
    public override bool UpdateState(double temperatureF, double temperatureB)
    {
      LastState_F = CurrentState_F;
      LastState_B = CurrentState_B;
      CurrentState_F = getState(temperatureF);
      CurrentState_B = getState(temperatureB);      

      bool phaseChanged = false;
      if ((CurrentState_F != LastState_F) || (CurrentState_B != LastState_B))
        phaseChanged = true;
      
      if (phaseChanged)
      {
        double res = 
          Thickness / layers[CurrentState_F].ThermalConductivity + 
          Thickness / layers[CurrentState_B].ThermalConductivity;
        HeatConductance = 2d / res;
        double th1000 = Thickness * 1000d;
        HeatCapacity_B = 0.5 * layers[CurrentState_B].VolSpecificHeat * th1000;
        HeatCapacity_F = 0.5 * layers[CurrentState_F].VolSpecificHeat * th1000;

        ThermalConductivity = HeatConductance * Thickness;
        VolSpecificHeat = (HeatCapacity_F + HeatCapacity_B) / th1000;
      }

      return phaseChanged;
    }

    #endregion

  }
}
