﻿/* SunShade.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Weather;

namespace Popolo.ThermalLoad
{
  /// <summary>日除け</summary>
  public class SunShade
  {

    #region 列挙型

    /// <summary>日除けの形状</summary>
    public enum Shapes
    {
      /// <summary>無し</summary>
      None = 0,
      /// <summary>水平庇</summary>
      Horizontal = 1,
      /// <summary>水平庇（無限大長）</summary>
      LongHorizontal = 2,
      /// <summary>袖壁（左）</summary>
      VerticalLeft = 3,
      /// <summary>袖壁（右）</summary>
      VerticalRight = 4,
      /// <summary>袖壁（両方）</summary>
      VerticalBoth = 5,
      /// <summary>袖壁（無限大長:左）</summary>
      LongVerticalLeft = 6,
      /// <summary>袖壁（無限大長:右）</summary>
      LongVerticalRight = 7,
      /// <summary>袖壁（無限大長:両方）</summary>
      LongVerticalBoth = 8,
      /// <summary>格子ルーバー</summary>
      Grid = 9
    }

    #endregion

    #region プロパティ

    /// <summary>傾斜面を取得する</summary>
    public ImmutableIncline Incline { get; private set; }

    /// <summary>形状を取得する</summary>
    public Shapes Shape { private set; get; }

    /// <summary>窓高さ[m]を取得する</summary>
    public double WinHeight { private set; get; }

    /// <summary>窓幅[m]を取得する</summary>
    public double WinWidth { private set; get; }

    /// <summary>張り出し幅[m]を取得する</summary>
    public double Overhang { private set; get; }

    /// <summary>上部マージン[m]を取得する</summary>
    public double TopMargin { private set; get; }

    /// <summary>下部マージン[m]を取得する</summary>
    public double BottomMargin { private set; get; }

    /// <summary>左側マージン[m]を取得する</summary>
    public double LeftMargin { private set; get; }

    /// <summary>右側マージン[m]を取得する</summary>
    public double RightMargin { private set; get; }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタを隠蔽</summary>
    private SunShade() { }

    #endregion

    #region インスタンスメソッド

    /// <summary>日影面積率[-]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>日影面積率[-]</returns>
    public double GetShadowRate(ImmutableSun sun)
    {
      //日の出前と日没後はすべて影      
      if (sun.Altitude <= 0) return 1;
      //日除けが無ければ影は無し
      if (Shape == Shapes.None) return 0;
      //太陽が裏面にある場合にはすべて影
      if (Incline.GetDirectSolarRadiationRate(sun) <= 0) return 1;

      double dpW = Overhang * Math.Tan(Incline.HorizontalAngle - sun.Orientation);
      double dpH = Overhang * Incline.GetTangentProfileAngle(sun);
      double sr = 0;
      switch (Shape)
      {
        case Shapes.Horizontal:
          sr = computeShadowArea_H(dpW, dpH, WinWidth, WinHeight, LeftMargin, TopMargin, RightMargin);
          break;
        case Shapes.VerticalLeft:
          sr = computeShadowArea_H(dpH, dpW, WinHeight, WinWidth, TopMargin, LeftMargin, BottomMargin);
          break;
        case Shapes.VerticalRight:
          sr = computeShadowArea_H(dpH, -dpW, WinHeight, WinWidth, TopMargin, RightMargin, BottomMargin);
          break;
        case Shapes.VerticalBoth:
          if (dpW < 0)
            sr = computeShadowArea_H(dpH, -dpW, WinHeight, WinWidth, TopMargin, RightMargin, BottomMargin);
          else sr = computeShadowArea_H(dpH, dpW, WinHeight, WinWidth, TopMargin, LeftMargin, BottomMargin);
          break;
        case Shapes.LongHorizontal:
          sr = computeShadowArea_LH(dpH, WinWidth, WinHeight, TopMargin);
          break;
        case Shapes.LongVerticalLeft:
          sr = computeShadowArea_LH(dpW, WinHeight, WinWidth, LeftMargin);
          break;
        case Shapes.LongVerticalRight:
          sr = computeShadowArea_LH(-dpW, WinHeight, WinWidth, RightMargin);
          break;
        case Shapes.LongVerticalBoth:
          if (dpW < 0) sr = computeShadowArea_LH(-dpW, WinHeight, WinWidth, RightMargin);
          else sr = computeShadowArea_LH(dpW, WinHeight, WinWidth, LeftMargin);
          break;
        case Shapes.Grid:
          sr = computeShadowArea_Grid
            (dpW, dpH, WinWidth, WinHeight, LeftMargin, TopMargin, RightMargin, BottomMargin);
          break;
      }
      return sr / (WinHeight * WinWidth);
    }

    #endregion

    #region 初期化処理

    /// <summary>サイズ0の日除けを作成する</summary>
    /// <returns>サイズ0の日除け</returns>
    public static SunShade MakeEmptySunShade()
    {
      SunShade ss = new SunShade();
      ss.Shape = Shapes.None;
      return ss;
    }

    /// <summary>水平庇（無限長）を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="tMargin">上部マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>水平庇（無限長）</returns>
    public static SunShade MakeHorizontalSunShade
      (double wWidth, double wHeight, double depth, double tMargin, ImmutableIncline incline)
    {
      SunShade ss = new SunShade();
      ss.Shape = Shapes.LongHorizontal;
      ss.WinWidth = wWidth;
      ss.WinHeight = wHeight;
      ss.Overhang = depth;
      ss.TopMargin = tMargin;
      ss.Incline = incline;
      return ss;
    }

    /// <summary>水平庇を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="lMargin">左側マージン[m]</param>
    /// <param name="rMargin">右側マージン[m]</param>
    /// <param name="tMargin">上部マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>水平庇</returns>
    public static SunShade MakeHorizontalSunShade(double wWidth, double wHeight, double depth, 
      double lMargin, double rMargin, double tMargin, ImmutableIncline incline)
    {
      SunShade ss = MakeHorizontalSunShade(wWidth, wHeight, depth, tMargin, incline);
      ss.Shape = Shapes.Horizontal;
      ss.LeftMargin = lMargin;
      ss.RightMargin = rMargin;
      return ss;
    }

    /// <summary>袖壁（無限長）を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="sMargin">横側マージン[m]</param>
    /// <param name="isLeftSide">左側か否か（右の場合はfalse）</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>袖壁（無限長）</returns>
    public static SunShade MakeVerticalSunShade(double wWidth, double wHeight, double depth, 
      double sMargin, bool isLeftSide, ImmutableIncline incline)
    {
      SunShade ss = new SunShade();
      ss.WinWidth = wWidth;
      ss.WinHeight = wHeight;
      ss.Overhang = depth;
      if (isLeftSide)
      {
        ss.Shape = Shapes.LongVerticalLeft;
        ss.LeftMargin = sMargin;
      }
      else
      {
        ss.Shape = Shapes.LongVerticalRight;
        ss.RightMargin = sMargin;
      }
      ss.Incline = incline;
      return ss;
    }

    /// <summary>袖壁を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="sMargin">横側マージン[m]</param>
    /// <param name="isLeftSide">左側か否か（右の場合はfalse）</param>
    /// <param name="tMargin">上部マージン[m]</param>
    /// <param name="bMargin">下部マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>袖壁</returns>
    public static SunShade MakeVerticalSunShade(double wWidth, double wHeight, double depth, 
      double sMargin, bool isLeftSide, double tMargin, double bMargin, ImmutableIncline incline)
    {
      SunShade ss = MakeVerticalSunShade(wWidth, wHeight, depth, sMargin, isLeftSide, incline);
      if (isLeftSide) ss.Shape = Shapes.VerticalLeft;
      else ss.Shape = Shapes.VerticalRight;
      ss.TopMargin = tMargin;
      ss.BottomMargin = bMargin;
      return ss;
    }

    /// <summary>袖壁（無限長）を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="lMargin">左側マージン[m]</param>
    /// <param name="rMargin">右側マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>袖壁（無限長）</returns>
    public static SunShade MakeVerticalSunShade(double wWidth, double wHeight, double depth, 
      double lMargin, double rMargin, ImmutableIncline incline)
    {
      SunShade ss = MakeVerticalSunShade(wWidth, wHeight, depth, lMargin, true, incline);
      ss.RightMargin = rMargin;
      ss.Shape = Shapes.LongVerticalBoth;
      return ss;
    }

    /// <summary>袖壁を作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="lMargin">左側マージン[m]</param>
    /// <param name="rMargin">右側マージン[m]</param>
    /// <param name="tMargin">上部マージン[m]</param>
    /// <param name="bMargin">下部マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>袖壁</returns>
    public static SunShade MakeVerticalSunShade(double wWidth, double wHeight, double depth, 
      double lMargin, double rMargin, double tMargin, double bMargin, ImmutableIncline incline)
    {
      SunShade ss = MakeVerticalSunShade(wWidth, wHeight, depth, lMargin, rMargin, incline);
      ss.Shape = Shapes.VerticalBoth;
      ss.TopMargin = tMargin;
      ss.BottomMargin = bMargin;
      return ss;
    }

    /// <summary>格子ルーバーを作成する</summary>
    /// <param name="wWidth">窓幅[m]</param>
    /// <param name="wHeight">窓高[m]</param>
    /// <param name="depth">張り出し幅[m]</param>
    /// <param name="lMargin">左側マージン[m]</param>
    /// <param name="rMargin">右側マージン[m]</param>
    /// <param name="tMargin">上部マージン[m]</param>
    /// <param name="bMargin">下部マージン[m]</param>
    /// <param name="incline">傾斜面</param>
    /// <returns>格子ルーバー</returns>
    public static SunShade MakeGridSunShade(double wWidth, double wHeight, double depth, 
      double lMargin, double rMargin, double tMargin, double bMargin, ImmutableIncline incline)
    {
      SunShade ss = new SunShade();
      ss.Shape = Shapes.Grid;
      ss.WinWidth = wWidth;
      ss.WinHeight = wHeight;
      ss.Overhang = depth;
      ss.LeftMargin = lMargin;
      ss.RightMargin = rMargin;
      ss.TopMargin = tMargin;
      ss.BottomMargin = bMargin;
      ss.Incline = incline;
      return ss;
    }

    #endregion

    #region private staticメソッド

    /// <summary>水平庇の日陰面積[m2]を計算する</summary>
    /// <param name="dpW">庇端部から影端部までの水平距離[m]</param>
    /// <param name="dpH">庇端部から影端部までの垂直距離[m]</param>
    /// <param name="dwW">窓幅[m]</param>
    /// <param name="dwH">窓高さ[m]</param>
    /// <param name="dmWA">庇と窓の左端距離[m]</param>
    /// <param name="dmH">庇と窓の上端距離[m]</param>
    /// <param name="dmWB">庇と窓の右端距離[m]</param>
    /// <returns>水平庇の日陰面積[m2]</returns>
    private static double computeShadowArea_H
      (double dpW, double dpH, double dwW, double dwH, double dmWA, double dmH, double dmWB)
    {
      if (dpH <= dmH) return 0;

      double dmW;
      if (0 < dpW) dmW = dmWA;
      else dmW = dmWB;
      dpW = Math.Abs(dpW);

      double dshHA, dshHB, dshWA, dshWB;
      double dshHAd, dshHBd, dshWAd, dshWBd;
      if (dmW < dpW) dshHAd = dmW * (dpH / dpW) - dmH;
      else dshHAd = dpH - dmH;
      dshHA = Math.Min(dwH, Math.Max(0, dshHAd));
      if (dmW + dwW < dpW) dshHBd = (dmW + dwW) * (dpH / dpW) - dmH;
      else dshHBd = dpH - dmH;
      dshHB = Math.Min(dwH, Math.Max(0, dshHBd));
      dshWAd = (dmW + dwW) - dmH * (dpW / dpH);
      dshWA = Math.Min(dwW, Math.Max(0, dshWAd));
      if (dpH < dmH + dwH) dshWBd = (dmW + dwW) - dpW;
      else dshWBd = (dmW + dwW) - (dmH + dwH) * dpW / dpH;
      dshWB = Math.Min(dwW, Math.Max(0, dshWBd));

      return dshWA * dshHA + 0.5 * (dshWA + dshWB) * (dshHB - dshHA);
    }

    /// <summary>無限長の水平庇の日陰面積[m2]を計算する</summary>
    /// <param name="dpH">庇端部から影端部までの垂直距離[m]</param>
    /// <param name="dwW">窓幅[m]</param>
    /// <param name="dwH">窓高さ[m]</param>
    /// <param name="dmH">庇と窓の上端距離[m]</param>
    /// <returns>無限長の水平庇の日陰面積[m2]</returns>
    private static double computeShadowArea_LH(double dpH, double dwW, double dwH, double dmH)
    {
      if (dpH <= dmH) return 0;
      return dwW * Math.Min(dpH - dmH, dwH);
    }

    /// <summary>格子ルーバーの日陰面積[m2]を計算する</summary>
    /// <param name="dpW">庇端部から影端部までの水平距離[m]</param>
    /// <param name="dpH">庇端部から影端部までの垂直距離[m]</param>
    /// <param name="dwW">窓幅[m]</param>
    /// <param name="dwH">窓高さ[m]</param>
    /// <param name="dmWA">庇と窓の左端距離[m]</param>
    /// <param name="dmHA">庇と窓の上端距離[m]</param>
    /// <param name="dmWB">庇と窓の右端距離[m]</param>
    /// <param name="dmHB">庇と窓の下端距離[m]</param>
    /// <returns>格子ルーバーの日陰面積[m2]</returns>
    private static double computeShadowArea_Grid
      (double dpW, double dpH, double dwW, double dwH, double dmWA, double dmHA, double dmWB, double dmHB)
    {
      double dWA = Math.Min(Math.Max(0, dpW - dmWA), dwW);
      double dWB = Math.Min(Math.Max(0, -(dpW + dmWB)), dwW);
      double dHA = Math.Min(Math.Max(0, dpH - dmHA), dwH);
      double dHB = Math.Min(Math.Max(0, -(dpH + dmHB)), dwH);
      return dwW * (dHA + dHB) + (dWA + dWB) * (dwH - dHA - dHB);
    }

    #endregion

  }
}
