﻿/* WallLayer.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

namespace Popolo.ThermalLoad
{
  /// <summary>壁層</summary>
  public class WallLayer: ICloneable, ImmutableWallLayer
  {

    #region 定数宣言

    /// <summary>湿り空気の密度[kg/m3]</summary>
    private const double MOISTAIR_DENSITY = 1.2;

    #endregion

    #region 列挙型定義

    /// <summary>壁材料</summary>
    public enum Materials
    {
      /// <summary>セメント・モルタル</summary>
      Mortar,
      /// <summary>鉄筋コンクリート</summary>
      ReinforcedConcrete,
      /// <summary>軽量骨材コンクリート1種</summary>
      LightweightAggregateConcrete1,
      /// <summary>軽量骨材コンクリート2種</summary>
      LightweightAggregateConcrete2,
      /// <summary>軽量気泡コンクリート（ALC）</summary>
      AutomaticLevelControl,
      /// <summary>普通れんが</summary>
      Brick,
      /// <summary>耐火れんが</summary>
      FireBrick,
      /// <summary>銅</summary>
      Copper,
      /// <summary>アルミニウム合金</summary>
      Aluminum,
      /// <summary>鋼材</summary>
      Steel,
      /// <summary>鉛</summary>
      Lead,
      /// <summary>ステンレス鋼</summary>
      StainlessSteel,
      /// <summary>フロートガラス(窓ガラスではない）</summary>
      FloatGlass,
      /// <summary>PVC(塩化ビニル)</summary>
      PolyvinylChloride,
      /// <summary>天然木材1類（桧、杉、えぞ松等）</summary>
      Wood1,
      /// <summary>天然木材2類（松、ラワン等）</summary>
      Wood2,
      /// <summary>天然木材3類（ナラ、サクラ、ブナ等）</summary>
      Wood3,
      /// <summary>合板</summary>
      Plywood,
      /// <summary>断熱木毛セメント</summary>
      WoodWoolCement,
      /// <summary>木片セメント</summary>
      WoodChipCement,
      /// <summary>ハードボード</summary>
      HardBoard,
      /// <summary>パーティクルボード</summary>
      ParticleBoard,
      /// <summary>せっこうボード</summary>
      PlasterBoard,
      /// <summary>せっこうプラスター</summary>
      GypsumPlaster,
      /// <summary>漆喰</summary>
      WhiteWash,
      /// <summary>土壁</summary>
      SoilWall,
      /// <summary>繊維質上塗材</summary>
      FiberCoating,
      /// <summary>畳床</summary>
      Tatami,
      /// <summary>タイル</summary>
      Tile,
      /// <summary>プラスチック（P）タイル</summary>
      PlasticTile,
      /// <summary>住宅用グラスウール断熱材 10K相当</summary>
      GlassWoolInsulation_10K,
      /// <summary>住宅用グラスウール断熱材 16K相当</summary>
      GlassWoolInsulation_16K,
      /// <summary>住宅用グラスウール断熱材 24K相当</summary>
      GlassWoolInsulation_24K,
      /// <summary>住宅用グラスウール断熱材 32K相当</summary>
      GlassWoolInsulation_34K,
      /// <summary>高性能グラスウール断熱材 16K相当</summary>
      HighGradeGlassWoolInsulation_16K,
      /// <summary>高性能グラスウール断熱材 24K相当</summary>
      HighGradeGlassWoolInsulation_24K,
      /// <summary>吹込用グラスウール断熱材1種 13K相当</summary>
      BlowingGlassWoolInsulation_13K,
      /// <summary>吹込用グラスウール断熱材2種 18K相当</summary>
      BlowingGlassWoolInsulation_18K,
      /// <summary>吹込用グラスウール断熱材2種 30K相当</summary>
      BlowingGlassWoolInsulation_30K,
      /// <summary>吹込用グラスウール断熱材2種 35K相当</summary>
      BlowingGlassWoolInsulation_35K,
      /// <summary>住宅用ロックウール断熱材 マット</summary>
      RockWoolInsulationMat,
      /// <summary>住宅用ロックウール断熱材 フェルト</summary>
      RockWoolInsulationFelt = 42,
      /// <summary>住宅用ロックウール断熱材 ボード</summary>
      RockWoolInsulationBoard = 43,
      /// <summary>吹込用ロックウール断熱材 25K</summary>
      BlowingRockWoolInsulation_25K,
      /// <summary>吹込用ロックウール断熱材 35K</summary>
      BlowingRockWoolInsulation_35K,
      /// <summary>ロックウール化粧吸音板</summary>
      RockWoolAcousticBoard,
      /// <summary>吹付けロックウール</summary>
      SprayedRockWool,
      /// <summary>ビーズ法ポリスチレンフォーム特号</summary>
      BeadMethodPolystyreneFoam_S,
      /// <summary>ビーズ法ポリスチレンフォーム1号</summary>
      BeadMethodPolystyreneFoam_1,
      /// <summary>ビーズ法ポリスチレンフォーム2号</summary>
      BeadMethodPolystyreneFoam_2,
      /// <summary>ビーズ法ポリスチレンフォーム3号</summary>
      BeadMethodPolystyreneFoam_3,
      /// <summary>ビーズ法ポリスチレンフォーム4号</summary>
      BeadMethodPolystyreneFoam_4,
      /// <summary>押出法ポリスチレンフォーム 1種</summary>
      ExtrudedPolystyreneFoam_1,
      /// <summary>押出法ポリスチレンフォーム 2種</summary>
      ExtrudedPolystyreneFoam_2,
      /// <summary>押出法ポリスチレンフォーム 3種</summary>
      ExtrudedPolystyreneFoam_3,
      /// <summary>硬質ウレタンフォーム保温版1種1号</summary>
      RigidUrethaneFoam_1_1,
      /// <summary>硬質ウレタンフォーム保温版1種2号</summary>
      RigidUrethaneFoam_1_2,
      /// <summary>硬質ウレタンフォーム保温版1種3号</summary>
      RigidUrethaneFoam_1_3,
      /// <summary>硬質ウレタンフォーム保温版2種1号</summary>
      RigidUrethaneFoam_2_1,
      /// <summary>硬質ウレタンフォーム保温版2種2号</summary>
      RigidUrethaneFoam_2_2,
      /// <summary>硬質ウレタンフォーム保温版2種3号</summary>
      RigidUrethaneFoam_2_3,
      /// <summary>硬質ウレタンフォーム（現場発泡品）</summary>
      RigidUrethaneFoam_OnSite,
      /// <summary>ポリエチレンフォーム A</summary>
      PolyethyleneFoam_A,
      /// <summary>ポリエチレンフォーム B</summary>
      PolyethyleneFoam_B,
      /// <summary>フェノールフォーム保温版 1種1号</summary>
      PhenolicFoam_1_1,
      /// <summary>フェノールフォーム保温版 1種2号</summary>
      PhenolicFoam_1_2,
      /// <summary>フェノールフォーム保温版 2種1号</summary>
      PhenolicFoam_2_1,
      /// <summary>フェノールフォーム保温版 2種2号</summary>
      PhenolicFoam_2_2,
      /// <summary>A級インシュレーションボード</summary>
      InsulationBoard_A,
      /// <summary>タタミボード</summary>
      TatamiBoard,
      /// <summary>シージングボード</summary>
      SheathingInsulationBoard,
      /// <summary>吹込用セルローズファイバー断熱材1</summary>
      CelluloseFiberInsulation_1,
      /// <summary>吹込用セルローズファイバー断熱材2</summary>
      CelluloseFiberInsulation_2,
      /// <summary>土壌（ローム質）</summary>
      Soil,
      /// <summary>EPS</summary>
      ExpandedPolystyrene,
      /// <summary>外装材</summary>
      CoveringMaterial,
      /// <summary>合成樹脂・リノリウム</summary>
      Linoleum,
      /// <summary>カーペット</summary>
      Carpet,
      /// <summary>石綿スレート</summary>
      AsbestosPlate,
      /// <summary>密閉空気層</summary>
      SealedAirGap,
      /// <summary>非密閉空気層</summary>
      AirGap,
      /// <summary>ポリスチレンフォーム</summary>
      PolystyreneFoam,
      /// <summary>スチレン発泡板</summary>
      StyreneFoam,
      /// <summary>ゴムタイル</summary>
      RubberTile,
      /// <summary>瓦</summary>
      Kawara,
      /// <summary>軽量コンクリート</summary>
      LightweightConcrete,
      /// <summary>防水層（アスファルトルーフィング）</summary>
      Asphalt,
      /// <summary>フレキシブルボード</summary>
      FrexibleBoard,
      /// <summary>珪酸カルシウム板</summary>
      CalciumSilicateBoard,
      /// <summary>高性能フェノールボード</summary>
      PhenolicFoam,
      /// <summary>花崗岩</summary>
      Granite,
      /// <summary>アクリル樹脂</summary>
      AcrylicResin
    }

    #endregion

    #region プロパティ

    /// <summary>名称を設定・取得する</summary>
    public string Name { get; set; }

    /// <summary>物性が変化しうるか</summary>
    public bool IsVariableProperties { get; protected set; }

    /// <summary>熱伝導率[W/(mK)]を取得する</summary>
    public double ThermalConductivity { get; protected set; }

    /// <summary>湿気伝導率[(kg/s) / ((kg/kg)·m)]を取得する</summary>
    public double MoistureConductivity { get; protected set; }

    /// <summary>容積比熱[kJ/(m^3K)]を取得する</summary>
    public double VolSpecificHeat { get; protected set; }

    /// <summary>熱コンダクタンス[W/(m2K)]を取得する</summary>
    public double HeatConductance { get; protected set; }

    /// <summary>顕熱容量（F側）[J/(m2K)]を取得する</summary>
    public double HeatCapacity_F { get; protected set; }

    /// <summary>顕熱容量（B側）[J/(m2K)]を取得する</summary>
    public double HeatCapacity_B { get; protected set; }

    /// <summary>水分容量[kg/m2]を取得する</summary>
    public double WaterCapacity { get; protected set; }

    /// <summary>単位絶対湿度差に対する吸湿係数[kg/m2]を取得する</summary>
    public double KappaC { get; private set; }

    /// <summary>単位温度差に対する放湿係数[kg/(m2K)]を取得する</summary>
    public double NuC { get; private set; }

    /// <summary>壁層の厚み[m]を取得する</summary>
    public double Thickness { get; protected set; }

    #endregion

    #region インスタンスメソッド

    /// <summary>インスタンスを初期化する</summary>
    protected WallLayer() { }

    /// <summary>新しいインスタンスを初期化する</summary>
    /// <param name="material">材料</param>
    /// <param name="thickness">壁層の厚み[m]</param>
    public WallLayer(Materials material, double thickness)
    {
      switch (material)
      {
        case Materials.Mortar:
          initialize("Mortar", 1.512, 1591.0, thickness);
          break;
        case Materials.ReinforcedConcrete:
          initialize("Reinforced Concrete", 1.600, 1896.0, thickness);
          break;
        case Materials.LightweightAggregateConcrete1:
          initialize("Lightweight Aggregate Concrete 1", 0.810, 1900.0, thickness);
          break;
        case Materials.LightweightAggregateConcrete2:
          initialize("Lightweight Aggregate Concrete 2", 0.580, 1599.0, thickness);
          break;
        case Materials.AutomaticLevelControl:
          initialize("Automatic Level Control", 0.170, 661.4, thickness);
          break;
        case Materials.Brick:
          initialize("Brick", 0.620, 1386.0, thickness);
          break;
        case Materials.FireBrick:
          initialize("FireBrick", 0.990, 1553.0, thickness);
          break;
        case Materials.Copper:
          initialize("Copper", 370.100, 3144.0, thickness);
          break;
        case Materials.Aluminum:
          initialize("Aluminum", 200.000, 2428.0, thickness);
          break;
        case Materials.Steel:
          initialize("Steel", 53.010, 3759.0, thickness);
          break;
        case Materials.Lead:
          initialize("Lead", 35.010, 1469.0, thickness);
          break;
        case Materials.StainlessSteel:
          initialize("Stainless Steel", 15.000, 3479.0, thickness);
          break;
        case Materials.FloatGlass:
          initialize("Float Glass", 1.000, 1914.0, thickness);
          break;
        case Materials.PolyvinylChloride:
          initialize("Polyvinyl Chloride", 0.170, 1023.0, thickness);
          break;
        case Materials.Wood1:
          initialize("Wood (Cedar)", 0.120, 519.1, thickness);
          break;
        case Materials.Wood2:
          initialize("Wood (Pine, Lauan)", 0.150, 648.8, thickness);
          break;
        case Materials.Wood3:
          initialize("Wood (Cherry, Fagaceae)", 0.190, 845.6, thickness);
          break;
        case Materials.Plywood:
          initialize("Plywood", 0.190, 716.0, thickness);
          break;
        case Materials.WoodWoolCement:
          initialize("Wood Wool Cement", 0.100, 841.4, thickness);
          break;
        case Materials.WoodChipCement:
          initialize("Wood Chip Cement", 0.170, 1679.0, thickness);
          break;
        case Materials.HardBoard:
          initialize("Hard Board", 0.170, 1233.0, thickness);
          break;
        case Materials.ParticleBoard:
          initialize("Particle Board", 0.150, 715.8, thickness);
          break;
        case Materials.PlasterBoard:
          initialize("Plaster Board", 0.170, 1030.0, thickness);
          break;
        case Materials.GypsumPlaster:
          initialize("Gypsum Plaster", 0.600, 1637.0, thickness);
          break;
        case Materials.WhiteWash:
          initialize("White Wash", 0.700, 1093.0, thickness);
          break;
        case Materials.SoilWall:
          initialize("Soil Wall", 0.690, 1126.0, thickness);
          break;
        case Materials.FiberCoating:
          initialize("Fiber Coating", 0.120, 4.2, thickness);
          break;
        case Materials.Tatami:
          initialize("Tatami", 0.110, 527.4, thickness);
          break;
        case Materials.Tile:
          initialize("Tile", 1.300, 2018.0, thickness);
          break;
        case Materials.PlasticTile:
          initialize("Plastic Tile", 0.190, 4.2, thickness);
          break;
        case Materials.GlassWoolInsulation_10K:
          initialize("Glass Wool Insulation 10kg/m3", 0.050, 8.4, thickness);
          break;
        case Materials.GlassWoolInsulation_16K:
          initialize("Glass Wool Insulation 16kg/m3", 0.045, 13.4, thickness);
          break;
        case Materials.GlassWoolInsulation_24K:
          initialize("Glass Wool Insulation 24kg/m3", 0.038, 20.1, thickness);
          break;
        case Materials.GlassWoolInsulation_34K:
          initialize("Glass Wool Insulation 32kg/m3", 0.036, 26.8, thickness);
          break;
        case Materials.HighGradeGlassWoolInsulation_16K:
          initialize("High Grade Glass Wool Insulation 16kg/m3", 0.038, 13.4, thickness);
          break;
        case Materials.HighGradeGlassWoolInsulation_24K:
          initialize("High Grade Glass Wool Insulation 24kg/m3", 0.036, 20.1, thickness);
          break;
        case Materials.BlowingGlassWoolInsulation_13K:
          initialize("Blowing Glass Wool Insulation 13kg/m3", 0.052, 10.9, thickness);
          break;
        case Materials.BlowingGlassWoolInsulation_18K:
          initialize("Blowing Glass Wool Insulation 18kg/m3", 0.052, 16.7, thickness);
          break;
        case Materials.BlowingGlassWoolInsulation_30K:
          initialize("Blowing Glass Wool Insulation 30kg/m3", 0.040, 29.3, thickness);
          break;
        case Materials.BlowingGlassWoolInsulation_35K:
          initialize("Blowing Glass Wool Insulation 35kg/m3", 0.040, 37.7, thickness);
          break;
        case Materials.RockWoolInsulationMat:
          initialize("Rock Wool Insulation Mat", 0.038, 33.5, thickness);
          break;
        case Materials.RockWoolInsulationFelt:
          initialize("Rock Wool Insulation Felt", 0.038, 41.9, thickness);
          break;
        case Materials.RockWoolInsulationBoard:
          initialize("Rock Wool Insulation Board", 0.036, 58.6, thickness);
          break;
        case Materials.BlowingRockWoolInsulation_25K:
          initialize("Blowing Rock Wool Insulation 25kg/m3", 0.047, 20.9, thickness);
          break;
        case Materials.BlowingRockWoolInsulation_35K:
          initialize("Blowing Rock Wool Insulation 35kg/m3", 0.051, 29.3, thickness);
          break;
        case Materials.RockWoolAcousticBoard:
          initialize("Rock Wool Acoustic Board", 0.058, 293.9, thickness);
          break;
        case Materials.SprayedRockWool:
          initialize("Sprayed Rock Wool", 0.047, 167.9, thickness);
          break;
        case Materials.BeadMethodPolystyreneFoam_S:
          initialize("Bead Method Polystyrene Foam S", 0.034, 33.9, thickness);
          break;
        case Materials.BeadMethodPolystyreneFoam_1:
          initialize("Bead Method Polystyrene Foam 1", 0.036, 37.7, thickness);
          break;
        case Materials.BeadMethodPolystyreneFoam_2:
          initialize("Bead Method Polystyrene Foam 2", 0.037, 31.4, thickness);
          break;
        case Materials.BeadMethodPolystyreneFoam_3:
          initialize("Bead Method Polystyrene Foam 3", 0.040, 25.1, thickness);
          break;
        case Materials.BeadMethodPolystyreneFoam_4:
          initialize("Bead Method Polystyrene Foam 4", 0.043, 18.8, thickness);
          break;
        case Materials.ExtrudedPolystyreneFoam_1:
          initialize("Extruded Polystyrene Foam 1", 0.040, 25.1, thickness);
          break;
        case Materials.ExtrudedPolystyreneFoam_2:
          initialize("Extruded Polystyrene Foam 2", 0.034, 25.1, thickness);
          break;
        case Materials.ExtrudedPolystyreneFoam_3:
          initialize("Extruded Polystyrene Foam 3", 0.028, 25.1, thickness);
          break;
        case Materials.RigidUrethaneFoam_1_1:
          initialize("Rigid Urethane Foam 1_1", 0.024, 56.1, thickness);
          break;
        case Materials.RigidUrethaneFoam_1_2:
          initialize("Rigid Urethane Foam 1_2", 0.024, 44.0, thickness);
          break;
        case Materials.RigidUrethaneFoam_1_3:
          initialize("Rigid Urethane Foam 1_3", 0.026, 31.4, thickness);
          break;
        case Materials.RigidUrethaneFoam_2_1:
          initialize("Rigid Urethane Foam 2_1", 0.023, 56.1, thickness);
          break;
        case Materials.RigidUrethaneFoam_2_2:
          initialize("Rigid Urethane Foam 2_2", 0.023, 44.0, thickness);
          break;
        case Materials.RigidUrethaneFoam_2_3:
          initialize("Rigid Urethane Foam 2_3", 0.024, 31.4, thickness);
          break;
        case Materials.RigidUrethaneFoam_OnSite:
          initialize("Rigid Urethane Foam (OnSite)", 0.026, 49.8, thickness);
          break;
        case Materials.PolyethyleneFoam_A:
          initialize("Polyethylene Foam A", 0.038, 62.8, thickness);
          break;
        case Materials.PolyethyleneFoam_B:
          initialize("Polyethylene Foam B", 0.042, 62.8, thickness);
          break;
        case Materials.PhenolicFoam_1_1:
          initialize("Phenolic Foam 1_1", 0.033, 37.7, thickness);
          break;
        case Materials.PhenolicFoam_1_2:
          initialize("Phenolic Foam 1_2", 0.030, 37.7, thickness);
          break;
        case Materials.PhenolicFoam_2_1:
          initialize("Phenolic Foam 2_1", 0.036, 56.5, thickness);
          break;
        case Materials.PhenolicFoam_2_2:
          initialize("Phenolic Foam 2_2", 0.034, 56.5, thickness);
          break;
        case Materials.InsulationBoard_A:
          initialize("Insulation Board A", 0.049, 324.8, thickness);
          break;
        case Materials.TatamiBoard:
          initialize("Tatami Board", 0.045, 15.1, thickness);
          break;
        case Materials.SheathingInsulationBoard:
          initialize("Sheathing Insulation Board", 0.052, 390.1, thickness);
          break;
        case Materials.CelluloseFiberInsulation_1:
          initialize("Cellulose Fiber Insulation 1", 0.040, 37.7, thickness);
          break;
        case Materials.CelluloseFiberInsulation_2:
          initialize("Cellulose Fiber Insulation 2", 0.040, 62.8, thickness);
          break;
        case Materials.Soil:
          initialize("Soil", 1.047, 3340.0, thickness);
          break;
        case Materials.ExpandedPolystyrene:
          initialize("Expanded Polystyrene", 0.035, 300.0, thickness);
          break;
        case Materials.CoveringMaterial:
          initialize("Covering Material", 0.140, 1680.0, thickness);
          break;
        case Materials.Linoleum:
          initialize("Linoleum", 0.190, 1470.0, thickness);
          break;
        case Materials.Carpet:
          initialize("Carpet", 0.080, 318.0, thickness);
          break;
        case Materials.AsbestosPlate:
          initialize("Asbestos Plate", 1.200, 1820.0, thickness);
          break;
        case Materials.SealedAirGap:
          initialize("Sealed AirGap", 5.800, 0.838, thickness);
          break;
        case Materials.AirGap:
          initialize("Air Gap", 11.600, 0.838, thickness);
          break;
        case Materials.PolystyreneFoam:
          initialize("Polystyrene Foam", 0.035, 80.0, thickness);
          break;
        case Materials.StyreneFoam:
          initialize("Styrene Foam", 0.035, 10.0, thickness);
          break;
        case Materials.RubberTile:
          initialize("Rubber Tile", 0.400, 784.0, thickness);
          break;
        case Materials.Kawara:
          initialize("Kawara", 1.000, 1506.0, thickness);
          break;
        case Materials.LightweightConcrete:
          initialize("Lightweight Concrete", 0.780, 1607.0, thickness);
          break;
        case Materials.Asphalt:
          initialize("Asphalt", 0.110, 920.0, thickness);
          break;
        case Materials.FrexibleBoard:
          initialize("Frexible Board", 0.350, 1600.0, thickness);
          break;
        case Materials.CalciumSilicateBoard:
          initialize("Calcium Silicate Board", 0.130, 680.0, thickness);
          break;
        case Materials.PhenolicFoam:
          initialize("Phenolic Foam", 0.020, 37.7, thickness);
          break;
        case Materials.Granite:
          initialize("Granite", 4.300, 2.9, thickness);
          break;
        case Materials.AcrylicResin:
          initialize("Acrylic Resin", 0.210, 1666.0, thickness);
          break;
        default:
          throw new Exception("wall material is not defined");
      }
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="thermalConductivity">熱伝導率[W/(mK)]</param>
    /// <param name="volSpecificHeat">容積比熱[kJ/(m3K)]</param>
    /// <param name="thickness">壁層の厚み[m]</param>
    public WallLayer(string name, double thermalConductivity, double volSpecificHeat, double thickness)
    { initialize(name, thermalConductivity, volSpecificHeat, thickness); }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="thermalConductivity">熱伝導率[W/(mK)]</param>
    /// <param name="volSpecificHeat">容積比熱[kJ/(m3K)]</param>
    /// <param name="moistureConductivity">湿気伝導率[(kg/s)/((kg/kg)·m)]</param>
    /// <param name="voidage">空隙率[-]</param>
    /// <param name="kappa">吸湿係数[kg/m3]</param>
    /// <param name="nu">放湿係数[kg/(m3K)]</param>
    /// <param name="thickness">壁層の厚み[m]</param>
    public WallLayer(string name, double thermalConductivity, double volSpecificHeat,
      double moistureConductivity, double voidage, double kappa, double nu, double thickness)
    {
      initialize(name, thermalConductivity, volSpecificHeat, thickness);
      MoistureConductivity = moistureConductivity;
      WaterCapacity = 0.5 * voidage * thickness * MOISTAIR_DENSITY;
      KappaC = 0.5 * kappa * thickness;
      NuC = 0.5 * nu * thickness;
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="thermalConductivity">熱伝導率[W/(mK)]</param>
    /// <param name="volumetricSpecificHeat">容積比熱[kJ/(m3K)]</param>
    /// <param name="thickness">壁層の厚み[m]</param>
    protected void initialize
      (string name, double thermalConductivity, double volumetricSpecificHeat, double thickness)
    {
      Name = name;
      ThermalConductivity = thermalConductivity;
      VolSpecificHeat = volumetricSpecificHeat;
      Thickness = thickness;
      HeatConductance = ThermalConductivity / Thickness;
      HeatCapacity_F = HeatCapacity_B = 0.5 * VolSpecificHeat * Thickness * 1000d;
    }

    /// <summary>壁層端部の温度に応じて物性値を更新する</summary>
    /// <param name="temperature1">温度1</param>
    /// <param name="temperature2">温度2</param>
    /// <returns>物性値変更の有無</returns>
    public virtual bool UpdateState(double temperature1, double temperature2){ return false; }

    /// <summary>壁層端部の温湿度に応じて物性値を更新する</summary>
    /// <param name="temperature1">温度1</param>
    /// <param name="temperature2">温度2</param>
    /// <param name="humidity1">絶対湿度1</param>
    /// <param name="humidity2">絶対湿度2</param>
    /// <returns>物性値変更の有無</returns>
    public virtual bool UpdateState
      (double temperature1, double temperature2, double humidity1, double humidity2){ return false; }

    #endregion

    #region ICloneable実装

    /// <summary>複製を生成する</summary>
    /// <returns>複製</returns>
    public virtual object Clone() { return MemberwiseClone(); }

    #endregion

  }

  #region 読み取り専用interface

  /// <summary>読み取り専用の壁層</summary>
  public interface ImmutableWallLayer
  {

    #region プロパティ

    /// <summary>名称を設定・取得する</summary>
    string Name { get; }

    /// <summary>物性が変化しうるか</summary>
    bool IsVariableProperties { get; }

    /// <summary>熱伝導率[W/(mK)]を取得する</summary>
    double ThermalConductivity { get; }

    /// <summary>湿気伝導率[(kg/s) / ((kg/kg)·m)]を取得する</summary>
    double MoistureConductivity { get; }

    /// <summary>容積比熱[kJ/(m^3K)]を取得する</summary>
    double VolSpecificHeat { get; }

    /// <summary>熱コンダクタンス[W/(m2K)]を取得する</summary>
    double HeatConductance { get; }

    /// <summary>顕熱容量（F側）[J/(m2K)]を取得する</summary>
    double HeatCapacity_F { get; }

    /// <summary>顕熱容量（B側）[J/(m2K)]を取得する</summary>
    double HeatCapacity_B { get; }

    /// <summary>水分容量[kg/m2]を取得する</summary>
    double WaterCapacity { get; }

    /// <summary>単位絶対湿度差に対する吸湿係数[kg/m2]を取得する</summary>
    double KappaC { get; }

    /// <summary>単位温度差に対する放湿係数[kg/(m2K)]を取得する</summary>
    double NuC { get; }

    /// <summary>壁層の厚み[m]を取得する</summary>
    double Thickness { get; }

    #endregion

  }

  #endregion

}
