﻿/* Window.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Weather;

namespace Popolo.ThermalLoad
{
  /// <summary>窓</summary>
  public class Window: ImmutableWindow
  {

    #region 列挙型定義

    /// <summary>ガラス種別</summary>
    public enum GlassTypes
    {
      /// <summary>透明フロート</summary>
      Transparent,
      /// <summary>熱線吸収</summary>
      HeatAbsorbing,
      /// <summary>熱線反射</summary>
      HeatReflecting,
      /// <summary>Low-E</summary>
      LowEmissivity
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>入射角特性近似係数</summary>
    private double[][] tau_CF, tau_CB, rho_CF, rho_CB;

    /// <summary>中空層熱抵抗[(m2K)/W]</summary>
    private double[] agapRes;

    /// <summary>ガラス熱抵抗[(m2K)/W]</summary>
    private double[] glassRes;

    /// <summary>各層の光学特性
    /// 0:透過,1:反射,2:吸収
    /// F:正面,B:裏,Dir:直達,Dif:拡散</summary>
    private double[,] opFDir, opBDir, opFDif, opBDif;

    /// <summary>各層の吸収率リスト</summary>
    private double[] absFDir, absFDif, absBDif;

    /// <summary>各層のガラスの垂直入射時の透過率と反射率</summary>
    private double[,] taurhoF, taurhoB;

    /// <summary>従前の太陽高度と太陽方位</summary>
    private double lstAlt, lstOri;

    /// <summary>中空層日射遮蔽物リスト</summary>
    private IShadingDevice[] sDevices;

    /// <summary>窓面積[m2]</summary>
    private double area = 1;

    /// <summary>窓面積[m2]を設定・取得する</summary>
    public double Area
    {
      set { if (0 < value) { area = value; } }
      get { return area; }
    }

    /// <summary>外表面側傾斜面を取得する</summary>
    public ImmutableIncline OutsideIncline { get; private set; }

    /// <summary>屋外からの直達日射の総合透過率[-]を取得する</summary>
    public double DirectSolarIncidentTransmittance { get; private set; }

    /// <summary>屋外からの直達日射の総合反射率[-]を取得する</summary>
    public double DirectSolarIncidentReflectance { get; private set; }

    /// <summary>屋外からの直達日射の吸収日射取得率[-]を取得する</summary>
    public double DirectSolarIncidentAbsorptance { get; private set; }

    /// <summary>屋外からの拡散日射の総合透過率[-]を取得する</summary>
    public double DiffuseSolarIncidentTransmittance { get; private set; }

    /// <summary>屋外からの拡散日射の総合反射率[-]を取得する</summary>
    public double DiffuseSolarIncidentReflectance { get; private set; }

    /// <summary>屋外からの拡散日射の吸収日射取得率[-]を取得する</summary>
    public double DiffuseSolarIncidentAbsorptance { get; private set; }

    /// <summary>屋内からの拡散日射の総合透過率[-]を取得する</summary>
    public double DiffuseSolarLostTransmittance { get; private set; }

    /// <summary>屋内からの拡散日射の総合反射率[-]を取得する</summary>
    public double DiffuseSolarLostReflectance { get; private set; }

    /// <summary>屋内からの拡散日射の吸収日射取得率[-]を取得する</summary>
    public double DiffuseSolarLostAbsorptance { get; private set; }

    /// <summary>ガラスの層数を取得する</summary>
    public int GlazingNumber { get; private set; }
    
    /// <summary>日除けを設定・取得する</summary>
    public SunShade SunShade { get; set; }

    /// <summary>F側とB側の対流・放射熱伝達率[W/(m2K)]</summary>
    private double cCoefF, rCoefF, cCoefB, rCoefB;

    /// <summary>F側対流熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double ConvectiveCoefficientF
    {
      get { return cCoefF; }
      set
      {
        cCoefF = value;
        updateFilmCoefficent();
      }
    }

    /// <summary>F側放射熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double RadiativeCoefficientF
    {
      get { return rCoefF; }
      set
      {
        rCoefF = value;
        updateFilmCoefficent();
      }
    }

    /// <summary>F側総合熱伝達率[W/(m2K)]を取得する</summary>
    public double FilmCoefficientF
    { get { return 1d / (2 * agapRes[0]); } }

    /// <summary>F側の短波長放射率[-]を取得する</summary>
    public double ShortWaveEmissivityF
    { get { throw new Exception("Not Implemented"); } }

    /// <summary>F側の長波長放射率[-]を設定・取得する</summary>
    public double LongWaveEmissivityF { get; set; } = 0.9;

    /// <summary>F側相当温度を設定・取得する</summary>
    public double SolAirTemperatureF { get; set; }

    /// <summary>B側対流熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double ConvectiveCoefficientB
    {
      get { return cCoefB; }
      set
      {
        cCoefB = value;
        updateFilmCoefficent();
      }
    }

    /// <summary>B側放射熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double RadiativeCoefficientB
    {
      get { return rCoefB; }
      set
      {
        rCoefB = value;
        updateFilmCoefficent();
      }
    }

    /// <summary>B側総合熱伝達率[W/(m2K)]を取得する</summary>
    public double FilmCoefficientB
    { get { return 1d / (2 * agapRes[agapRes.Length - 1]); } }

    /// <summary>B側の短波長放射率[-]を取得する</summary>
    public double ShortWaveEmissivityB
    { get { return 1 - DiffuseSolarLostReflectance; } }

    /// <summary>B側の長波長放射率[-]を設定・取得する</summary>
    public double LongWaveEmissivityB { get; set; } = 0.9;

    /// <summary>B側相当温度を設定・取得する</summary>
    public double SolAirTemperatureB { get; set; }

    /// <summary>内側表面を取得する</summary>
    internal wallWindowSurface InsideSurface { get; private set; }

    /// <summary>外側表面を取得する</summary>
    internal wallWindowSurface OutsideSurface { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>N層の多層ガラスの新しいインスタンスを初期化する</summary>
    /// <param name="area">面積[m2]</param>
    /// <param name="transmittanceF">正面側透過率リスト[-]（0=F, N-1=B）</param>
    /// <param name="reflectanceF">正面側反射率リスト[-]（0=F, N-1=B）</param>
    /// <param name="transmittanceB">裏側透過率リスト[-]（0=F, N-1=B）</param>
    /// <param name="reflectanceB">裏側反射率リスト[-]（0=F, N-1=B）</param>
    /// <param name="outsideIncline">外側傾斜面</param>
    public Window(double area, double[] transmittanceF, double[] reflectanceF, 
      double[] transmittanceB, double[] reflectanceB, ImmutableIncline outsideIncline)
    {
      Area = area;
      GlazingNumber = transmittanceF.Length;
      this.SunShade = SunShade.MakeEmptySunShade();
      this.OutsideIncline = outsideIncline;
      lstAlt = lstOri = -999;

      tau_CF = new double[GlazingNumber][];
      tau_CB = new double[GlazingNumber][];
      rho_CF = new double[GlazingNumber][];
      rho_CB = new double[GlazingNumber][];
      taurhoF = new double[GlazingNumber, 2];
      taurhoB = new double[GlazingNumber, 2];
      opFDir = new double[GlazingNumber * 2 + 1, 3];
      opBDir = new double[GlazingNumber * 2 + 1, 3];
      opFDif = new double[GlazingNumber * 2 + 1, 3];
      opBDif = new double[GlazingNumber * 2 + 1, 3];
      absFDir = new double[GlazingNumber * 2 + 1];
      absFDif = new double[GlazingNumber * 2 + 1];
      absBDif = new double[GlazingNumber * 2 + 1];
      agapRes = new double[GlazingNumber * 2 + 2];
      glassRes = new double[GlazingNumber];

      //空の日射遮蔽で初期化
      sDevices = new IShadingDevice[GlazingNumber + 1];
      for (int i = 0; i < sDevices.Length; i++)
      {
        sDevices[i] = new EmptyShadingDevice();
        int i2 = i * 2;
        opFDir[i2, 0] = opBDir[i2, 0] = opFDif[i2, 0] = opBDif[i2, 0] = 1.0;
        opFDir[i2, 1] = opBDir[i2, 1] = opFDif[i2, 1] = opBDif[i2, 1] =
          opFDir[i2, 2] = opBDir[i2, 2] = opFDif[i2, 2] = opBDif[i2, 2] = 0.0;
      }

      //垂直入射時の透過率と反射率を保存
      for (int i = 0; i < GlazingNumber; i++)
      {
        taurhoF[i, 0] = transmittanceF[i];
        taurhoB[i, 0] = transmittanceB[i];
        taurhoF[i, 1] = reflectanceF[i];
        taurhoB[i, 1] = reflectanceB[i];
      }

      //熱抵抗を初期化
      RadiativeCoefficientF = RadiativeCoefficientB = 4.5;
      ConvectiveCoefficientF = 18.5;
      ConvectiveCoefficientB = 7.5;
      updateFilmCoefficent();
      agapRes[agapRes.Length - 1] = agapRes[agapRes.Length - 2] = 0.5 * 1 / 6.7;
      for (int i = 2; i < agapRes.Length - 2; i++) agapRes[i] = 0.5 * 1 / 6.7;
      for (int i = 0; i < glassRes.Length; i++) glassRes[i] = 0.006;

      //内外表面作成
      OutsideSurface = new wallWindowSurface(this, true);
      InsideSurface = new wallWindowSurface(this, false);

      //入射角特性を透明フロートガラスで初期化
      for (int i = 0; i < GlazingNumber; i++) SetAngleDependence(i, GlassTypes.Transparent);
    }

    /// <summary>N層の多層ガラスの新しいインスタンスを初期化する</summary>
    /// <param name="area">面積[m2]</param>
    /// <param name="transmittance">透過率リスト[-]（0=F, N-1=B）</param>
    /// <param name="reflectance">反射率リスト[-]（0=F, N-1=B）</param>
    /// <param name="outsideIncline">外側傾斜面</param>
    public Window(double area, double[] transmittance, double[] reflectance, ImmutableIncline outsideIncline) :
      this(area, transmittance, reflectance, transmittance, reflectance, outsideIncline)
    { }

    #endregion

    #region インスタンスメソッド

    /// <summary>光学特性を更新する</summary>
    /// <param name="sun">太陽</param>
    public void UpdateOpticalProperties(ImmutableSun sun)
    {
      //ガラス・日射遮蔽物単体の光学特性の更新処理//////////////////////////////
      double cos = OutsideIncline.GetDirectSolarRadiationRate(sun);
      if (sun.Altitude <= 0)
      {
        DirectSolarIncidentTransmittance = 0;
        DirectSolarIncidentReflectance = 1;
        DirectSolarIncidentAbsorptance = 0;
        return;
      }
      bool sunMoved = (sun.Altitude != lstAlt || sun.Orientation != lstOri);
      if (sunMoved)
      {
        lstAlt = sun.Altitude;
        lstOri = sun.Orientation;

        //ガラスの直達日射入射角特性を反映
        if (0 < cos)
        {
          for (int i = 0; i < GlazingNumber; i++)
          {
            double tauF, tauB, rhoF, rhoB;
            tauF = tauB = rhoF = rhoB = 0;
            int lenth = tau_CF[i].Length - 1;
            for (int j = lenth; 0 <= j; j--)
            {
              tauF = cos * (tauF + tau_CF[i][j]);
              tauB = cos * (tauB + tau_CB[i][j]);
              rhoF = cos * (rhoF + rho_CF[i][j]);
              rhoB = cos * (rhoB + rho_CB[i][j]);
            }
            opFDir[2 * i + 1, 0] = tauF * taurhoF[i, 0];
            opFDir[2 * i + 1, 1] = 1 - (1 - taurhoF[i, 1]) * rhoF;
            opFDir[2 * i + 1, 2] = 1 - (opFDir[2 * i + 1, 0] + opFDir[2 * i + 1, 1]);
            opBDir[2 * i + 1, 0] = tauB * taurhoB[i, 0];
            opBDir[2 * i + 1, 1] = 1 - (1 - taurhoB[i, 1]) * rhoB;
            opBDir[2 * i + 1, 2] = 1 - (opBDir[2 * i + 1, 0] + opBDir[2 * i + 1, 1]);
          }
        }

        //日射遮蔽物のプロファイル角を更新
        double pAngle = OutsideIncline.GetProfileAngle(sun);
        for (int i = 0; i < sDevices.Length; i++) sDevices[i].ProfileAngle = pAngle;
      }

      //日射遮蔽物の光学特性を更新
      bool sPropChanged = false;
      for (int i = 0; i < sDevices.Length; i++)
      {
        int i2 = i * 2;
        if (sDevices[i].HasPropertyChanged)
        {
          sDevices[i].ComputeOpticalProperties(false, true, out opFDir[i2, 0], out opFDir[i2, 1]);
          sDevices[i].ComputeOpticalProperties(false, false, out opBDir[i2, 0], out opBDir[i2, 1]);
          sDevices[i].ComputeOpticalProperties(true, true, out opFDif[i2, 0], out opFDif[i2, 1]);
          sDevices[i].ComputeOpticalProperties(true, false, out opBDif[i2, 0], out opBDif[i2, 1]);
          opFDir[i2, 2] = 1 - (opFDir[i2, 0] + opFDir[i2, 1]);
          opBDir[i2, 2] = 1 - (opBDir[i2, 0] + opBDir[i2, 1]);
          opFDif[i2, 2] = 1 - (opFDif[i2, 0] + opFDif[i2, 1]);
          opBDif[i2, 2] = 1 - (opBDif[i2, 0] + opBDif[i2, 1]);
          sPropChanged = true;
        }
      }

      //太陽位置と日射遮蔽物光学特性不変の場合は終了
      if (!sunMoved && !sPropChanged) return;

      //総合光学特性の更新処理//////////////////////////////////////////////////
      double adF, adB;
      double ttlTF, ttlRF, ttlTB, ttlRB;
      //直達日射に関する特性を更新
      if (0 < cos)
      {
        double[] bf = new double[absFDir.Length];
        computeTotalOProperties
          (opFDir, opBDir, out ttlTF, out ttlRF, ref absFDir, out ttlTB, out ttlRB, ref bf);
        DirectSolarIncidentTransmittance = ttlTF;
        DirectSolarIncidentReflectance = ttlRF;
        integrateAbsorption(absFDir, agapRes, glassRes, out adF, out adB);
        DirectSolarIncidentAbsorptance = adB;
      }
      else
      {
        DirectSolarIncidentTransmittance = 0;
        DirectSolarIncidentReflectance = 1;
        DirectSolarIncidentAbsorptance = 0;
      }

      //遮蔽物光学特性特性変化時には拡散日射に関する特性を更新
      if (sPropChanged) updateDiffuseTotalProperties();
    }

    /// <summary>拡散日射に関する総合特性を更新する</summary>
    private void updateDiffuseTotalProperties()
    {
      double adF, adB, ttlTF, ttlRF, ttlTB, ttlRB;
      computeTotalOProperties
        (opFDif, opBDif, out ttlTF, out ttlRF, ref absFDif, out ttlTB, out ttlRB, ref absBDif);
      DiffuseSolarIncidentTransmittance = ttlTF;
      DiffuseSolarIncidentReflectance = ttlRF;
      DiffuseSolarLostTransmittance = ttlTF;
      DiffuseSolarLostReflectance = ttlRF;
      integrateAbsorption(absFDif, agapRes, glassRes, out adF, out adB);
      DiffuseSolarIncidentAbsorptance = adB;
      integrateAbsorption(absBDif, agapRes, glassRes, out adF, out adB);
      DiffuseSolarLostAbsorptance = adB;
    }

    /// <summary>窓全体の熱抵抗[m2K/W]を取得する</summary>
    /// <returns>窓全体の熱抵抗[m2K/W]</returns>
    public double GetResistance()
    {
      double rg = 0;
      for (int i = 0; i < glassRes.Length; i++) rg += glassRes[i];
      for (int i = 0; i < agapRes.Length; i++) rg += agapRes[i];
      return rg;
    }

    /// <summary>総合特性を計算する</summary>
    /// <param name="opPropF">F側入射に対する光学特性</param>
    /// <param name="opPropB">B側入射に対する光学特性</param>
    /// <param name="ttlTF">出力:F側入射に対する総合透過率</param>
    /// <param name="ttlRF">出力:F側入射に対する総合反射率</param>
    /// <param name="ttlAF">出力:F側入射に対する総合吸収率</param>
    /// <param name="ttlTB">出力:B側入射に対する総合透過率</param>
    /// <param name="ttlRB">出力:B側入射に対する総合反射率</param>
    /// <param name="ttlAB">出力:B側入射に対する総合吸収率</param>
    private static void computeTotalOProperties
      (double[,] opPropF, double[,] opPropB, out double ttlTF, out double ttlRF, ref double[] ttlAF,
      out double ttlTB, out double ttlRB, ref double[] ttlAB)
    {
      int ln = opPropF.GetLength(0);
      double[] ttlTBi = new double[ln];
      double[] ttlRFi = new double[ln];

      ttlTF = opPropF[ln - 1, 0];
      ttlRFi[ln - 1] = opPropF[ln - 1, 1];
      ttlTBi[ln - 1] = opPropB[ln - 1, 0];
      ttlRB = opPropB[ln - 1, 1];
      for (int i = ln - 2; 0 <= i; i--)
      {
        double xr = 1 / (1 - ttlRB * opPropF[i, 1]);
        ttlRFi[i] = ttlRFi[i + 1] + ttlTF * xr * opPropF[i, 1] * ttlTBi[i + 1];
        ttlRB = opPropB[i, 1] + opPropB[i, 0] * xr * ttlRB * opPropF[i, 0];
        ttlTF = ttlTF * xr * opPropF[i, 0];
        ttlTBi[i] = opPropB[i, 0] * xr * ttlTBi[i + 1];
      }

      ttlTF = opPropF[0, 0];
      ttlRF = opPropF[0, 1];
      ttlTB = opPropB[0, 0];
      ttlRB = opPropB[0, 1];
      ttlAF[0] = opPropF[0, 2];
      ttlAB[0] = 0;
      for (int j = 1; j < ln; j++)
      {
        double tfm1 = ttlTF;
        double tbm1 = ttlTB;
        double rbm1 = ttlRB;
        double xr = 1 / (1 - ttlRB * opPropF[j, 1]);
        ttlRF = ttlRF + ttlTF * xr * opPropF[j, 1] * ttlTB;
        ttlRB = opPropB[j, 1] + opPropB[j, 0] * xr * ttlRB * opPropF[j, 0];
        ttlTF = ttlTF * xr * opPropF[j, 0];
        ttlTB = opPropB[j, 0] * xr * ttlTB;
        double bf = 1 / (1 - rbm1 * ttlRFi[j]);
        ttlAF[j] = tfm1 * opPropF[j, 2] * bf;
        ttlAF[j - 1] += tfm1 * ttlRFi[j] * opPropB[j - 1, 2] * bf;
        ttlAB[j - 1] += ttlTBi[j] * opPropB[j - 1, 2] * bf;
        ttlAB[j] = ttlTBi[j] * rbm1 * opPropF[j, 2] * bf;
      }
    }

    #endregion

    #region モデル設定関連の処理

    /// <summary>日射遮蔽物を設定する</summary>
    /// <param name="number">設定する層番号（屋外側:0,屋内側:N+1）</param>
    /// <param name="sDevice">遮蔽物</param>
    public void SetShadingDevice(int number, IShadingDevice sDevice) { sDevices[number] = sDevice; }

    /// <summary>ガラスの熱抵抗[m2K/W]を設定する</summary>
    /// <param name="number">層番号</param>
    /// <param name="resistance">ガラスの熱抵抗[m2K/W]</param>
    public void SetGlassResistance(int number, double resistance)
    {
      glassRes[number] = resistance;
      updateAbsorptance();
    }

    /// <summary>中空層の熱抵抗[m2K/W]を設定する</summary>
    /// <param name="number">層番号（層の右側）</param>
    /// <param name="resistance">中空層の熱抵抗[m2K/W]</param>
    public void SetAirGapResistance(int number, double resistance)
    {
      agapRes[2 * number + 2] = agapRes[2 * number + 3] = 0.5 * resistance;
      updateAbsorptance();
    }

    /// <summary>総合熱伝達率[W/(m2K)]を更新する</summary>
    private void updateFilmCoefficent()
    {
      agapRes[0] = agapRes[1] = 0.5 / (ConvectiveCoefficientF + RadiativeCoefficientF);
      agapRes[agapRes.Length - 1] = 
        agapRes[agapRes.Length - 2] = 0.5 / (ConvectiveCoefficientB + RadiativeCoefficientB);
      updateAbsorptance();
    }

    /// <summary>吸収日射取得率を更新する</summary>
    private void updateAbsorptance()
    {
      double adF, adB;
      integrateAbsorption(absFDir, agapRes, glassRes, out adF, out adB);
      DirectSolarIncidentAbsorptance = adB;
      integrateAbsorption(absFDif, agapRes, glassRes, out adF, out adB);
      DiffuseSolarIncidentAbsorptance = adB;
      integrateAbsorption(absBDif, agapRes, glassRes, out adF, out adB);
      DiffuseSolarLostAbsorptance = adB;
    }

    /// <summary>総合吸収率をF側とB側に按分する</summary>
    /// <param name="ttlA">総合吸収率リスト</param>
    /// <param name="agapResist">中空層の熱抵抗リスト</param>
    /// <param name="glassResistance">ガラスの熱抵抗リスト</param>
    /// <param name="ttlAF">F側按分量</param>
    /// <param name="ttlAB">B側按分量</param>
    private static void integrateAbsorption
      (double[] ttlA, double[] agapResist, double[] glassResistance, out double ttlAF, out double ttlAB)
    {
      double rSum1, rSum2 = 0;
      rSum1 = rSum2 = ttlAF = ttlAB = 0;
      for (int i = 0; i < agapResist.Length; i++) rSum1 += agapResist[i];
      for (int i = 0; i < glassResistance.Length; i++) rSum1 += glassResistance[i];
      for (int i = 0; i < ttlA.Length; i++)
      {
        rSum2 += agapResist[i];
        if (i % 2 == 1) rSum2 += 0.5 * glassResistance[(i - 1) / 2];
        else if (i != 0) rSum2 += 0.5 * glassResistance[i / 2 - 1];
        double fRate = rSum2 / rSum1;
        ttlAB += ttlA[i] * fRate;
        ttlAF += ttlA[i] * (1 - fRate);
      }
    }

    #endregion

    #region 入射角特性関連の処理

    /// <summary>入射角特性を設定する</summary>
    /// <param name="layerNumber">層番号</param>
    /// <param name="coefTF">F側透過特性の近似係数</param>
    /// <param name="coefTB">B側透過特性の近似係数</param>
    /// <param name="coefRF">F側反射特性の近似係数</param>
    /// <param name="coefRB">B側反射特性の近似係数</param>
    public void SetAngleDependence
      (int layerNumber, double[] coefTF, double[] coefTB, double[] coefRF, double[] coefRB)
    {
      int ln = layerNumber;
      tau_CF[ln] = coefTF;
      tau_CB[ln] = coefTB;
      rho_CF[ln] = coefRF;
      rho_CB[ln] = coefRB;

      //拡散日射の規準化透過率と規準化反射率の計算
      double difCTF = 0;
      double difCTB = 0;
      double difCRF = 0;
      double difCRB = 0;
      for (int j = 0; j < tau_CF[ln].Length; j++)
      {
        difCTF += tau_CF[ln][j] / (j + 3);
        difCTB += tau_CB[ln][j] / (j + 3);
        difCRF += rho_CF[ln][j] / (j + 3);
        difCRB += rho_CB[ln][j] / (j + 3);
      }
      difCTF *= 2;
      difCTB *= 2;
      difCRF *= 2;
      difCRB *= 2;

      //拡散日射の透過率と反射率を計算
      opFDif[2 * ln + 1, 0] = difCTF * taurhoF[ln, 0];
      opFDif[2 * ln + 1, 1] = difCRF * taurhoF[ln, 1];
      opFDif[2 * ln + 1, 2] = 1 - (opFDif[2 * ln + 1, 0] + opFDif[2 * ln + 1, 1]);
      opBDif[2 * ln + 1, 0] = difCTB * taurhoB[ln, 0];
      opBDif[2 * ln + 1, 1] = difCRB * taurhoB[ln, 1];
      opBDif[2 * ln + 1, 2] = 1 - (opBDif[2 * ln + 1, 0] + opBDif[2 * ln + 1, 1]);

      //拡散日射に関する総合特性を更新
      updateDiffuseTotalProperties();
    }

    /// <summary>入射角特性を設定する</summary>
    /// <param name="layerNumber">層番号</param>
    /// <param name="type">ガラス種類</param>
    public void SetAngleDependence(int layerNumber, GlassTypes type)
    {
      int ln = layerNumber;
      switch (type)
      {
        case GlassTypes.HeatAbsorbing:
          SetAngleDependence(layerNumber,
            new double[] { 1.760, 3.770, -14.901, 16.422, -6.052 },
            new double[] { 1.760, 3.770, -14.901, 16.422, -6.052 },
            new double[] { 5.189, -12.392, 16.593, -11.851, 3.461 },
            new double[] { 5.189, -12.392, 16.593, -11.851, 3.461 }
            );
          return;
        case GlassTypes.HeatReflecting:
          SetAngleDependence(layerNumber,
            new double[] { 3.297, -1.122, -8.408, 12.206, -4.972 },
            new double[] { 3.297, -1.122, -8.408, 12.206, -4.972 },
            new double[] { 5.842, -15.264, -21.642, -15.948, 4.727 },
            new double[] { 5.842, -15.264, -21.642, -15.948, 4.727 }
            );
          return;
        case GlassTypes.LowEmissivity:
          SetAngleDependence(layerNumber,
            new double[] { 2.273, 1.631, -10.358, 11.769, -4.316 },
            new double[] { 2.273, 1.631, -10.358, 11.769, -4.316 },
            new double[] { 5.084, -12.646, 18.213, -13.967, 4.316 },
            new double[] { 4.387, -9.175, 11.152, -7.416, 2.052 }
            );
          return;
        default:
          SetAngleDependence(layerNumber,
            new double[] { 2.552, 1.364, -11.388, 13.617, -5.146 },
            new double[] { 2.552, 1.364, -11.388, 13.617, -5.146 },
            new double[] { 5.189, -12.392, 16.593, -11.851, 3.461 },
            new double[] { 5.189, -12.392, 16.593, -11.851, 3.461 }
            );
          return;
      }
    }

    #endregion

  }

  #region 読み取り専用窓

  /// <summary>読み取り専用の窓</summary>
  public interface ImmutableWindow
  {
    /// <summary>窓面積[m2]を設定・取得する</summary>
    double Area { get; }

    /// <summary>外表面側傾斜面を取得する</summary>
    ImmutableIncline OutsideIncline { get; }

    /// <summary>屋外からの直達日射の総合透過率[-]を取得する</summary>
    double DirectSolarIncidentTransmittance { get; }

    /// <summary>屋外からの直達日射の総合反射率[-]を取得する</summary>
    double DirectSolarIncidentReflectance { get; }

    /// <summary>屋外からの直達日射の吸収日射取得率[-]を取得する</summary>
    double DirectSolarIncidentAbsorptance { get; }

    /// <summary>屋外からの拡散日射の総合透過率[-]を取得する</summary>
    double DiffuseSolarIncidentTransmittance { get; }

    /// <summary>屋外からの拡散日射の総合反射率[-]を取得する</summary>
    double DiffuseSolarIncidentReflectance { get; }

    /// <summary>屋外からの拡散日射の吸収日射取得率[-]を取得する</summary>
    double DiffuseSolarIncidentAbsorptance { get; }

    /// <summary>屋内からの拡散日射の総合透過率[-]を取得する</summary>
    double DiffuseSolarLostTransmittance { get; }

    /// <summary>屋内からの拡散日射の総合反射率[-]を取得する</summary>
    double DiffuseSolarLostReflectance { get; }

    /// <summary>屋内からの拡散日射の吸収日射取得率[-]を取得する</summary>
    double DiffuseSolarLostAbsorptance { get; }

    /// <summary>ガラスの層数を取得する</summary>
    int GlazingNumber { get; }

    /// <summary>日除けを取得する</summary>
    SunShade SunShade { get; }

    /// <summary>F側相当温度を取得する</summary>
    double SolAirTemperatureF { get; }

    /// <summary>B側相当温度を取得する</summary>
    double SolAirTemperatureB { get; }

    /// <summary>F側対流熱伝達率[W/(m2K)]を設定・取得する</summary>
    double ConvectiveCoefficientF { get; }

    /// <summary>B側対流熱伝達率[W/(m2K)]を設定・取得する</summary>
    double ConvectiveCoefficientB { get; }

    /// <summary>F側放射熱伝達率[W/(m2K)]を設定・取得する</summary>
    double RadiativeCoefficientF { get; }

    /// <summary>B側放射熱伝達率[W/(m2K)]を設定・取得する</summary>
    double RadiativeCoefficientB { get; }

    /// <summary>F側総合熱伝達率[W/(m2K)]を取得する</summary>
    double FilmCoefficientF { get; }

    /// <summary>B側総合熱伝達率[W/(m2K)]を取得する</summary>
    double FilmCoefficientB { get; }

    /// <summary>F側の短波長放射率[-]を取得する</summary>
    double ShortWaveEmissivityF { get; }

    /// <summary>F側の長波長放射率[-]を設定・取得する</summary>
    double LongWaveEmissivityF { get; }

    /// <summary>B側の短波長放射率[-]を取得する</summary>
    double ShortWaveEmissivityB { get; }

    /// <summary>B側の長波長放射率[-]を設定・取得する</summary>
    double LongWaveEmissivityB { get; }
  }

  #endregion

}
