﻿/* wallWindowSurface.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Weather;

namespace Popolo.ThermalLoad
{
  /// <summary>壁・窓表面</summary>
  internal class wallWindowSurface
  {

    #region インスタンス変数・プロパティ

    /// <summary>F側か否か</summary>
    private bool isSideF;

    /// <summary>壁表面か否かを取得する</summary>
    public bool IsWall { get; private set; }

    /// <summary>壁を取得する</summary>
    public Wall Wall { get; private set; }

    /// <summary>窓を取得する</summary>
    public Window Window { get; private set; }

    /// <summary>番号を設定・取得する</summary>
    public int Index { get; set; }

    /// <summary>地中壁か否か</summary>
    public bool IsGroundWall { get; set; } = false;

    /// <summary>隣室温度差係数[-]を取得する</summary>
    public double AdjacentSpaceFactor { get; set; } = -1.0;

    /// <summary>傾斜面を設定・取得する</summary>
    public ImmutableIncline Incline { get; set; }

    /// <summary>属するゾーン番号を設定・取得する</summary>
    public int ZoneIndex { get; set; }

    /// <summary>裏側表面を取得する</summary>
    public wallWindowSurface ReverseSideSurface
    {
      get
      {
        if (IsWall && isSideF) return Wall.SurfaceB;
        else if (IsWall && !isSideF) return Wall.SurfaceF;
        else if (!IsWall && isSideF) return Window.InsideSurface;
        else return Window.OutsideSurface;
      }
    }

    /// <summary>表面積[m2]を取得する</summary>
    public double Area
    {
      get
      {
        if (IsWall) return Wall.Area;
        else return Window.Area;
      }
    }

    /// <summary>総合熱伝達率[W/(m2K)]を取得する</summary>
    public double FilmCoefficient
    {
      get
      {
        if (IsWall && isSideF) return Wall.FilmCoefficientF;
        else if (IsWall && !isSideF) return Wall.FilmCoefficientB;
        else if (!IsWall && isSideF) return Window.FilmCoefficientF;
        else return Window.FilmCoefficientB;
      }
    }

    /// <summary>放射熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double RadiativeCoefficient
    {
      get
      {
        if (IsWall && isSideF) return Wall.RadiativeCoefficientF;
        else if (IsWall && !isSideF) return Wall.RadiativeCoefficientB;
        else if (!IsWall && isSideF) return Window.RadiativeCoefficientF;
        else return Window.RadiativeCoefficientB;
      }
      set
      {
        if (IsWall && isSideF) Wall.RadiativeCoefficientF = value;
        else if (IsWall && !isSideF) Wall.RadiativeCoefficientB = value;
        else if (!IsWall && isSideF) Window.RadiativeCoefficientF = value;
        else Window.RadiativeCoefficientB = value;
      }
    }

    /// <summary>対流熱伝達率[W/(m2K)]を設定・取得する</summary>
    public double ConvectiveCoefficient
    {
      get
      {
        if (IsWall && isSideF) return Wall.ConvectiveCoefficientF;
        else if (IsWall && !isSideF) return Wall.ConvectiveCoefficientB;
        else if (!IsWall && isSideF) return Window.ConvectiveCoefficientF;
        else return Window.ConvectiveCoefficientB;
      }
      set
      {
        if (IsWall && isSideF) Wall.ConvectiveCoefficientF = value;
        else if (IsWall && !isSideF) Wall.ConvectiveCoefficientB = value;
        else if (!IsWall && isSideF) Window.ConvectiveCoefficientF = value;
        else Window.ConvectiveCoefficientB = value;
      }
    }

    /// <summary>湿気伝達率[(kg/s)/((kg/kg)m2)]を取得する</summary>
    public double MoistureCoefficient
    {
      get
      {
        if (IsWall && isSideF) return Wall.MoistureCoefficientF;
        else if (IsWall && !isSideF) return Wall.MoistureCoefficientB;
        else return 0;
      }
    }

    /// <summary>短波長放射率[-]を取得する</summary>
    public double ShortWaveEmissivity
    {
      get
      {
        if (IsWall && isSideF) return Wall.ShortWaveAbsorptanceF;
        else if (IsWall && !isSideF) return Wall.ShortWaveAbsorptanceB;
        else if (!IsWall && isSideF) return Window.ShortWaveEmissivityF;
        else return Window.ShortWaveEmissivityB;
      }
    }

    /// <summary>長波長放射率[-]を取得する</summary>
    public double LongWaveEmissivity {
      get
      {
        if (IsWall && isSideF) return Wall.LongWaveEmissivityF;
        else if (IsWall && !isSideF) return Wall.LongWaveEmissivityB;
        else if (!IsWall && isSideF) return Window.LongWaveEmissivityF;
        else return Window.LongWaveEmissivityB;
      }
    }

    /// <summary>相当温度[C]を設定・取得する</summary>
    public double SolAirTemperature
    {
      set
      {
        if (IsWall && isSideF) Wall.SolAirTemperatureF = value;
        else if (IsWall && !isSideF) Wall.SolAirTemperatureB = value;
        else if (!IsWall && isSideF) Window.SolAirTemperatureF = value;
        else Window.SolAirTemperatureB = value;
      }
      get
      {
        if (IsWall && isSideF) return Wall.SolAirTemperatureF;
        else if (IsWall && !isSideF) return Wall.SolAirTemperatureB;
        else if (!IsWall && isSideF) return Window.SolAirTemperatureF;
        else return Window.SolAirTemperatureB;
      }
    }

    /// <summary>表面温度[C]を取得する</summary>
    public double SurfaceTemperature
    {
      get
      {
        if (IsWall) return IF2 + FFS2 * SolAirTemperature + BFS2 * ReverseSideSurface.SolAirTemperature
            + FFL2 * HumidityRatio + BFL2 * ReverseSideSurface.HumidityRatio;
        else return FFS2 * SolAirTemperature + BFS2 * ReverseSideSurface.SolAirTemperature;
      }
    }

    /// <summary>絶対湿度[kg/kg]を設定・取得する</summary>
    public double HumidityRatio
    {
      set
      {
        if (IsWall && isSideF) Wall.HumidityRatioF = value;
        else if (IsWall && !isSideF) Wall.HumidityRatioB = value;
        else return;
      }
      get
      {
        if (IsWall && isSideF) return Wall.HumidityRatioF;
        else if (IsWall && !isSideF) return Wall.HumidityRatioB;
        else return 0;
      }
    }

    /// <summary>表側の相当温度の係数（温度）を取得する</summary>
    public double FFS2
    {
      get
      {
        if (IsWall && isSideF) return Wall.FFS2_F;
        else if (IsWall && !isSideF) return Wall.BFS2_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 1 - 1d / (Window.GetResistance() * Window.FilmCoefficientB);
      }
    }

    /// <summary>表側の相当温度の係数（湿度）を取得する</summary>
    public double FFS3
    {
      get
      {
        if (IsWall && isSideF) return Wall.FFS3_F;
        else if (IsWall && !isSideF) return Wall.BFS3_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>表側の絶対湿度の係数（温度）を取得する</summary>
    public double FFL2
    {
      get
      {
        if (IsWall && isSideF) return Wall.FFL2_F;
        else if (IsWall && !isSideF) return Wall.BFL2_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>表側の絶対湿度の係数（湿度）を取得する</summary>
    public double FFL3
    {
      get
      {
        if (IsWall && isSideF) return Wall.FFL3_F;
        else if (IsWall && !isSideF) return Wall.BFL3_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>裏側の相当温度の係数（温度）を取得する</summary>
    public double BFS2
    {
      get
      {
        if (IsWall && isSideF) return Wall.BFS2_F;
        else if (IsWall && !isSideF) return Wall.FFS2_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 1d / (Window.GetResistance() * Window.FilmCoefficientB);
      }
    }

    /// <summary>裏側の相当温度の係数（湿度）を取得する</summary>
    public double BFS3
    {
      get
      {
        if (IsWall && isSideF) return Wall.BFS3_F;
        else if (IsWall && !isSideF) return Wall.FFS3_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>裏側の絶対湿度の係数（温度）を取得する</summary>
    public double BFL2
    {
      get
      {
        if (IsWall && isSideF) return Wall.BFL2_F;
        else if (IsWall && !isSideF) return Wall.FFL2_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>裏側の絶対湿度の係数（湿度）を取得する</summary>
    public double BFL3
    {
      get
      {
        if (IsWall && isSideF) return Wall.BFL3_F;
        else if (IsWall && !isSideF) return Wall.FFL3_B;
        else if (!IsWall && isSideF) throw new Exception("Not Implemented");
        else return 0;
      }
    }

    /// <summary>時間遅れ要素の係数（温度）を取得する</summary>
    public double IF2
    {
      get
      {
        if (IsWall && isSideF) return Wall.IF2_F;
        else if (IsWall && !isSideF) return Wall.IF2_B;
        else if (!IsWall && isSideF) return 0;
        else return 0;
      }
    }

    /// <summary>時間遅れ要素の係数（温度）を取得する</summary>
    public double IF3
    {
      get
      {
        if (IsWall && isSideF) return Wall.IF3_F;
        else if (IsWall && !isSideF) return Wall.IF3_B;
        else if (!IsWall && isSideF) return 0;
        else return 0;
      }
    }

    /// <summary>総合熱伝達率に占める対流熱伝達率の比率[-]を取得する</summary>
    public double ConvectiveRate
    { get { return ConvectiveCoefficient / FilmCoefficient; } }

    /// <summary>総合熱伝達率に占める放射熱伝達率の比率[-]を取得する</summary>
    public double RadiativeRate { get { return 1 - ConvectiveRate; } }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="wall">壁</param>
    /// <param name="isSideF">F側表面か否か</param>
    public wallWindowSurface(Wall wall, bool isSideF)
    {
      IsWall = true;
      Wall = wall;
      this.isSideF = isSideF;
      Index = -1;
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="window">窓</param>
    /// <param name="isSideF">F側表面か否か</param>
    public wallWindowSurface(Window window, bool isSideF)
    {
      IsWall = false;
      Window = window;
      this.isSideF = isSideF;
      Index = -1;
    }

    #endregion

  }
}
